
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gbmh.cpp,v 1.1.1.1 1998/11/21 19:38:05 odinn Exp $
//  ------------------------------------------------------------------
//  Boyer-Moore-Horspool pattern match.
//  ------------------------------------------------------------------
//  Based on Public Domain version by Thad Smith 7/21/1992, based on a
//  7/92 public domain BMH version by Raymond Gardner.
//  With fixes found by Jeff Dunlop 10/21/93.
//  ------------------------------------------------------------------


#include <ctype.h>
#include <limits.h>
#include <gstrall.h>
#include <gmemdbg.h>
#include <gbmh.h>


//  ------------------------------------------------------------------

gbmh::gbmh() {

  pat = NULL;
}


//  ------------------------------------------------------------------

gbmh::~gbmh() {

  delete[] pat;
}


//  ------------------------------------------------------------------

void gbmh::init(const char* pattern, bool ignorecase) {

  ignore_case = ignorecase;
  patlen = strlen(pattern);

  delete[] pat;
  pat = new char [patlen+1];
  throw_new(pat);

  // Copy pattern
  strcpy(pat, pattern);
  if(ignore_case)
    strupr(pat);

  // initialize skip array
  int i;
  for(i=0; i<256; i++)
    skip[i] = patlen;
  for(i=0; i<patlen; i++) {
    skip[pat[i]] = patlen - i - 1;
    if(ignore_case)
      skip[tolower(pat[i])] = patlen - i - 1;
  }
  char lastpatchar = pat[patlen - 1];
  skip[lastpatchar] = INT_MAX;
  if(ignore_case)
    skip[tolower(lastpatchar)] = INT_MAX;

  // Horspool's fixed second shift
  skip2 = patlen;
  for(i=0; i<patlen-1; ++i)
    if(pat[i] == lastpatchar)
      skip2 = patlen - i - 1;
}


//  ------------------------------------------------------------------

bool gbmh::find(const char* buffer) {

  int buflen = strlen(buffer);

  int i = patlen - 1 - buflen;
  if(i >= 0)
    return false;

  buffer += buflen;

  while(1) {

    while((i += skip[buffer[i]]) < 0)
      ;

    if(i < (INT_MAX - buflen))
      return false;

    i -= INT_MAX;
    int j = patlen - 1;
    const char* s = buffer + (i - j);

    if(ignore_case) {
      while(--j >= 0 and toupper(s[j]) == pat[j])
        ;
    }
    else {
      while(--j >= 0 and s[j] == pat[j])
        ;
    }

    if(j < 0)
      return true;

    if((i += skip2) >= 0)
      return false;
  }
}


//  ------------------------------------------------------------------

#define BmhLARGE 32767                // flag for last character match

void strbmhinit(const char* __pat, BmhArg* __arg, int __ignore) {

  __arg->ignore = __ignore;

  int i, lastpatchar;
  int patlen = __arg->patlen = strlen(__pat);
  int* skip = __arg->skip;
  char* pat = __arg->pat;

  // Copy pattern

  if(__arg->ignore) {
    for(i=0; i<patlen; i++)
      pat[i] = (char)toupper(__pat[i]);
  }
  else {
    strcpy(pat, __pat);
  }

  // initialize skip array

  for(i=0; i<256; i++)
    skip[i] = patlen;

  if(__arg->ignore) {
    for(i=0; i<patlen-1; i++) {
      skip[        pat[i] ] = patlen - i - 1;
      skip[tolower(pat[i])] = patlen - i - 1;
    }
    lastpatchar = pat[patlen - 1];
    skip[        lastpatchar ] = BmhLARGE;
    skip[tolower(lastpatchar)] = BmhLARGE;
  }
  else {
    for(i=0; i<patlen; i++)
      skip[pat[i]] = patlen - i - 1;
    lastpatchar = pat[patlen - 1];
    skip[lastpatchar] = BmhLARGE;
  }

  __arg->skip2 = patlen;  // Horspool's fixed second shift

  for(i=0; i<patlen-1; ++i) {
    if(pat[i] == lastpatchar)
      __arg->skip2 = patlen - i - 1;
  }
}


//  ------------------------------------------------------------------

const char* strbmhsearch(const char* __buffer, int __buflen, const BmhArg* __arg) {

  int i, j;
  const char* s;
  const char* pat = __arg->pat;
  const int* skip = __arg->skip;
  int patlen = __arg->patlen;

  i = patlen - 1 - __buflen;
  if(i >= 0)
    return NULL;

  __buffer += __buflen;

  for(;;) {

    while((i += skip[((byte*)__buffer)[i]]) < 0)
      ;   // mighty fast inner loop

    if(i < (BmhLARGE - __buflen))
      return NULL;

    i -= BmhLARGE;
    j = patlen - 1;
    s = __buffer + (i - j);
    if(__arg->ignore) {
      while(--j >= 0 and toupper(s[j]) == pat[j])
        ;
    }
    else {
      while(--j >= 0 and s[j] == pat[j])
        ;
    }
    if(j < 0)
      return s;
    if((i += __arg->skip2) >= 0)
      return NULL;
  }
}


//  ------------------------------------------------------------------

