
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gfilutl1.cpp,v 1.1.1.1 1998/11/21 19:38:06 odinn Exp $
//  ------------------------------------------------------------------
//  File utility functions
//  ------------------------------------------------------------------


#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#if defined(__WATCOMC__)
#include <sys/utime.h>
#else
#include <utime.h>
#endif

#include <gio.h>
#include <gdos.h>
#include <gshare.h>
#include <gdirect.h>

#include <gfile.h>
#include <gtimall.h>
#include <gfilutil.h>

#if defined(__OS2__)
#define INCL_BASE
#include <os2.h>
#endif

#if defined(__WIN32__)
#include <windows.h>
#endif


//  ------------------------------------------------------------------

char* AddBackslash(char* p) {

  strchg(p, GOLD_WRONG_SLASH_CHR, GOLD_SLASH_CHR);
  if(p[strlen(p)-1] != GOLD_SLASH_CHR)
    strcat(p, GOLD_SLASH_STR);

  return p;
}


//  ------------------------------------------------------------------

char* StripBackslash(char* p) {

  int x = strlen(p) - 1;

  strchg(p, GOLD_WRONG_SLASH_CHR, GOLD_SLASH_CHR);
  if(p[x] == GOLD_SLASH_CHR)
    p[x] = NUL;

  return p;
}


//  ------------------------------------------------------------------
//  Get size of file

long GetFilesize(const char* file) {

  struct stat info;
  if(stat(file, &info) == 0)
    return info.st_size;
  return -1;
}


//  ------------------------------------------------------------------
//  Get timestamp of file

long GetFiletime(const char* file) {

  struct stat st;
  if(stat(file, &st) == 0) {
    ulong mtime = st.st_mtime;
    struct tm& f = *localtime((time_t*) &mtime);
    FFTime t;
    t.ft_year  = f.tm_year - 80;
    t.ft_month = f.tm_mon + 1;
    t.ft_day   = f.tm_mday;
    t.ft_hour  = f.tm_hour;
    t.ft_min   = f.tm_min;
    t.ft_tsec  = f.tm_sec / 2;
    return (long)t.number();
  }
  return 0;
}


//  ------------------------------------------------------------------
//  Get creation timestamp of file

#if 0
long GetFileCreationTime(const char* file) {

  struct stat st;
  if(stat(file, &st) == 0) {
    ulong ctime = st.st_ctime;
    struct tm& f = *glocaltime(&ctime);
    FFTime t;
    t.ft_year  = f.tm_year - 80;
    t.ft_month = f.tm_mon + 1;
    t.ft_day   = f.tm_mday;
    t.ft_hour  = f.tm_hour;
    t.ft_min   = f.tm_min;
    t.ft_tsec  = f.tm_sec / 2;
    return (long)t.number();
  }
  return 0;
}
#endif


//  ------------------------------------------------------------------
//  Check if file exists

int fexist(const char* filename) {
  
  return *filename ? ((access(filename, 0) == 0) and not is_dir(filename)) : false;
}


//  ------------------------------------------------------------------
//  Get size of open file

long fsize(FILE* fp) {

  dword pos,len=0L;

  if(fp) {
    fseek(fp, 0L, SEEK_CUR);  /* Dummy */
    pos = ftell(fp);
    fseek(fp, 0L, SEEK_END);
    len = ftell(fp);
    fseek(fp, pos, SEEK_SET);
  }
  if(len==0xFFFFFFFFL)
    len = 0L;

  return len;
}


//  ------------------------------------------------------------------
//  Compare timestamps of two files

long FiletimeCmp(const char* file1, const char* file2) {

  long ft1 = GetFiletime(file1);
  long ft2 = GetFiletime(file2);

  return ft1 - ft2;
}


//  ------------------------------------------------------------------
//  Check if a pathname is a directory

int is_dir(const char* path) {

  // Check if it's a root path (X:\)
  #if defined(GOLD_HAS_DRIVES)
  if((path[1] == ':') and (path[2] == '\\') and (path[3] == NUL))
    return true;  // The root is a directory
  #endif

  Path tmp;
  strcpy(tmp, path);
  StripBackslash(tmp);

  struct stat st;
  if(stat(tmp, &st) == 0)
    return (st.st_mode & S_IFDIR) ? true : false;
  return false;
}


//  ------------------------------------------------------------------
//  Add path to filename if no path is present

const char* AddPath(const char* path, const char* file) {

  if(strpbrk(file, "/\\")) {
    // Don't add path if the filename already contains one
    return file;
  }
  else if(*path and ((*file == '.') or strchr("/\\", path[strlen(path)-1]))) {
    // Build path+filename if path ends with a slash or backslash
    return stradd(path, file);
  }
  else {
    // Use filename in path or file
    return *path ? path : file;
  }
}


//  ------------------------------------------------------------------
//  Make path with filename

char* MakePath(char* pathfile, const char* path, const char* file) {

  return strcpy(pathfile, AddPath(path, file));
}


//  ------------------------------------------------------------------
//  Add path to filename, if no path is set

void MakePathname(char* pathname, const char* path, const char* name) {

  char newpath[GMAXPATH]; //, drive[GMAXDRIVE], dir[GMAXDIR], file[GMAXFILE], ext[GMAXEXT];
  Path tmpname;
  Path tmppath;
  strcpy(tmpname, name);
  strcpy(tmppath, path);

  if(strblank(tmpname)) {
    *pathname = NUL;
    return;
  }

  strbtrim(tmppath);
  strchg(pathname, GOLD_WRONG_SLASH_CHR, GOLD_SLASH_CHR);
  strchg(tmppath,  GOLD_WRONG_SLASH_CHR, GOLD_SLASH_CHR);
  strchg(tmpname,  GOLD_WRONG_SLASH_CHR, GOLD_SLASH_CHR);
  //_splitpath(tmpname, drive, dir, file, ext);
  if(strpbrk(tmpname, GOLD_SLASH_STR))
    strcpy(pathname, tmpname);
  else {
    strcpy(newpath, tmppath);
    if(*newpath)
      AddBackslash(newpath);
    strcpy(pathname, stradd(newpath, tmpname));
  }
}


//  ------------------------------------------------------------------

#if defined(__GNUC__)
int eof(int h) {

  return tell(h) > filelength(h);
}
#endif


//  ------------------------------------------------------------------
//  Shareable fopen() for compilers that need it

FILE *fsopen(const char *path, const char *type, int shflag) {

  return _fsopen(path, type, shflag);
}

#if defined(__GNUC__) // and not defined(__EMX__)

FILE* _fsopen(const char *path, const char *type, int shflag) {

  FILE* fp=NULL;
  int fh=-1, access=0, mode=S_STDRD, c, n;

  switch(type[0]) {
    case 'r':
      access |= (O_RDONLY|O_EXCL);
      break;
    case 'w':
      access |= (O_TRUNC|O_CREAT|O_WRONLY);
      mode |= S_STDRW;
      break;
    case 'a':
      access |= (O_APPEND|O_CREAT|O_WRONLY);
      mode |= S_STDRW;
      break;
    default:
      return NULL;
  }
  n=0;
  do {
    c = type[++n];
    switch(c) {
      case '+':
        access &= (~O_RDONLY);
        access &= (~O_WRONLY);
        access |= O_RDWR;
        break;
      case 'b':
        access |= O_BINARY;
        break;
      case 't':
        access |= O_TEXT;
        break;
      default:
        break;
    }
  } while((n < 3) && (c));

  fh = sopen(path, access, shflag, mode);
  if(fh != -1)
    fp = fdopen(fh, (char*)type);

  return fp;
}

#endif


//  ------------------------------------------------------------------

int ftrunc(FILE* fp, long size) {

  return chsize(fileno(fp),(size));
}


//  ------------------------------------------------------------------

void TouchFile(const char* filename) {

  if(not fexist(filename))
    close(open(filename, O_WRONLY|O_CREAT|O_TRUNC, S_STDRW));
  else {
    struct utimbuf ut;
    ut.actime = ut.modtime = time(NULL);
    utime(filename, &ut);
  }
}


//  ------------------------------------------------------------------

char* PathCopy(char* __dst, const char* __src) {

  return AddBackslash(strxcpy(__dst, __src, sizeof(Path)));
}


//  ------------------------------------------------------------------

int PathInvalid(const char* __path) {

  return strieql(__path, GFILE_INVALID) or (*__path == NUL);
}


//  ------------------------------------------------------------------

int TestLockPath(const char* __path) {

  int _canlock = false;

  Path _file;
  strcpy(_file, __path);
  strcat(_file, "GTMPXXXX.XXX");
  mktemp(_file);

  int _fh = sopen(_file, O_RDWR|O_CREAT|O_TRUNC|O_BINARY, SH_DENYNO, S_STDRW);
  if(_fh != -1) {
    if(lock(_fh, 0L, 1L) == -1)
      _canlock = NO;
    else {
      _canlock = YES;
      unlock(_fh, 0L, 1L);
    }
    close(_fh);
    remove(_file);
  }

  return _canlock;
}


//  ------------------------------------------------------------------

void flushfile(int fh) {

  int nfd = dup(fh);
  if(nfd != -1)
    close(nfd);
}


//  ------------------------------------------------------------------

const char* CleanFilename(const char* __file) {

  static const char* invalidfilename = "<invalid>";
  if((__file == NULL) or (__file == (char*)0xFFFFFFFFL) or (__file == (char*)0xEEEEEEEEL))
    return invalidfilename;

  const char* _tmp = strrchr(__file, GOLD_SLASH_CHR);
  return _tmp ? (_tmp+1) : __file;
}


//  ------------------------------------------------------------------

int ForceDeleteFile(const char* file) {

  #if defined(__OS2__)
  return DosForceDelete((PSZ)file);
  #else
  return remove(file);
  #endif
}


//  ------------------------------------------------------------------

void WipeFile(const char* file, int options) {

  uint n;
  byte buf[512];

  switch(options) {
    default:
      for(n=0; n<512; n++)
        buf[n] = (byte)(rand() % 256);
  }

  int fh = sopen(file, O_RDWR|O_BINARY, SH_DENYRW, S_STDRW);
  if(fh != -1) {
    uint blocks = (uint)(filelength(fh) / 512L) + 1;
    for(n=0; n<blocks; n++)
      write(fh, buf, 512);
    chsize(fh, 0);
    close(fh);
    ForceDeleteFile(file);
  }
}


//  ------------------------------------------------------------------

int strschg_environ(char* s) {

  char* b;
  char* e;
  char* replace;
  char find[41];
  char findrep[41];
  int replaced = 0;

  do {
    b = strchr(s, '%');
    if(b) {
      e = strchr(b+1, '%');
      if(e) {
        strxcpy(find, b+1, (size_t)(e-b));
        strxcpy(findrep, b, (size_t)(e-b)+2);
        replace = getenv(find);
        replace = replace ? replace : "";
        strschg(s, findrep, replace);
        replaced++;
      }
    }
  } while(b and e);

  replace = getenv("HOME");
  if(replace) {
    strschg(s, "~", replace);
    replaced++;
  }

  return replaced;
}


//  ------------------------------------------------------------------

int gchdir(const char* dir) {

  #if defined(GOLD_HAS_DRIVES)
  if(dir[1] == ':') {
    #if defined(__EMX__)
    _chdrive(*dir);
    #else
    uint drives;
    _dos_setdrive(toupper(*dir)-'@', &drives);
    #endif
  }
  #endif
  int e = chdir(dir);
  if(e) {
    Path p;
    strcpy(p, dir);
    StripBackslash(p);
    e = chdir(p);
  }
  return e;
}


//  ------------------------------------------------------------------

