
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gftnadrs.cpp,v 1.1.1.1 1998/11/21 19:38:06 odinn Exp $
//  ------------------------------------------------------------------
//  FidoNet Technology Network - 5D address class.
//  ------------------------------------------------------------------


#include <gstrgstr.h>
#include <gfile.h>
#include <gutlmisc.h>
#include <gftnadrs.h>


//  ------------------------------------------------------------------

gftnaddrs::gftnaddrs()                    { operator=(0); }
gftnaddrs::gftnaddrs(const char* s)       { operator=(s); }
gftnaddrs::gftnaddrs(const gstring& s)    { operator=(s); }
gftnaddrs::gftnaddrs(const ftn_addr& a)   { operator=(a); }
gftnaddrs::gftnaddrs(const gftnaddrs& a)  { operator=(a); }
gftnaddrs::gftnaddrs(const ftn_aka& a)    { operator=(a); }


//  ------------------------------------------------------------------

bool gftnaddrs::getaddrpart(word& part, const char* s) {

  if(*s == '*') {
    part = ftn::wildcard_all;
    return true;
  }
  else if(*s == '?') {
    part = ftn::wildcard_first;
    return true;
  }
  else if(isdigit(*s)) {
    part = (word)atow(s);
    return true;
  }
  return false;
}


//  ------------------------------------------------------------------

gftnaddrs& gftnaddrs::operator=(int n) {

  zone = net = node = point = (word)n;
  domain = "";
  return *this;
}


//  ------------------------------------------------------------------

gftnaddrs& gftnaddrs::operator=(const char* s) {

  reset();
  return set(s);
}


//  ------------------------------------------------------------------

gftnaddrs& gftnaddrs::operator=(const gstring& s) {

  reset();
  return set(s.c_str());
}


//  ------------------------------------------------------------------

gftnaddrs& gftnaddrs::operator=(const ftn_addr& a) {

  zone  = a.zone;
  net   = a.net;
  node  = a.node;
  point = a.point;
  return *this;
}


//  ------------------------------------------------------------------

gftnaddrs& gftnaddrs::operator=(const gftnaddrs& a) {

  zone   = a.zone;
  net    = a.net;
  node   = a.node;
  point  = a.point;
  domain = a.domain;
  return *this;
}


//  ------------------------------------------------------------------

gftnaddrs& gftnaddrs::operator=(const ftn_aka& a) {

  operator=(a.addr);
  domain = a.domain;
  return *this;
}


//  ------------------------------------------------------------------

gftnaddrs& gftnaddrs::set(const char* s) {

  const char* ptr = s;
  bool gotzone = false;
  bool gotnet = false;
  bool gotnode = false;
  bool gotpoint = false;
  bool gotdomain = false;

  if(*s == '*') {
    operator=(ftn::wildcard_all);
    domain = '*';
    ptr++;
  }

  while(not isspace(*ptr) and *ptr) {

    switch(*ptr) {

      case '<':
        return *this;

      case '{':
        while((*ptr != '}') and *ptr)
          ptr++;
        break;

      case '#':
        domain.copy(s, ptr++);
        gotdomain = true;
        s = ptr;
        break;

      case ':':
        gotzone = getaddrpart(zone, s);
        s = ++ptr;
        if(*s)
          gotnet = getaddrpart(net, s);
        break;

      case '/':
        if(not gotnet)
          gotnet = getaddrpart(net, s);
        s = ++ptr;
        if(*s)
          gotnode = getaddrpart(node, s);
        break;

      case '.':
        if((not gotnode) and (*s != '.'))
          gotnode = getaddrpart(node, s);
        s = ++ptr;
        if(*s)
          getaddrpart(point, s);
        gotpoint = true;
        break;

      case '@':
        if(not (gotzone or gotnet or gotnode or gotpoint))
          gotnode = getaddrpart(node, s);
        s = ++ptr;
        while(not isspace(*ptr) and *ptr)
          ptr++;
        domain.copy(s, ptr);
        gotdomain = true;
        break;

      default:
        ptr++;
    }
  }

  if(not (gotzone or gotnet or gotnode or gotpoint))
    gotnode = getaddrpart(node, s);

  if(gotzone and not gotnet)
    net = 0;

  if(gotzone and not gotnode) {
    node = (word)((net == (word)ftn::wildcard_all) ? ftn::wildcard_all : 0);
    gotnode = true;
  }

  if(gotnode and not gotpoint)
    point = (word)((node == (word)ftn::wildcard_all) ? ftn::wildcard_all : 0);

  if(gotzone and not gotdomain)
    domain = "";

  return *this;
}


//  ------------------------------------------------------------------

const gstring& gftnaddrs::build_address(bool normal) {

  char buf[20];
  address = "";

  if(net) {

    if(not normal and domain) {
      address = domain;
      address += '#';
    }

    if(zone) {
      if(zone == ftn::wildcard_all)
        address += '*';
      else if(zone == ftn::wildcard_first)
        address += '?';
      else
        address += ultoa(zone, buf, 10);
      address += ':';
    }

    if(net == ftn::wildcard_all)
      address += '*';
    else if(net == ftn::wildcard_first)
      address += '?';
    else
      address += ultoa(net, buf, 10);
    address += '/';

    if(node == ftn::wildcard_all)
      address += '*';
    else if(node == ftn::wildcard_first)
      address += '?';
    else
      address += ultoa(node, buf, 10);

    if(point) {
      address += '.';
      if(point == ftn::wildcard_all)
        address += '*';
      else if(point == ftn::wildcard_first)
        address += '?';
      else
        address += ultoa(point, buf, 10);
    }

    if(normal and domain) {
      address += '@';
      address += domain;
    }
  }

  return address;
}


//  ------------------------------------------------------------------

bool gftnaddrs::match(const gftnaddrs& mask) const {

  if(not ((mask.zone == ftn::wildcard_all) or (mask.zone == zone)))
    return false;
  if(not ((mask.net == ftn::wildcard_all) or (mask.net == net)))
    return false;
  if(not ((mask.node == ftn::wildcard_all) or (mask.node == node)))
    return false;
  if(not ((mask.point == ftn::wildcard_all) or (mask.point == point)))
    return false;
  if(not ((*mask.domain == '*') or (mask.domain.ieql(domain))))
    return false;

  return true;
}


//  ------------------------------------------------------------------

int gftnaddrs::compare(const gftnaddrs& other, bool fourdee) const {

  int n = 1;
  int d = compare_two(zone, other.zone);
  if(d == 0) {
    n++;
    d = compare_two(net, other.net);
    if(d == 0) {
      n++;
      d = compare_two(node, other.node);
      if(d == 0) {
        n++;
        d = compare_two(point, other.point);
        if((d == 0) and not fourdee) {
          n++;
          d = domain.icmp(other.domain);
        }
      }
    }
  }

  return d != 0 ? (d > 0 ? n : -n) : 0;
}


//  ------------------------------------------------------------------

int gftnaddrs::compare(const ftn_addr& other) const {

  int n = 1;
  int d = compare_two(zone, other.zone);
  if(d == 0) {
    n++;
    d = compare_two(net, other.net);
    if(d == 0) {
      n++;
      d = compare_two(node, other.node);
      if(d == 0) {
        n++;
        d = compare_two(point, other.point);
      }
    }
  }

  return d != 0 ? (d > 0 ? n : -n) : 0;
}


//  ------------------------------------------------------------------

gfile& operator<<(gfile& fp, const gftnaddrs& a) {

  fp << a.zone << a.net << a.node << a.point << a.domain;
  return fp;
}


//  ------------------------------------------------------------------

gfile& operator>>(gfile& fp, gftnaddrs& a) {

  fp >> a.zone >> a.net >> a.node >> a.point >> a.domain;
  return fp;
}


//  ------------------------------------------------------------------

gfile& operator<<(gfile& fp, const gftnnode& a) {

  fp << a.name;
  fp << a.addr;
  return fp;
}


//  ------------------------------------------------------------------

gfile& operator>>(gfile& fp, gftnnode& a) {

  fp >> a.name;
  fp >> a.addr;
  return fp;
}


//  ------------------------------------------------------------------

