
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gmd5.cpp,v 1.1.1.1 1998/11/21 19:38:06 odinn Exp $
//  ------------------------------------------------------------------
//  RSA Data Security, Inc.,MD5 message-digest algorithm
//  ------------------------------------------------------------------
//
//  Copyright (C) 1991-2, RSA Data Security, Inc. Created 1991. All
//  rights reserved.
//
//  License to copy and use this software is granted provided that it
//  is identified as the "RSA Data Security, Inc. MD5 Message-Digest
//  Algorithm" in all material mentioning or referencing this software
//  or this function.
//
//  License is also granted to make and use derivative works provided
//  that such works are identified as "derived from the RSA Data
//  Security, Inc. MD5 Message-Digest Algorithm" in all material
//  mentioning or referencing the derived work.
//
//  RSA Data Security, Inc. makes no representations concerning either
//  the merchantability of this software or the suitability of this
//  software for any particular purpose. It is provided "as is"
//  without express or implied warranty of any kind.
//
//  These notices must be retained in any copies of any part of this
//  documentation and/or software.
//
//  ------------------------------------------------------------------
//  This C++ implementation is ported from the original C version.
//  Ported by Odinn Sorensen.
//  ------------------------------------------------------------------


#include <string.h>
#include <gutlmisc.h>
#include <gmd5.h>


//  ------------------------------------------------------------------

char gmd5::padding[64] = {
  0x80, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0
};


//  ------------------------------------------------------------------
//  F, G, H and I are basic MD5 functions.

inline dword gmd5::calc_f(dword x, dword y, dword z) { return (x & y) | (~x & z); }
inline dword gmd5::calc_g(dword x, dword y, dword z) { return (x & z) | (y & ~z); }
inline dword gmd5::calc_h(dword x, dword y, dword z) { return x ^ y ^ z; }
inline dword gmd5::calc_i(dword x, dword y, dword z) { return y ^ (x | ~z); }


//  ------------------------------------------------------------------
//  ROTATE_LEFT rotates x left n bits.

inline dword gmd5::rotate_left(dword x, int n) { return (x << n) | (x >> (32-n)); }


//  ------------------------------------------------------------------
//  FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
//  Rotation is separate from addition to prevent recomputation.

//  ------------------------------------------------------------------

inline void gmd5::calc_ff(dword& a, dword b, dword c, dword d, dword x, int s, dword ac) {

  a += calc_f(b, c, d) + x + ac;
  a = rotate_left(a, s);
  a += b;
}


//  ------------------------------------------------------------------

inline void gmd5::calc_gg(dword& a, dword b, dword c, dword d, dword x, int s, dword ac) {

  a += calc_g(b, c, d) + x + ac;
  a = rotate_left(a, s);
  a += b;
}


//  ------------------------------------------------------------------

inline void gmd5::calc_hh(dword& a, dword b, dword c, dword d, dword x, int s, dword ac) {

  a += calc_h(b, c, d) + x + ac;
  a = rotate_left(a, s);
  a += b;
}


//  ------------------------------------------------------------------

inline void gmd5::calc_ii(dword& a, dword b, dword c, dword d, dword x, int s, dword ac) {

  a += calc_i(b, c, d) + x + ac;
  a = rotate_left(a, s);
  a += b;
}


//  ------------------------------------------------------------------
//  MD5 basic transformation. Transforms state based on block.

void gmd5::transform(const char* block) {

  const int S11 =  7;
  const int S12 = 12;
  const int S13 = 17;
  const int S14 = 22;
  const int S21 =  5;
  const int S22 =  9;
  const int S23 = 14;
  const int S24 = 20;
  const int S31 =  4;
  const int S32 = 11;
  const int S33 = 16;
  const int S34 = 23;
  const int S41 =  6;
  const int S42 = 10;
  const int S43 = 15;
  const int S44 = 21;

  dword x[16];
  dword a = state[0];
  dword b = state[1];
  dword c = state[2];
  dword d = state[3];

  decode(x, block, 64);

  // Round 1
  calc_ff(a, b, c, d, x[ 0], S11, 0xD76AA478UL); // 1
  calc_ff(d, a, b, c, x[ 1], S12, 0xE8C7B756UL); // 2
  calc_ff(c, d, a, b, x[ 2], S13, 0x242070DBUL); // 3
  calc_ff(b, c, d, a, x[ 3], S14, 0xC1BDCEEEUL); // 4
  calc_ff(a, b, c, d, x[ 4], S11, 0xF57C0FAFUL); // 5
  calc_ff(d, a, b, c, x[ 5], S12, 0x4787C62AUL); // 6
  calc_ff(c, d, a, b, x[ 6], S13, 0xA8304613UL); // 7
  calc_ff(b, c, d, a, x[ 7], S14, 0xFD469501UL); // 8
  calc_ff(a, b, c, d, x[ 8], S11, 0x698098D8UL); // 9
  calc_ff(d, a, b, c, x[ 9], S12, 0x8B44F7AFUL); // 10
  calc_ff(c, d, a, b, x[10], S13, 0xFFFF5BB1UL); // 11
  calc_ff(b, c, d, a, x[11], S14, 0x895CD7BEUL); // 12
  calc_ff(a, b, c, d, x[12], S11, 0x6B901122UL); // 13
  calc_ff(d, a, b, c, x[13], S12, 0xFD987193UL); // 14
  calc_ff(c, d, a, b, x[14], S13, 0xA679438EUL); // 15
  calc_ff(b, c, d, a, x[15], S14, 0x49B40821UL); // 16

  // Round 2
  calc_gg(a, b, c, d, x[ 1], S21, 0xF61E2562UL); // 17
  calc_gg(d, a, b, c, x[ 6], S22, 0xC040B340UL); // 18
  calc_gg(c, d, a, b, x[11], S23, 0x265E5A51UL); // 19
  calc_gg(b, c, d, a, x[ 0], S24, 0xE9B6C7AAUL); // 20
  calc_gg(a, b, c, d, x[ 5], S21, 0xD62F105DUL); // 21
  calc_gg(d, a, b, c, x[10], S22, 0x02441453UL); // 22
  calc_gg(c, d, a, b, x[15], S23, 0xD8A1E681UL); // 23
  calc_gg(b, c, d, a, x[ 4], S24, 0xE7D3FBC8UL); // 24
  calc_gg(a, b, c, d, x[ 9], S21, 0x21E1CDE6UL); // 25
  calc_gg(d, a, b, c, x[14], S22, 0xC33707D6UL); // 26
  calc_gg(c, d, a, b, x[ 3], S23, 0xF4D50D87UL); // 27
  calc_gg(b, c, d, a, x[ 8], S24, 0x455A14EDUL); // 28
  calc_gg(a, b, c, d, x[13], S21, 0xA9E3E905UL); // 29
  calc_gg(d, a, b, c, x[ 2], S22, 0xFCEFA3F8UL); // 30
  calc_gg(c, d, a, b, x[ 7], S23, 0x676F02D9UL); // 31
  calc_gg(b, c, d, a, x[12], S24, 0x8D2A4C8AUL); // 32

  // Round 3
  calc_hh(a, b, c, d, x[ 5], S31, 0xFFFA3942UL); // 33
  calc_hh(d, a, b, c, x[ 8], S32, 0x8771F681UL); // 34
  calc_hh(c, d, a, b, x[11], S33, 0x6D9D6122UL); // 35
  calc_hh(b, c, d, a, x[14], S34, 0xFDE5380CUL); // 36
  calc_hh(a, b, c, d, x[ 1], S31, 0xA4BEEA44UL); // 37
  calc_hh(d, a, b, c, x[ 4], S32, 0x4BDECFA9UL); // 38
  calc_hh(c, d, a, b, x[ 7], S33, 0xF6BB4B60UL); // 39
  calc_hh(b, c, d, a, x[10], S34, 0xBEBFBC70UL); // 40
  calc_hh(a, b, c, d, x[13], S31, 0x289B7EC6UL); // 41
  calc_hh(d, a, b, c, x[ 0], S32, 0xEAA127FAUL); // 42
  calc_hh(c, d, a, b, x[ 3], S33, 0xD4EF3085UL); // 43
  calc_hh(b, c, d, a, x[ 6], S34, 0x04881D05UL); // 44
  calc_hh(a, b, c, d, x[ 9], S31, 0xD9D4D039UL); // 45
  calc_hh(d, a, b, c, x[12], S32, 0xE6DB99E5UL); // 46
  calc_hh(c, d, a, b, x[15], S33, 0x1FA27CF8UL); // 47
  calc_hh(b, c, d, a, x[ 2], S34, 0xC4AC5665UL); // 48

  // Round 4
  calc_ii(a, b, c, d, x[ 0], S41, 0xF4292244UL); // 49
  calc_ii(d, a, b, c, x[ 7], S42, 0x432AFF97UL); // 50
  calc_ii(c, d, a, b, x[14], S43, 0xAB9423A7UL); // 51
  calc_ii(b, c, d, a, x[ 5], S44, 0xFC93A039UL); // 52
  calc_ii(a, b, c, d, x[12], S41, 0x655B59C3UL); // 53
  calc_ii(d, a, b, c, x[ 3], S42, 0x8F0CCC92UL); // 54
  calc_ii(c, d, a, b, x[10], S43, 0xFFEFF47DUL); // 55
  calc_ii(b, c, d, a, x[ 1], S44, 0x85845DD1UL); // 56
  calc_ii(a, b, c, d, x[ 8], S41, 0x6FA87E4FUL); // 57
  calc_ii(d, a, b, c, x[15], S42, 0xFE2CE6E0UL); // 58
  calc_ii(c, d, a, b, x[ 6], S43, 0xA3014314UL); // 59
  calc_ii(b, c, d, a, x[13], S44, 0x4E0811A1UL); // 60
  calc_ii(a, b, c, d, x[ 4], S41, 0xF7537E82UL); // 61
  calc_ii(d, a, b, c, x[11], S42, 0xBD3AF235UL); // 62
  calc_ii(c, d, a, b, x[ 2], S43, 0x2AD7D2BBUL); // 63
  calc_ii(b, c, d, a, x[ 9], S44, 0xEB86D391UL); // 64

  state[0] += a;
  state[1] += b;
  state[2] += c;
  state[3] += d;

  // Zeroize sensitive information.
  memset(x, 0, sizeof(x));
}


//  ------------------------------------------------------------------
//  Encodes input (dword) into output (char). Assumes len is a
//  multiple of 4.

void gmd5::encode(char* output, const dword* input, uint len) {

  uint i, j;
  for(i=0,j=0; j<len; i++,j+=4) {
    output[j] = (char)(input[i] & 0xFF);
    output[j+1] = (char)((input[i] >> 8) & 0xFF);
    output[j+2] = (char)((input[i] >> 16) & 0xFF);
    output[j+3] = (char)((input[i] >> 24) & 0xFF);
  }
}


//  ------------------------------------------------------------------
//  Decodes input (char) into output (dword). Assumes len is a
//  multiple of 4.

void gmd5::decode(dword* output, const char* input, uint len) {

  uint i, j;
  for(i=0,j=0; j<len; i++,j+=4)
    output[i] = ((dword)input[j]) | (((dword)input[j+1]) << 8) | (((dword)input[j+2]) << 16) | (((dword)input[j+3]) << 24);
}


//  ------------------------------------------------------------------

gmd5::gmd5() {

  memset(digest, 0, 16);
}


//  ------------------------------------------------------------------
//  MD5 initialization. Begins an MD5 operation, writing a new context.

void gmd5::begin() {

  count[0] = count[1] = 0;

  // Load magic initialization constants.
  state[0] = 0x67452301;
  state[1] = 0xEFCDAB89;
  state[2] = 0x98BADCFE;
  state[3] = 0x10325476;
}


//  ------------------------------------------------------------------
//  MD5 block update operation. Continues an MD5 message-digest
//  operation, processing another message block, and updating the
//  context.

void gmd5::update(ulong input) {

  char ulbuf[40];
  update(ultoa(input, ulbuf, 10));
}

void gmd5::update(const char* input) {

  update(input, strlen(input));
}

void gmd5::update(const char* input, uint inputlen) {

  // Compute number of chars mod 64
  uint index = (uint)((count[0] >> 3) & 0x3F);

  // Update number of bits
  if((count[0] += ((dword)inputlen << 3)) < ((dword)inputlen << 3))
    count[1]++;
  count[1] += ((dword)inputlen >> 29);

  uint partlen = 64 - index;

  // Transform as many times as possible.
  uint i = 0;
  if(inputlen >= partlen) {
    memcpy(&buffer[index], input, partlen);
    transform(buffer);
    for(i=partlen; (i+63)<inputlen; i+=64)
      transform(&input[i]);
    index = 0;
  }

  // Buffer remaining input
  memcpy(&buffer[index], &input[i], inputlen-i);
}


//  ------------------------------------------------------------------
//  MD5 finalization. Ends an MD5 message-digest operation, writing
//  the message digest and zeroizing the context.

void gmd5::end() {

  char bits[8];

  // Save number of bits
  encode(bits, count, 8);

  // Pad out to 56 mod 64.
  uint index = (uint)((count[0] >> 3) & 0x3F);
  uint padlen = (index < 56) ? (56 - index) : (120 - index);
  update(padding, padlen);

  // Append length (before padding)
  update(bits, 8);

  // Store state in digest
  encode(digest, state, 16);

  // Zeroize sensitive information.
  memset(state, 0, sizeof(state));
  memset(count, 0, sizeof(count));
  memset(buffer, 0, sizeof(buffer));
}


//  ------------------------------------------------------------------

int gmd5::operator==(gmd5& m) {

  return !memcmp(digest, m.digest, 16);
}


//  ------------------------------------------------------------------

