
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gmemblks.cpp,v 1.1.1.1 1998/11/21 19:38:06 odinn Exp $
//  ------------------------------------------------------------------
//  Memory blocks.
//  ------------------------------------------------------------------


#ifdef __BORLANDC__
//#pragma option -Od
#endif

#include <gmemall.h>
#include <gfile.h>
#include <gmemdbg.h>
#include <gmemblks.h>


//  ------------------------------------------------------------------

gmemblock::gmemblock(int block_size, int line_size, int line_extra) {

  blocksize = block_size;
  lineextra = line_extra;
  linesize = line_size;
  linelen = sizeof(lineindex) + lineextra;

  reset(true);
}


//  ------------------------------------------------------------------

gmemblock::~gmemblock() {

  reset();
}


//  ------------------------------------------------------------------

void gmemblock::reset(bool init) {

  if(not init) {
    free(blockidx);
    free(block);
    if(views) {
      for(int i=0; i<views; i++)
        delete view[i];
      free(view);
    }
  }
  view = NULL;
  views = 0;
  block = NULL;
  blockidx = NULL;
  blocklen = 0;
  blocklines = 0;
  blockreserved = 0;
  alloclen = 0;
  alloclines = 0;
}


//  ------------------------------------------------------------------

void gmemblock::adjust_block(uint len) {

  if(len > alloclen) {
    alloclen = blocksize * ((blocksize-1+len) / blocksize);
    block = (char*)realloc(block, alloclen);  throw_alloc(block);
  }
}


//  ------------------------------------------------------------------

gmemblock::lineindex* gmemblock::adjust_index(int lines) {

  if(lines > alloclines) {
    int len = lines * linelen;
    int size = linesize * (((linesize-1+len) / linesize));
    blockidx = (char*)realloc(blockidx, size);  throw_alloc(blockidx);
    alloclines = size / linelen;
  }

  return (lineindex*)blockidx;
}


//  ------------------------------------------------------------------

void gmemblock::adjust_offsets(int index, int difference) {

  for(int n=index; n<blocklines; n++)
    blockindex(n).offset += difference;
}


//  ------------------------------------------------------------------

int gmemblock::add(const char* s, uint len) {

  uint pos = blocklen;

  adjust_block(blocklen+len);
  memcpy(block+pos, s, len);

  adjust_index(blocklines+1);

  blockindex(blocklines).offset = pos;
  blockindex(blocklines).length = len;

  blocklen += len;

  return blocklines++;
}


//  ------------------------------------------------------------------

int gmemblock::insert(int n, const char* s, uint newlen) {

  if(n >= blocklines)
    return add(s, newlen);

  adjust_block(blocklen+newlen);
  adjust_index(blocklines+1);

  uint oldpos = blockindex(n).offset;

  memmove(block+oldpos+newlen, block+oldpos, blocklen-oldpos);
  memcpy(block+oldpos, s, newlen);

  memmove(blockidxptr(n+1), blockidxptr(n), (blocklines-n)*linelen);

  blockindex(n).offset = oldpos;
  blockindex(n).length = newlen;

  blocklen += newlen;
  blocklines++;

  adjust_offsets(n+1, newlen);

  for(int i=0; i<views; i++) {
    if(n < view[i]->first)
      view[i]->first++;
    if(n < view[i]->last)
      view[i]->last++;
  }

  return n;
}


//  ------------------------------------------------------------------

int gmemblock::set(int n, const char* s, uint newlen) {

  uint oldlen = blockindex(n).length;

  int difference = newlen - oldlen;
  if(difference) {
    adjust_block(blocklen+difference);
    if((n+1) < blocklines) {
      uint oldpos = blockindex(n+1).offset;
      uint newpos = oldpos + difference;
      memmove(block+newpos, block+oldpos, blocklen-oldpos);
    }
    blocklen += difference;
    blockindex(n).length = newlen;
  }

  memcpy(block+blockindex(n).offset, s, newlen);

  adjust_offsets(n+1, difference);

  return n;
}


//  ------------------------------------------------------------------

char* gmemblock::get(int n, char* s) const {

  return get(n, s, blockindex(n).length);
}


//  ------------------------------------------------------------------

char* gmemblock::get(int n, char* s, uint l) const {

  return (char*)memcpy(s, get(n), l);
}


//  ------------------------------------------------------------------

void gmemblock::erase(int first, int last) {

  uint len = 0;
  for(int n=first; n<last; n++) {
    uint tlen = blockindex(n).length;
    len += tlen;
  }

  uint firstpos = blockindex(first).offset;

  memmove(block+firstpos, block+firstpos+len, blocklen-firstpos);
  memmove(blockidxptr(first), blockidxptr(last), (blocklines-last)*linelen);

  blocklines -= last-first;
  blocklen -= len;

  adjust_offsets(first, -len);

  for(int i=0; i<views; i++) {
    uint& vfirst = view[i]->first;
    uint& vlast = view[i]->last;
    if(first <= vfirst) {
      if(last <= vfirst) {
        vfirst -= last - first;
        vlast -= last - first;
      }
      else if(last <= vlast) {
        vfirst = first;
        vlast = first + vlast - last;
      }
      else {
        vfirst = vlast = first;
      }
    }
    else {
      if(last <= vlast)
        vlast -= last - first;
      else
        vlast = first;
    }
  }
}


//  ------------------------------------------------------------------

gmemblock& gmemblock::operator=(const gmemblock& o) {

  reset();

  blocklen = o.blocklen;
  blocklines = o.blocklines;
  alloclen = o.alloclen;
  alloclines = o.alloclines;
  block = (char*)malloc(alloclen);
  throw_new(block);
  memcpy(block, o.block, alloclen);
  blockidx = (char*)malloc(alloclines*linelen);
  throw_new(blockidx);
  memcpy(blockidx, o.blockidx, alloclines*linelen);
  views = o.views;
  view = (gmemblockview**)calloc(views, sizeof(gmemblockview*));
  throw_new(view);
  for(int i=0; i<views; i++) {
    view[i] = new gmemblockview(o.view[i]);
    throw_new(view[i]);
  }

  return *this;
}


//  ------------------------------------------------------------------

void gmemblock::save(gfile& fp) {

  fp << blocklen << blocklines << views;

  fp.fwrite(block, blocklen);
  fp.fwrite(blockidx, blocklines*linelen);

  for(int i=0; i<views; i++)
    fp.fwrite(view[i], sizeof(gmemblockview));
}


//  ------------------------------------------------------------------

void gmemblock::load(gfile& fp) {

  reset();

  fp >> blocklen >> blocklines >> views;

  adjust_block(blocklen);
  adjust_index(blocklines);

  fp.fread(block, blocklen);
  fp.fread(blockidx, blocklines*linelen);

  view = (gmemblockview**)calloc(views, sizeof(gmemblockview*));
  throw_new(view);
  for(int i=0; i<views; i++) {
    view[i] = new gmemblockview;
    throw_new(view[i]);
    fp.fread(view[i], sizeof(gmemblockview));
  }
}


//  ------------------------------------------------------------------

gmemblockview* gmemblock::add_view(uint first, uint last) {

  gmemblockview* newview = new gmemblockview(first, last);
  throw_new(newview);
  view = (gmemblockview**)realloc(view, (views+1)*sizeof(gmemblockview*));
  throw_new(view);
  view[views++] = newview;
  return newview;
}


//  ------------------------------------------------------------------

void gmemblock::remove_view(gmemblockview* v) {

  for(int i=0; i<views; i++) {
    if(view[i] == v) {
      delete view[i];
      for(--views; i<views; i++)
        view[i] = view[i+1];
      break;
    }
  }
}


//  ------------------------------------------------------------------

