
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gmemory.h,v 1.1.1.1 1998/11/21 19:38:08 odinn Exp $
//  ------------------------------------------------------------------
//  Memory allocation stuff a'la STL.
//  ------------------------------------------------------------------


#ifndef __GMEMORY_H
#define __GMEMORY_H


//  ------------------------------------------------------------------

#include <limits.h>
#include <gmemdbg.h>


//  ------------------------------------------------------------------

#ifndef __WATCOMC__
inline void* operator new (size_t, void* p) {

  return p;
}
#endif


//  ------------------------------------------------------------------

template <class T>
inline T* gallocate(int size, T*) {

  T* tmp = (T*)(::operator new((unsigned int)(size * sizeof(T))));
  if(tmp == 0) {
    //cerr << "out of memory" << endl;
    throw_new(tmp);
  }
  return tmp;
}


//  ------------------------------------------------------------------

template <class T>
inline T* gallocate(long size, T*) {

  T* tmp = (T*)(::operator new((unsigned long)(size * sizeof(T))));
  if(tmp == 0) {
    //cerr << "out of memory" << endl;
    throw_new(tmp);
  }
  return tmp;
}


//  ------------------------------------------------------------------

template <class T>
inline void gdeallocate(T* buffer) {

  ::operator delete(buffer);
}


//  ------------------------------------------------------------------

template <class T>
class gallocator {

public:

  typedef T         value_type;
  typedef T*        pointer;
  typedef const T*  const_pointer;
  typedef T&        reference;
  typedef const T&  const_reference;
  typedef size_t    size_type;
  typedef ptrdiff_t difference_type;

  pointer allocate(size_type n = 0)   { return ::gallocate((difference_type)n, (pointer)0); }
  void    deallocate(pointer p)       { ::gdeallocate(p); }

  pointer       address(reference x)              { return (pointer)&x; }
  const_pointer const_address(const_reference x)   { return (const_pointer)&x; }

  size_type init_page_size()  { return maximum_of_two(size_type(1), size_type(4096/sizeof(T))); }

  size_type max_size() const  { return maximum_of_two(size_type(1), size_type(UINT_MAX/sizeof(T))); }

};


//  ------------------------------------------------------------------
//  A specialization.

class gallocator<void> {

public:

  typedef void* pointer;
};


//  ------------------------------------------------------------------
//  Construct object in-place

template <class T1, class T2>
inline void construct(T1* p, const T2& value) {

  new (p) T1(value);
}


//  ------------------------------------------------------------------
//  Destroy object that was constructed in-place

template <class T>
inline void destroy(T* pointer) {

  #if (defined(__BORLANDC__) and (__BORLANDC__ < 0x500))
  pointer->T::~T();
  #else
  pointer->~T();
  #endif
}


//  ------------------------------------------------------------------
//  Destroy objects that were constructed in-place

template <class Pointer>
void destroy(Pointer first, Pointer last) {

  while(first != last) {
    destroy(first);
    ++first;
  }
}


//  ------------------------------------------------------------------
//  Some specializations of destroy for builtin types.

inline void destroy(void*)             {;}
inline void destroy(char*)             {;}
inline void destroy(unsigned char*)    {;}
inline void destroy(short*)            {;}
inline void destroy(unsigned short*)   {;}
inline void destroy(int*)              {;}
inline void destroy(unsigned int*)     {;}
inline void destroy(long*)             {;}
inline void destroy(unsigned long*)    {;}
inline void destroy(float*)            {;}
inline void destroy(double*)           {;}
inline void destroy(long double*)      {;}
inline void destroy(void**)            {;}
inline void destroy(char**)            {;}
inline void destroy(unsigned char**)   {;}
inline void destroy(short**)           {;}
inline void destroy(unsigned short**)  {;}
inline void destroy(int**)             {;}
inline void destroy(unsigned int**)    {;}
inline void destroy(long**)            {;}
inline void destroy(unsigned long**)   {;}
inline void destroy(float**)           {;}
inline void destroy(double**)          {;}
inline void destroy(long double**)     {;}
inline void destroy(void***)           {;}
inline void destroy(char***)           {;}
inline void destroy(unsigned char***)  {;}
inline void destroy(short***)          {;}
inline void destroy(unsigned short***) {;}
inline void destroy(int***)            {;}
inline void destroy(unsigned int***)   {;}
inline void destroy(long***)           {;}
inline void destroy(unsigned long***)  {;}
inline void destroy(float***)          {;}
inline void destroy(double***)         {;}
inline void destroy(long double***)    {;}
#ifndef GSTD_NO_BOOL
inline void destroy(bool*)             {;}
inline void destroy(bool**)            {;}
inline void destroy(bool***)           {;}
#endif


//  ------------------------------------------------------------------

template <class InputIterator, class ForwardIterator>
ForwardIterator uninitialized_copy(InputIterator first, InputIterator last, ForwardIterator result) {

  while(first != last)
    construct(&*result++, *first++);

  return result;
}


//  ------------------------------------------------------------------

template <class ForwardIterator, class T>
void uninitialized_fill(ForwardIterator first, ForwardIterator last, const T& x) {

  while(first != last)
    construct(&*first++, x);
}


//  ------------------------------------------------------------------

template <class ForwardIterator, class Size, class T>
ForwardIterator uninitialized_fill_n(ForwardIterator first, Size n, const T& x) {

  while(n--)
    construct(&*first++, x);

  return first;
}


//  ------------------------------------------------------------------

template <class iIter, class oIter>
oIter copy(iIter first, iIter last, oIter result) {

  while(first != last)
    *result++ = *first++;

  return result;
}


//  ------------------------------------------------------------------

template <class bIter1, class bIter2>
bIter2 copy_backward(bIter1 first, bIter1 last, bIter2 result) {

  while(first != last)
    *(--result) = *(--last);
  return result;
}


//  ------------------------------------------------------------------

template <class fIter, class T>
void fill(fIter first, fIter last, const T& value) {

  while(first != last)
    *first++ = value;
}


//  ------------------------------------------------------------------

#endif

//  ------------------------------------------------------------------

