
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gsndvoc.h,v 1.1.1.1 1998/11/21 19:38:09 odinn Exp $
//  ------------------------------------------------------------------
//  CT-VOICE.DRV Sound Blaster Interface.
//  ------------------------------------------------------------------
//  This C++ class provides a clean object-oriented interface to the
//  Sound Blaster CT-VOICE.DRV driver.
//
//  This implementation is based on ASM/C code examples provided by
//  Creative Technology and various other sources. The example VOC
//  player is adapted (and greatly expanded) from a C code example
//  also by Creative Technology.
//  
//  Converted to C++ object and inline assembler by Odinn Sorensen.
//  ------------------------------------------------------------------
//  Sound Blaster and Sound Blaster Pro are a TradeMarks of Creative
//  Labs, Inc. and Creative Technology Pte Ltd.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------
//  Only include once!

#ifndef __GSNDVOC_H
#define __GSNDVOC_H


//  ------------------------------------------------------------------
//  Required headers

#include <gdefs.h>


//  ------------------------------------------------------------------
//  VOC file format header.
//  ------------------------------------------------------------------
//
//  This header is based on information in the Sound Blaster Developer
//  Kit from Creative Technology, in the form of the archives
//  SBK-1.ARJ and MANUAL.ARJ, which appear to be an independent
//  transcription of the printed manual + the kit files.
//
//  The kit I looked at did not include information about the new
//  features of Sound Blaster Pro, such as stereo in VOC's, higher
//  sampling rates etc.
//
//  Any stuff in here about Sound Blaster Pro comes from my own
//  experiments, just so you know ...
//
//  ------------------------------------------------------------------
//
//  Sound Blaster and Sound Blaster Pro are a TradeMarks of Creative
//  Labs, Inc. and Creative Technology Pte Ltd.
//
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------
//  CREATIVE VOICE FILE (VOC) FORMAT
//
//  The Creative Voice File is organized in two main blocks, the
//  Header Block and Data Block.
//
//  The Header Block contains identifier, version number and pointer
//  to the start of the Data Block.
//
//  The Data Block is divided into various sub-blocks of various types
//  such as Voice Data, Silence, Marker, ASCII text, Repeat and End
//  Repeat, and Terminator.
//
//  The CT-VOICE driver only processes the Data Block. It is IMPORTANT
//  that you pass the adress of the Data Block and not the Voice File
//  when calling the driver to perform voice output.
//
//  ------------------------------------------------------------------

struct vocHdr {         // Offs  Size
                        // -------------------------------------------
  char  filetype[20];   //    0    20  File Type Description
                        //
                        //             The following message is stored
                        //             here: "Creative Voice File",1AH
                        //
  word  dataoffset;     //   20     2  Offset of Data Block from the
                        //             start of voice file
                        //
                        //             This WORD points to the Data
                        //             Block. It helps the application
                        //             programs to locate the Data
                        //             Block in case the size of
                        //             Header Block is changed.
                        //
                        //             For this version, the value
                        //             here is always 1AH
                        //
  word  vocversion;     //   22     2  Voice file format version
                        //             number
                        //
                        //             This version number allows your
                        //             program to identify different
                        //             organization formats of voice
                        //             data file in case of future
                        //             enchancement.
                        //
                        //             LSB and MSB are the minor and
                        //             major version number
                        //             respectively. Current version
                        //             is 1.10
                        //
  word  vocid;          //   24     2  Voice file identification code
                        //
                        //             This code allows your program
                        //             to check that this file is a
                        //             VOC file. It's content is the
                        //             complement of the file format
                        //             version number then plus 1234h
                        //
                        //             For current version of 1.10 it
                        //             is complement (010Ah) + 1234H =
                        //             1129H
                        // -------------------------------------------
};


//  ------------------------------------------------------------------
//  DATA BLOCK
//
//  The Data Block is sub-divided into multiple sub-blocks of data.
//
//  The first byte of each sub-block is called the Block Type
//  (BLKTYPE). It indicates the type of data contained in the
//  sub-block.
//
//  The next three bytes is the 24-bit Block Length (BLKLEN). It is
//  the number of bytes in the sub-block _excluding_ the BLKTYPE and
//  BLKLEN fields. Only the Terminator sub-block does not have the
//  BLKLEN field.
//
//  Depending on BLKTYPE, the rest of the data in the sub-block may be
//  the voice attributes (such as sampling rate and packing), voice
//  data or other information (such as Marker and ASCII text).
//
//  Your program needs not interpret all the BLKTYPE. For those
//  BLKTYPEs it does not recognize, it can move to the next sub-block
//  based on the BLKLEN easily.
//
//  In most of the applications, your program can simply pass the
//  address of the Data Block to the CT-VOICE driver which will
//  interpret the Voice file and perform the output accordingly.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------
//  Definition of VOC blktypes

#define VOC_BLK_TERMINATOR   0x00
#define VOC_BLK_VOICEDATA    0x01
#define VOC_BLK_VOICECONT    0x02
#define VOC_BLK_SILENCE      0x03
#define VOC_BLK_MARKER       0x04
#define VOC_BLK_ASCIITEXT    0x05
#define VOC_BLK_REPEATLOOP   0x06
#define VOC_BLK_ENDLOOP      0x07
#define VOC_BLK_STEREODATA   0x08


//  ------------------------------------------------------------------
//  Block Type 0 - Terminator
//
//  00
//
//  This 1-byte sub-block terminates the entire Data Block. It
//  indicates that there is no other sub-block after it. CT-VOICE
//  driver terminates the voice outpt when this BLKTYPE is
//  encountered.
//
//  This BLKTYPE must be the last block of the voice file.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------
//  Block Type 1 - Voice Data
//
//  01, BLKLEN(24-bits), SR, PACK, VOICE DATA
//
//  This BLKTYPE indicates that this sub-block is a new set of voice
//  data. It may use different sampling rate or packing method.
//
//  The SR field (1 byte) is the Sampling Rate. It is defined as:
//
//       SR = 256 - 1000000 / sampling rate
//
//  For instance, if the sampling rate is 10,000 Hz, SR is set to 156.
//
//  The PACK field (1 byte) is the Packing method used for this
//  sub-block of voice data. It is defined as:
//
//        0      : 8-bit raw or unpacked
//        1      : 4-bit packed
//        2      : 2.6-bit packed
//        3      : 2-bit packed
//
//  The voice data length is calculated as follow:
//
//  voice data length = BLKLEN - 2
//  ------------------------------------------------------------------

struct vocBlk1 {
  byte  type;
  word  len_lo;
  byte  len_hi;
  byte  samplerate;
  byte  packtype;
};


//  ------------------------------------------------------------------
//  Block Type 2 - Voice Continuation
//
//  02, BLKLEN(24-bits), VOICE DATA
//
//  This sub-block is a continuation of the voice data from the last
//  Voice Data sub-block.
//
//  The length of voice data is BLKLEN.
//  ------------------------------------------------------------------

struct vocBlk2 {
  byte  type;
  word  len_lo;
  byte  len_hi;
};


//  ------------------------------------------------------------------
//  Block Type 3 - Silence
//
//  03, BLKLEN(24-bits), PERIOD(16-bits), SR
//
//  This sub-block defines silence period in the voice data.
//
//  The length of the silence period (PERIOD, 2 bytes) is stored after
//  the BLKLEN field in units of sampling cycle. Its value is one less
//  than the silence period.
//
//  The sampling rate (SR) is defined as above.
//
//  The BLKLEN for this sub-block is always 3.
//  ------------------------------------------------------------------

struct vocBlk3 {
  byte  type;
  word  len_lo;
  byte  len_hi;
  word  period;
  byte  samplerate;
};


//  ------------------------------------------------------------------
//  Block Type 4 - Marker
//
//  04, BLKLEN(24-bits), MARKER(16-bits)
//
//  This is a special sub-block that specifies a Marker (MARKER, 2
//  bytes) in the voice data.
//
//  During voice output, the CT-VOICE driver updates the status word
//  ct_voice_status with this MARKER value. Your program may check for
//  the desired MARKER value to perform synchronization with the voice
//  output process.
//
//  The MARKER value can be added to the Voice File using the Sound
//  Blaster utility program VEDIT.
//
//  The MARKER value must be within the range of 1 to 0FFFE hex. Note
//  that 0 and FFFF hex are reserved by the CT-VOICE driver. Also
//  refer to the following section Essentials of using CT-VOICE
//  driver.
//
//  The BLKLEN field is always set to 2.
//  ------------------------------------------------------------------

struct vocBlk4 {
  byte  type;
  word  len_lo;
  byte  len_hi;
  word  marker;
};


//  ------------------------------------------------------------------
//  Block Type 5 - ASCII Text
//
//  05, BLKLEN(24-bits), ASCII DATA, NULL
//
//  This sub-block specifies a null-terminated ASCII string in the
//  voice data file.
//
//  Most programs may choose to ignore this BLKTYPE. It is defined for
//  a program wich requires ASCII information on the voice such as
//  name, type or remarks on the voice data.
//  ------------------------------------------------------------------

struct vocBlk5 {
  byte  type;
  word  len_lo;
  byte  len_hi;
  char  string[1];
};


//  ------------------------------------------------------------------
//  Block Type 6 - Repeat Loop
//
//  06, BLKLEN(24-bits), COUNT(16-bits)
//
//  This sub-block specifies the beginning of a repeat loop. The data
//  sub-blocks between this block and the next End Repeat Loop
//  sub-block is repeated by the CT-VOICE driver for (COUNT + 1)
//  times.
//
//  The COUNT (2 bytes) field takes a value from 1 TO FFFE hex as loop
//  count. If COUNT is FFFF hex, then it is to be an endless loop. In
//  this case, refer to the following section Reference for SBC Voice
//  Functions on how to break out of the endless loop.
//
//  The BLKLEN field is always 2.
//  ------------------------------------------------------------------

struct vocBlk6 {
  byte  type;
  word  len_lo;
  byte  len_hi;
  word  count;
};


//  ------------------------------------------------------------------
//  Block Type 7 - End Repeat Loop
//
//  07, BLKLEN(24-bits)
//
//  This sub-blocks specifies the end of a Repeat Loop, Refer to Block
//  Type 6 for details.
//
//  The BLKLEN is always 0.
//  ------------------------------------------------------------------

struct vocBlk7 {
  byte  type;
  word  len_lo;
  byte  len_hi;
};


struct vocBlk8 {
  byte  type;
  word  len_lo;
  byte  len_hi;
  byte  byte0;
  byte  samplerate;
  byte  byte2;
  byte  stereo;
};


//  ------------------------------------------------------------------
//  Block Type X - Unknown block
//
//  XX
//
//  This block type is unknown to the current implementation, so the
//  only thing we can assume about it is the length, with which we can
//  skip over and ignore it safely...
//  ------------------------------------------------------------------

struct vocBlkX {
  byte  type;
  word  len_lo;
  byte  len_hi;
  byte  unknown[1];
};


//  ------------------------------------------------------------------
//  Union of all known VOC block types + an empty for skipping unknown

union vocBlk {
  byte    type;
  vocBlk1 type1;
  vocBlk2 type2;
  vocBlk3 type3;
  vocBlk4 type4;
  vocBlk5 type5;
  vocBlk6 type6;
  vocBlk7 type7;
  vocBlk8 type8;
  vocBlkX typex;
};


//  ------------------------------------------------------------------
//  Version of this class

#define CTV_CLASS_VERSION 0x0101


//  ------------------------------------------------------------------
//  Class status values

#define CTV_STATUS_NOERRORS 0x0000
#define CTV_STATUS_UNDEFERR 0xFFFF


//  ------------------------------------------------------------------
//  Driver function numbers

#define CTV_FUNC_GET_VERSION_NUM   0
#define CTV_FUNC_SET_BASE_IO_ADDX  1
#define CTV_FUNC_SET_INTERRUPT_NUM 2
#define CTV_FUNC_INITIALIZE_DRIVER 3
#define CTV_FUNC_ON_OFF_SPEAKER    4
#define CTV_FUNC_SET_STATUS_ADDX   5
#define CTV_FUNC_OUTPUT_VOICE      6
#define CTV_FUNC_INPUT_VOICE       7
#define CTV_FUNC_TERMINATE_PROCESS 8
#define CTV_FUNC_UNINSTALL_DRIVER  9
#define CTV_FUNC_PAUSE_VOICE       10
#define CTV_FUNC_CONTINUE_VOICE    11
#define CTV_FUNC_BREAK_LOOP        12
#define CTV_FUNC_SET_USER_FUNCTION 13


//  ------------------------------------------------------------------
//  Default SoundBlaster values

#define SB_DEF_IOPORT 220
#define SB_DEF_INTNUM 7


//  ------------------------------------------------------------------
//  Base class

class ctvClass {

  //  ----------------------------------------------------------------
  private:

    static word classInstances;
    static word classVersion;
    static word classStatus;

    static char* voiceDriver;            // "There can be only one..."

    static word ioPort;
    static word intNum;

    // Volatile because the driver can change the this variable while
    // the program is running. It tells the compiler not to assume
    // anything about the state of the variable, thereby disabling
    // certain types of optimizations where this variable is involved.

    static volatile word voiceStatus;

  //  ----------------------------------------------------------------
  protected:

    char* voiceBuffer;
    long voiceBufferSize;

    word setStatusWord(word* statptr);

  //  ----------------------------------------------------------------
  public:

    int debug;

    ctvClass(int dbg=false, int ioport=0, int intnum=0, char* filename=NULL);
    ~ctvClass();

    word getClassVersion();
    word getClassVersion(word& major, word& minor);

    word getDriverVersion();
    word getDriverVersion(word& major, word& minor);

    word getVoiceStatus() { return voiceStatus; }
    word getClassStatus() { return classStatus; }

    vocHdr* getVoiceHeader() { return (vocHdr*)voiceBuffer; }

    word loadDriver(char* filename=NULL);
    void removeDriver();
    word initDriver();

    word setIoPort(word ioport=SB_DEF_IOPORT);
    word setIntNum(word intnum=SB_DEF_INTNUM);

    word setSpeaker(word setting=true);

    word loadVoiceBuffer(char* filename, void* vocbuf=NULL, long vocbufsize=0L);
    word saveVoiceBuffer(char* filename, void* vocbuf=NULL, long vocbufsize=0L);
    void removeVoiceBuffer(void* vocbuf=NULL);

    word playVoiceBuffer(void* vocbuf=NULL);
    word recordVoiceBuffer(void* vocbuf=NULL, long vocbufsize=0L, word samplerate=22050);

    word stopVoice();
    word pauseVoice();
    word continueVoice();
    word breakLoop(word setting);

    byte cvtWrd2Sr(word wrd) { return (byte)(256L-(1000000L/(long)wrd)); }
    word cvtSr2Wrd(byte sr) { return (word)(1000000L/(256L-(long)sr)); }
    char* cvtPack2Str(byte packtype);

  //  ----------------------------------------------------------------
};


//  ------------------------------------------------------------------

#endif

//  ------------------------------------------------------------------

