
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gsock.h,v 1.2 1998/12/27 16:08:12 odinn Exp $
//  ------------------------------------------------------------------
//  TCP/IP classes.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------

#ifndef __G_TCPIP_H
#define __G_TCPIP_H

#include <gsocket.h>


//  ------------------------------------------------------------------

class ginternetclient {

protected:

  char* line_pointer;
  int multi_line_response;
  int response_code;

public:

  gsocket& socket;

  ginternetclient(gsocket& s);
  virtual ~ginternetclient();

  int error() { return socket.error(); }
  char* line_buffer() { return socket.linebuffer(); }

  void debug(int setting=true) { socket.debug = setting; }
  void blocking(int setting=true) { socket.blocking = setting; }

  // Open/close server connection.
  int open(const char* server, const char* service_name) { return socket.open(server, service_name); }
  int open(const char* server, int service_port) { return socket.open(server, service_port); }
  void close() { socket.close(); }

  // Return true if a connection has been made.
  int connected() { return socket.connected(); }

  // Get a text response from the server.
  // Return true if the server is sending a multi-line response.
  // Return false if there are no further response lines.
  // If false is returned, error() should be checked.
  virtual int get_response();

  // Get a text line from the server.
  // Return true if the server is expected to send more lines.
  // Return false if there is no more data.
  // If false is returned, error() should be checked.
  int get_line();

  // Return pointer into the line buffer.
  // Return NULL if there is no more data.
  const char* line() { return line_pointer; }

  // Send line to the server.
  // Return true if line was sent successfully, false if there was an error.
  int put_line(const char* line);

  // Translate line_buffer and return line_pointer into it.
  // Called by get_line().
  // Return pointer to the line_buffer.
  virtual char* translate_get_buffer();

  // Translate/copy line into line_buffer.
  // Called by put_line().
  virtual void translate_put_buffer(const char* line);
};


//  ------------------------------------------------------------------
//  NNTP class

class gnntp {

protected:

  enum {
    mode_none,
    mode_list,
    mode_list_active_times,
    mode_list_distributions,
    mode_list_distrib_pats,
    mode_list_newsgroups,
    mode_list_overview_fmt,
    mode_post,
    mode_xover
  };

  int mode;

  int rspcode;
  char rspbuffer[512];
  char cmdbuffer[512];
  char linebuffer[512];

  const char* lineptr;

  int global_posting_allowed;

  ulong number_of_articles;
  ulong first_article_number;
  ulong last_article_number;

  ulong current_article_number;
  char current_article_messageid[512];

  char* cracked_group;
  ulong cracked_last_article;
  ulong cracked_first_article;
  int cracked_posting_allowed;

  ulong cracked_timecreated;
  char* cracked_creator_email;

  char* cracked_distribution_value;
  char* cracked_distribution_explanation;

  ulong cracked_weight;
  char* cracked_group_pattern;

  char* cracked_group_description;

  char* cracked_header;
  int cracked_header_full;

  ulong cracked_article_number;
  char* cracked_subject;
  char* cracked_author;
  char* cracked_date;
  char* cracked_message_id;
  char* cracked_references;
  ulong cracked_byte_count;
  ulong cracked_line_count;
  char* cracked_xover_other;

  void append_cmd_argument(const char* argument);
  void build_command(const char* command, const char* argument);

  int send_command(int ok_code);
  int send_command(const char* command=NULL);
  int send_command(const char* command, int ok_code);
  int send_command(const char* command, int ok_code1, int ok_code2);

  int get_response();

  void get_rsp_num_msgid();

  int specify_article(const char* command);
  int specify_article(const char* command, ulong number);
  int specify_article(const char* command, const char* messageid);

  int last_or_next(const char* command);

  void crack_list();
  void crack_list_active_times();
  void crack_list_distributions();
  void crack_list_distrib_pats();
  void crack_list_newsgroups();
  void crack_list_overview_fmt();
  void crack_xover();

public:
  
  gsocket& socket;

  gnntp(gsocket& s);
  ~gnntp();

  void debug(int setting=true) { socket.debug = setting; }
  void blocking(int setting=true) { socket.blocking = setting; }

  // Open and close connection to the NNTP server
  int open(const char* server);
  void close();

  // Get a line in a multiline response from the server
  // Return NULL when there are no more lines
  const char* getline();
  const char* line_pointer() { return lineptr; }

  // Send line to the server. Sends ".CRLF" if line is NULL
  int putline(const char* line);
  
  // Return true if posting is allowed on the server
  int posting_allowed() { return global_posting_allowed; }

  // Return last command response
  int response_code() { return rspcode; }
  const char* response() { return rspbuffer; }
  const char* response_text() { return *rspbuffer ? rspbuffer+4 : rspbuffer; }

  // Return the number or messageid of the current article
  ulong article_number() { return current_article_number; }
  const char* article_messageid() { return current_article_messageid; }

  // Article selection commands
  int article(const char* messageid);
  int article(ulong number);
  int head(const char* messageid);
  int head(ulong number);
  int body(const char* messageid);
  int body(ulong number);
  int stat(const char* messageid);
  int stat(ulong number);

  // Newsgroup selection command
  int group(const char* newsgroup_name);

  // Ask for a summary of available commands
  int help();

  // Go to the previous or next article
  int last();
  int next();

  // Ask for a list of valid newsgroups
  int list(const char* pattern=NULL);

  // Ask for a list of article numbers in a newsgroup
  int listgroup(const char* newsgroup=NULL);

  // Ask for a list of new groups since a certain date/time
  int newgroups(const char* date, const char* time, const char* gmt, const char* distributions=NULL);

  // Ask for a list of articles since a certain date/time
  int newnews(const char* newsgroups, const char* date, const char* time, const char* gmt, const char* distribution=NULL);

  // Post an article
  int post();

  // Tell the server that to optimize for a news reading client
  int mode_reader();

  // Ask for the overview (xover) format
  int list_overview_fmt();

  // Get newsgroup descriptions
  int list_newsgroups(const char* pattern=NULL);
  int xgtitle(const char* pattern=NULL);

  // Get specific headers from specific articles
  int xhdr(const char* header, const char* range=NULL);

  // Get overview database information for the specified articles
  int xover(const char* range=NULL);

  int list_active_times();
  int list_subscriptions();
  int list_distrib_pats();
  int list_distributions();
};


//  ------------------------------------------------------------------
//  POP3 class

class gpop3 {

protected:

  int list_mode;

  int rspstatus;
  char rspbuffer[512];
  char cmdbuffer[512];
  char linebuffer[512];

  const char* lineptr;

  ulong number_of_messages;
  ulong maildrop_size;
  ulong msg_number;
  ulong msg_size;

  int get_response();

  int send_command(const char* command=NULL);

  int user(const char* userid);
  int pass(const char* password);

  void crack_list();

public:

  gsocket& socket;

  gpop3(gsocket& s);
  ~gpop3();

  void debug(int setting=true) { socket.debug = setting; }
  void blocking(int setting=true) { socket.blocking = setting; }

  // Open and close connection to the POP3 server
  int open(const char* server, const char* userid, const char* password);
  void close();

  // Get a line in a multiline response from the server
  // Return NULL when there are no more lines
  const char* getline();
  const char* line_pointer() { return lineptr; }

  // Return last command response
  int positive_status() { return rspstatus; }
  const char* response() { return rspbuffer; }
  const char* response_text() { return *rspbuffer ? (*rspbuffer == '+' ? rspbuffer+4 : rspbuffer+5) : rspbuffer; }

  ulong message_number() { return msg_number; }
  ulong message_size() { return msg_size; }

  int stat();
  int list(ulong number=0);
  int retr(ulong number);
  int dele(ulong number);
  int rset();
  int noop();
  int top(ulong number, ulong lines);

};


//  ------------------------------------------------------------------
//  SMTP class

class gsmtp {

protected:

  int rspcode;
  char rspbuffer[512];
  char cmdbuffer[512];
  char linebuffer[512];

  const char* lineptr;

  int get_response();

  int send_command(int ok_code);
  int send_command(const char* command=NULL);
  int send_command(const char* command, int ok_code);
  int send_command(int ok_code1, int ok_code2);

  void append_cmd_argument(const char* argument);

public:

  gsocket& socket;

  gsmtp(gsocket& s);
  ~gsmtp();

  void debug(int setting=true) { socket.debug = setting; }
  void blocking(int setting=true) { socket.blocking = setting; }

  int open(const char* server, const char* domain);
  void close();

  // Send line to the server. Sends ".CRLF" if line is NULL
  int putline(const char* line);
  
  // Return last command response
  int response_code() { return rspcode; }
  const char* response() { return rspbuffer; }
  const char* response_text() { return *rspbuffer ? rspbuffer+4 : rspbuffer; }

  int mail(const char* address);
  int rcpt(const char* address);
  int data();
  int rset();
  int help(const char* command=NULL);
};


//  ------------------------------------------------------------------

#endif

//  ------------------------------------------------------------------

