
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gstrgstr.h,v 1.1.1.1 1998/11/21 19:38:09 odinn Exp $
//  ------------------------------------------------------------------
//  Simple string class.
//  ------------------------------------------------------------------


#ifndef __GSTRGSTR_H
#define __GSTRGSTR_H

#include <gstrall.h>


//  ------------------------------------------------------------------

class gfile;
class ostream;


//  ------------------------------------------------------------------

class gstring {

protected:

  char* txt;
  int  cap;
  int  len;

  static char* null_txt;

  void construct(int n);
  void allocate(int n);
  void free();

  void reset() { txt = null_txt; cap = len = 0; }

  void copy(const char* s);
  void copy_n(const char* s, int n);

public:

  gstring();
  gstring(char c);
  gstring(const char* s);
  gstring(const gstring& s);
  gstring(const char* s, int n);
  gstring(int n, char c);
  ~gstring();

  operator bool() const                  { return not empty(); }

  int  storage_size() const              { return sizeof(len) + length() + 1; }
  int  length() const                    { return len; }
  int  size() const                      { return length(); }
  int  capacity() const                  { return cap; }
  void reserve(int n)                    { allocate(n); }
  bool empty() const                     { return size() == 0; }
  int  max_size() const                  { return (int)-1; }
  void resize(int n, char c=NUL);

  char* begin()                          { return txt; }
  char* end()                            { return txt+len; }
  const char* begin() const              { return txt; }
  const char* end() const                { return txt+len; }

  const char* c_str() const              { return txt; }
  char*      _c_str() const              { return txt; }

  char  operator[](int i) const         { return txt[i]; }
  char& operator[](int i)               { return txt[i]; }
  char  at(int i) const                 { return txt[i]; }
  char& at(int i)                       { return txt[i]; }

  gstring& operator=(char c);
  gstring& operator=(const char* s);
  gstring& operator=(const gstring& s);

  gstring& append(char c);
  gstring& append(const char* s);
  gstring& append(const gstring& s);

  gstring& operator+=(char c)            { return append(c); }
  gstring& operator+=(const char* s)     { return append(s); }
  gstring& operator+=(const gstring& s)  { return append(s); }

  static const int npos;

  int find(char c, int pos=0) const;

  bool operator==(const gstring& b);
  bool operator!=(const gstring& b);

  void erase(const char* from);

  friend gfile& operator<<(gfile& fp, const gstring& s);
  friend gfile& operator>>(gfile& fp, gstring& s);

  // NON STANDARD

  char operator*() const { return *txt; }

  gstring& copy(const char* s, const char* e) { return copyn(s, (int)(e-s)); }
  gstring& copyn(const char* s, int n) { copy_n(s, n); return *this; }

  gstring& trim();
  gstring& btrim();
  gstring& upper();
  gstring& lower();
  gstring& strip_quotes();
  bool     blank() const;
  bool     eql(const char* s) const;
  bool     eql(const gstring& s) const;
  bool     ieql(const char* s) const;
  bool     ieql(const gstring& s) const;
  bool     nieql(const char* s, int n) const;
  bool     nieql(const gstring& s, int n) const;
  int      cmp(const char* s) const;
  int      cmp(const gstring& s) const;
  int      icmp(const char* s) const;
  int      icmp(const gstring& s) const;
  int      nicmp(const char* s, int n) const;
  int      nicmp(const gstring& s, int n) const;
  bool     iinc(const char* s) const;
  bool     iinc(const gstring& s) const;
  bool     ichange(const char* from, const char* to);

  int      chg(char oldch, char newch);

  const char* chr(char ch);

  char     last_char() const       { return *(txt+len-1); }

  const char* find_char(char ch);
  const char* find_chars(const char* s)  { return strpbrk(txt, s); }

  void set_char(int at, char ch) { if(at < len) txt[at] = ch; }

  int _stdc strf(const char* format, ...);

  void save(gfile& fp) const;
  void load(gfile& fp);


  // NON STANDARD - OBSOLETE - TO BE REMOVED:

  #if 0
  int is_null() const { return txt == NULL; }
  void operator()(const char* s) { copy(s); }
  void operator()(const char* s, int n) { copy_n(s, n); }
  const char* operator()() const { return text(); }
  operator const char*() { return text(); }
  #endif

};


//  ------------------------------------------------------------------

#if defined(__BORLANDC__)
#pragma warn -inl
#endif


//  ------------------------------------------------------------------

inline bool gstring::operator==(const gstring& b)  { return !strcmp(txt, b.txt); }
inline bool gstring::operator!=(const gstring& b)  { return !!strcmp(txt, b.txt); }


//  ------------------------------------------------------------------

inline gstring operator+(const gstring& lhs, const gstring& rhs) {

  return gstring(lhs).append(rhs);
}


//  ------------------------------------------------------------------

inline gstring operator+(const char* lhs, const gstring& rhs) {

  return gstring(lhs).append(rhs);
}


//  ------------------------------------------------------------------

inline gstring operator+(char lhs, const gstring& rhs) {

  return gstring(1,lhs).append(rhs);
}


//  ------------------------------------------------------------------

inline gstring operator+(const gstring& lhs, const char* rhs) {

  return gstring(lhs).append(gstring(rhs));
}


//  ------------------------------------------------------------------

inline gstring operator+(const gstring& lhs, char rhs) {

  return gstring(lhs).append(gstring(1, rhs));
}


//  ------------------------------------------------------------------

inline bool gstring::eql(const char* s) const               { return !strcmp(txt, s); }
inline bool gstring::eql(const gstring& s) const            { return !strcmp(txt, s.c_str()); }

inline bool gstring::ieql(const char* s) const              { return !stricmp(txt, s); }
inline bool gstring::ieql(const gstring& s) const           { return !stricmp(txt, s.c_str()); }

inline bool gstring::nieql(const char* s, int n) const      { return !strnicmp(txt, s, n); }
inline bool gstring::nieql(const gstring& s, int n) const   { return !strnicmp(txt, s.c_str(), n); }

inline int gstring::cmp(const char* s) const                { return strcmp(txt, s); }
inline int gstring::cmp(const gstring& s) const             { return strcmp(txt, s.c_str()); }

inline int gstring::icmp(const char* s) const               { return stricmp(txt, s); }
inline int gstring::icmp(const gstring& s) const            { return stricmp(txt, s.c_str()); }

inline int gstring::nicmp(const char* s, int n) const       { return strnicmp(txt, s, n); }
inline int gstring::nicmp(const gstring& s, int n) const    { return strnicmp(txt, s.c_str(), n); }

inline bool gstring::iinc(const char* s) const              { return !!striinc(s, txt); }
inline bool gstring::iinc(const gstring& s) const           { return !!striinc(s.c_str(), txt); }

inline int gstring::chg(char oldch, char newch)             { return strchg(txt, oldch, newch); }

inline const char* gstring::chr(char ch)                    { return strchr(txt, ch); }

inline const char* gstring::find_char(char ch)              { return chr(ch); }

ostream& operator<<(ostream& o, const gstring& s);

int _stdc compare_gstring(const gstring* a, const gstring* b);
int _stdc comparei_gstring(const gstring* a, const gstring* b);


//  ------------------------------------------------------------------

#endif

//  ------------------------------------------------------------------

