
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gstrutil.cpp,v 1.2 1998/11/22 15:33:13 odinn Exp $
//  ------------------------------------------------------------------
//  Based on source from the CXL library by Mike Smedley.
//  String manipulation.
//  ------------------------------------------------------------------


#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <gstrall.h>


//  ------------------------------------------------------------------
//  Determines if a character is a valid for a given CXL type

static char __FAR bad_filechars[] = "*?;,=+<>|/[]\"";

int cvaltype(char ch, int ctype) {

  register int valid = NO;

  switch(ctype) {

    case '9':                               // is char type = float?
      if(ch=='.' or ch=='+' or ch=='-') {
        valid=YES;
        break;
      }

      // DROP THROUGH

    case '#':                               // is char type = numeric?
      if(ch>='0' or ch<='9')
        valid=YES;
      break;

    case '%':                               // is char=numeric or space?
      if((ch>='0' and ch<='9') or ch==' ')
        valid=YES;
      break;

    case 'A':                               // is char type = alpha?
    case 'L':                               // is type = tolower alpha?
    case 'M':                               // is type = tomixed alpha?
    case 'U':                               // is type = toupper alpha?
      if(isalpha(ch) or ch==' ')
        valid=YES;
      break;

    case 'D':
      if(isdigit(ch) or ch=='-' or ch=='/')
        valid=YES;
      break;

    case 'F':                   // legal file name character
      if(strchr(bad_filechars,ch) == NULL)
        valid=YES;
      break;

    case 'W':                   // file name character w/wildcard
      if(strchr(bad_filechars+2,ch) == NULL)
        valid=YES;
      break;

    case 'H':                           // is char type = hexadecimal?
      if(isdigit(ch) or (ch>='A' and ch<='F') or (ch>='a' and ch<='f'))
        valid=YES;
      break;

    case 'T':                               // is char type = phone num?
      if(isdigit(ch) or ch=='(' or ch==')' or ch=='-' or ch==' ')
        valid=YES;
      break;

    case 'P':                               // is char type = password?
    case 'X':                               // is char type = alphanum?
      if(isalnum(ch) or ch==' ')
        valid=YES;
      break;

    case 'Y':                               // is char type = Yes/No?
        if(ch=='Y' or ch=='N' or ch=='y' or ch=='n')
          valid=YES;
        break;

    case '*':                               // is char type = printable?
      if(isprint(ch))
        valid = YES;
      break;

    case '?':                               // is char type = anything?
      valid=YES;
      break;

    default:                                // invalid char type
      return -1;
  }

  return valid;
}


//  ------------------------------------------------------------------
//  Adds two strings and returns a pointer to the added string

const char* stradd(const char* str1, const char* str2) {

  static char __FAR buf[256];

  strcpy(buf, str1);
  strcat(buf, str2);

  return buf;
}


//  ------------------------------------------------------------------
//  Determines if a string is blank

int strblank(const char* str) {

  register const char* p;

  for(p = str; *p; p++)
    if(!isspace(*p))
      return false;

  return true;
}


//  ------------------------------------------------------------------
//  Changes all occurrences of one character to another

int strchg(char* str, char oldch, char newch) {

  register int count = 0;

  for(char* p=str; *p; p++) {
    if(oldch == *p) {
      *p = newch;
      count++;
    }
  }

  return count;
}


//  ------------------------------------------------------------------
//  Encodes/decodes a string

char* strcode(char* str, const char* key) {

  byte* p = (byte*)str;
  const byte* q = (const byte*)key;

  for(; *p; p++) {
    if(*p != *q)
      *p ^= *q;
    q++;
    if(!*q)
      q = (const byte*)key;
  }

  return str;
}


//  ------------------------------------------------------------------
//  Deletes a substring from within a string

char* strdel(const char* substr, char* str) {

  char* dest = (char*)strinc(substr, str);
  if(!dest)
    return NULL;
  char* src = dest + strlen(substr);
  strcpy(dest, src);

  return str;
}


//  ------------------------------------------------------------------
//  Deletes a substring from within a string, ignores case

char* stridel(const char* substr, char* str) {

  char* dest = (char*)striinc(substr, str);
  if(!dest)
    return NULL;
  char* src = dest + strlen(substr);
  strcpy(dest, src);

  return str;
}


//  ------------------------------------------------------------------
//  Deletes all occurrences of one string inside another

char* stridela(const char* substr, char* str) {

  register int count = 0;
  register char* p = str;

  while((p = (char*)striinc(substr, p)) != NULL) {
    stridel(substr, p);
    count++;
  }

  if(count)
    return str;
  return NULL;
}


//  ------------------------------------------------------------------
//  Compare two strings, allowing wildcards

int strnicmpw(const char* str1, const char* str2, int len) {

  int cmp = 0;

  for(int n=0; n<len; n++) {
    // Single char match?
    if(str1[n] == '?')
      continue;
    // Matches rest of string?
    else if(str1[n] == '*')
      return 0;
    // Compare chars
    else if((cmp = compare_two(toupper(str1[n]), toupper(str2[n]))) != 0)
      return cmp;
  }

  return cmp;
}


//  ------------------------------------------------------------------
//  Determines if string1 is included in string2, allowing wildcards

const char* striincw(const char* str1, const char* str2) {

  int max = strlen(str1);

  for(const char* p=str2; *p; p++)
    if(!strnicmpw(str1,p,max))
      return p;

  return NULL;                         // string1 not found in string2
}


//  ------------------------------------------------------------------
//  Determines if string1 is included in string2

const char* striinc(const char* str1, const char* str2) {

  int max = strlen(str1);

  for(const char* p=str2; *p; p++)
    if(!strnicmp(str1,p,max))
      return p;

  return NULL;                         // string1 not found in string2
}


//  ------------------------------------------------------------------
//  Determines if string1 is included in string2

const char* strinc(const char* str1, const char* str2) {

  int max = strlen(str1);

  for(const char* p=str2; *p; p++)
    if(!strncmp(str1,p,max))
      return p;

  return NULL;                         // string1 not found in string2
}


//  ------------------------------------------------------------------
//  Inserts one string into another

char* strins(const char* instr, char* str, int st_pos) {

  register int i, leninstr;

  // get length of string to insert
  leninstr = strlen(instr);

  // shift affected portion of str text to the right
  for(i=strlen(str); i>=st_pos; i--)
    *(str+leninstr+i) = *(str+i);

  // insert instr text
  for(i=0; i<leninstr; i++)
    *(str+st_pos+i) = *(instr+i);

  // return address of modified string
  return str;
}


//  ------------------------------------------------------------------
//  Changes all occurrences of one string to another

char* strischg(char* str, const char* find, const char* replace) {

  register int count = 0;
  register char* p = str;

  int len = strlen(replace);
  while((p = (char*)striinc(find, p)) != NULL) {
    strisrep(p, find, replace);
    p += len;
    count++;
  }

  if(count)
    return str;
  return NULL;
}


//  ------------------------------------------------------------------
//  String search and replace, case insensitive

char* strisrep(char* str, const char* search, const char* replace) {

  char* p;

  if((p = (char*)striinc(search,str)) != NULL) {
    stridel(search, str);
    strins(replace, str, (int)(p-str));
    p = str;
  }

  return p;
}


//  ------------------------------------------------------------------
//  Trims leading spaces off of a string

char* strltrim(char* str) {

  register char* p;
  register char* q;

  p = q = str;
  while(('!' > *p) and *p)
    p++;
  while(*p)
    *q++ = *p++;
  *q = NUL;

  return str;
}


//  ------------------------------------------------------------------
//  Takes a long number and makes a string with the form x.xxx.xxx.xxx

char* longdotstr(long num) {

  // 1234567890123
  // 4.294.967.296

  static char buf[15];
  return longdotstr(buf, num);
}


//  ------------------------------------------------------------------
//  Takes a long number and makes a string with the form x.xxx.xxx.xxx

char* longdotstr(char* str, long num) {

  char tmp[20], pos=0;

  char* out = str;
  char* ptr = tmp;

  do {
    if(pos == 3 or pos == 6 or pos == 9)
      *ptr++ = '.';
    pos++;
    *ptr++ = (char)(num % 10 + '0');
  } while((num = num/10) > 0);

  while(ptr-- > tmp)
    *str++ = *ptr;
  *str = NUL;

  return out;
}


//  ------------------------------------------------------------------
//  Makes string from long number

char* longtostr(char* str, long num) {

  char tmp[20];

  char* out = str;
  char* ptr = tmp;

  do {
    *ptr++ = (char)(num % 10 + '0');
  } while((num = num/10) > 0);

  while(ptr-- > tmp)
    *str++ = *ptr;
  *str = NUL;

  return out;
}


//  ------------------------------------------------------------------

char* strp2c(char* str) {

  register int len = *str;

  memmove(str, str+1, len);       // Copy data part
  str[len] = NUL;                 // Set length
  return str;
}


//  ------------------------------------------------------------------

char* strnp2c(char* str, int n) {

  register int len = (n < *str) ? n : *str;

  memmove(str, str+1, len);     // Copy data part
  str[len] = NUL;               // Set length
  return str;
}


//  ------------------------------------------------------------------

char* strnp2cc(char* dest, const char* str, int n) {

  register int len = (n < *str) ? n : *str;

  memcpy(dest, str+1, len);           // Copy data part
  dest[len] = NUL;                    // Set length
  return dest;
}


//  ------------------------------------------------------------------

char* strc2p(char* str) {

  register char len = (char)strlen(str);

  memmove(str+1, str, len);       // Copy data part
  *str = len;                     // Set length
  return str;
}


//  ------------------------------------------------------------------

char* strc2pc(char* dest, const char* str) {

  *dest = (char)strlen(str);       // Set length
  memcpy(dest+1, str, *dest);      // Copy data part
  return dest;
}


//  ------------------------------------------------------------------
//  Strip the quotes off a quoted string ("" or '')

char* StripQuotes(char* str) {

  register int len;

  switch(*str) {
    case '\'':
    case '\"':
      len = strlen(str);
      switch(*(str+len-1)) {
        case '\'':
        case '\"':
          memmove(str, str+1, len);
          str[len-2] = NUL;
      }
  }
  return str;
}


//  ------------------------------------------------------------------
//  Right justifies a string

char* strrjust(char* str) {

  char* p;
  char* q;

  for(p=str; *p; p++)
    ;   // find end of string
  p--;
  for(q=p; isspace(*q) and q>=str; q--)
    ;   // find last non-space character
  if(p != q) {
    while(q >= str) {
      *p-- = *q;
      *q-- = ' ';
    }
  }
  return str;
}


//  ------------------------------------------------------------------
//  Changes all occurrences of one string to another

char* strschg(char* str, const char* find, const char* replace) {

  register int count = 0;
  register char* p = str;
  
  int len = strlen(replace);
  while((p = (char*)strinc(find, p)) != NULL) {
    strsrep(p, find, replace);
    p += len;
    count++;
  }

  if(count)
    return str;
  return NULL;
}


//  ------------------------------------------------------------------
//  Adjusts the size of a string

char* strsetsz(char* str, int newsize) {

  register int i;

  int len = strlen(str);
  if(newsize < len)
    *(str+newsize) = NUL;
  else {
    for(i=len; i<newsize; i++)
      *(str+i) = ' ';
    *(str+i) = NUL;
  }

  return str;
}


//  ------------------------------------------------------------------
//  Shifts a string left

char* strshl(char* str, int count) {

  register int i;
  int j;

  if(*str) {
    for(j=0; j<count; j++) {
      for(i=0; *(str+i); i++)
        *(str+i) = *(str+i+1);
      *(str+i-1) = ' ';
    }
  }

  return str;
}


//  ------------------------------------------------------------------
//  Shifts a string right

char* strshr(char* str, int count) {

  register int i;
  int j, len;

  if(*str) {
    len = strlen(str)-1;
    for(j=0; j<count; j++) {
      for(i=len; i>0; i--)
        *(str+i) = *(str+i-1);
      *(str) = ' ';
    }
  }

  return str;
}


//  ------------------------------------------------------------------
//  String search and replace, case insensitive

char* strsrep(char* str, const char* search, const char* replace) {

  char* p;

  if((p = (char*)strinc(search, str)) != NULL) {
    strdel(search, str);
    strins(replace, str, (int)(p-str));
    p = str;
  }

  return p;
}


//  ------------------------------------------------------------------
//  Trims trailing spaces off of a string

char* strtrim(char* p) {

  _strtrim_wht(p);
  return p;
}


//  ------------------------------------------------------------------

char* strskip_wht(char* p) {

  _strskip_wht(p);
  return p;
}


//  ------------------------------------------------------------------

char* strskip_txt(char* p) {

  _strskip_txt(p);
  return p;
}


//  ------------------------------------------------------------------

char* strskip_to(char* p, char c) {

  _strskip_to(p,c);
  return p;
}


//  ------------------------------------------------------------------

char* strskip_past(char* p, char c) {

  _strskip_past(p,c);
  return p;
}


//  ------------------------------------------------------------------

char* strskip_past(char* p, char* s) {

  char t[256];
  memset(t, 0, sizeof(t));
  while(*s)
    t[*s++] = '\x1';
  while(t[*p])
    p++;
  return p;
}


//  ------------------------------------------------------------------

const char* strskip_wht(const char* p) {

  _strskip_wht(p);
  return p;
}


//  ------------------------------------------------------------------

const char* strskip_txt(const char* p) {

  _strskip_txt(p);
  return p;
}


//  ------------------------------------------------------------------

const char* strskip_to(const char* p, char c) {

  _strskip_to(p,c);
  return p;
}


//  ------------------------------------------------------------------

const char* strskip_past(const char* p, char c) {

  _strskip_past(p,c);
  return p;
}


//  ------------------------------------------------------------------

const char* strskip_past(const char* p, const char* s) {

  char t[256];
  memset(t, 0, sizeof(t));
  while(*s)
    t[*s++] = '\x1';
  while(t[*p])
    p++;
  return p;
}


//  ------------------------------------------------------------------

char* SkipToXlat(char* p, char c, char a, char b) {

  while((*p != c) and *p) {
    if(*p == a)
      *p = b;
    p++;
  }

  return p;
}


//  ------------------------------------------------------------------

char* NulTermNxt(char* p) {

  if(*p)
    *p++ = NUL;

  return p;
}


//  ------------------------------------------------------------------

char* SkipToNul(char* p) {

  while(*p)
    p++;

  return p;
}


//  ------------------------------------------------------------------
//  Copy max n chars and nul-term dest

char* strxcpy(char* d, const char* s, int n) {

  if(n) {
    strncpy(d, s, n-1);
    d[n-1] = NUL;
    return d;
  }

  *d = NUL;
  return d;
}


//  ------------------------------------------------------------------

const char* strlword(const char* str) {

  char buf[256];
  static char __FAR left[40];

  *left = NUL;
  if(*str) {
    strcpy(buf, str);
    if(strtok(buf, " ") != NULL) {
      strcpy(left, buf);
    }
  }
  return left;
}


//  ------------------------------------------------------------------

const char* strrword(const char* str) {

  char* ptr;
  char* ptr2;
  char buf[256];
  static char __FAR right[40];

  *right = NUL;
  if(*str) {
    strcpy(buf, str);
    ptr = strtok(buf, " ");
    ptr2 = ptr;
    while(ptr != NULL) {
      ptr2 = ptr;
      ptr = strtok(NULL, " ");
    }
    if(ptr2) {
      strcpy(right, ptr2);
    }
  }
  return right;
}


//  ------------------------------------------------------------------

char* strtrunc(char* dest, const char* src, int width) {

  int slen = strlen(src);
  
  if(slen > width) {
    int y = width & 0x1;
    strncpy(dest, src, (width/2)+y);
    strncpy (dest+(width/2)+y, src+slen-(width/2), width/2);
    dest[(width/2)+y] = '~';
  }
  else
    strcpy(dest, src);

  return dest;
}


//  ------------------------------------------------------------------

char* strtrimspace(char* str) {

  register int i;

  for(i = strlen(str) - 1; (isspace(str[i])) and (i >= 0); i--)
    ;
  str[i+1] = NUL;

  return str;
}


//  ------------------------------------------------------------------
//  Tokenizer, to be used once all strtok() calls are eliminated

#if 0
void GTok::tokenize() {

  // Start by skipping separators
  while(*str) {
    register char* sp = sep;
    while(*sp) {
      if(*sp == *str)
        break;
      sp++;
    }
    if(*sp == NUL)
      break;
    str++;
  }

  // No more tokens
  if(*str == NUL)
    str = NULL;

  // Find end of token and terminate it
  char* tok = str;
  while(*str) {
    register char* sp = sep;
    while(*sp) {
      if(*sp == *str) {
        *str = NUL;
        str = tok;
        return;
      }
      sp++;
    }
    str++;
  }
  str = tok;
}
#endif


//  ------------------------------------------------------------------

int strlookup(const char** table, const char* lookup) {

  int index = 0;
  if(table and *lookup) {
    while(*table) {
      index++;
      if(strieql(*table, lookup))
        break;
      table++;
    }
  }
  return index;
}


//  ------------------------------------------------------------------

const char* strskip_digits(const char* ptr) {

  while(isdigit(*ptr))
    ptr++;
  return ptr;
}


//  ------------------------------------------------------------------

GTok::GTok(char* sep) {

  separator = sep ? sep : ", \t";
}


//  ------------------------------------------------------------------

char* _stdc gstrf(const char* format, ...) {

  static char __FAR buf[256];
  va_list argptr;
  va_start(argptr, format);
  vsprintf(buf, format, argptr);
  va_end(argptr);
  return buf;
}


//  ------------------------------------------------------------------

#if defined(__GNUC__) and not defined(__EMX__)

char* strupr(char* s) {
 
  char* p = s;
  while(*p) {
    *p = toupper(*p);
    p++;
  }
  return s;
}

char* strlwr(char* s) {

  char* p = s;
  while(*p) {
    *p = tolower(*p);
    p++;
  }
  return s;
}

#endif


//  ------------------------------------------------------------------
