
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gtag.cpp,v 1.1.1.1 1998/11/21 19:38:07 odinn Exp $
//  ------------------------------------------------------------------
//  Tag number types and set class.
//  ------------------------------------------------------------------


#if defined(__BORLANDC__)
//#pragma option -Od
#endif

#include <stdlib.h>
#include <string.h>
#include <gmemdbg.h>
#include <gtag.h>


//  ------------------------------------------------------------------
//  Constructor.

gtag::gtag() {

  granularity = 10;

  tag = NULL;
  allocated = tags = 0;
}


//  ------------------------------------------------------------------
//  Destructor.

gtag::~gtag() {

  reset();
}


//  ------------------------------------------------------------------
//  Deallocate and reset internal data.

void gtag::reset() {

  throw_release(tag);
  allocated = tags = 0;
}


//  ------------------------------------------------------------------
//  Resize tag array.
//  Returns NULL if realloc failed.

ulong* gtag::resize(uint newtags) {

  register uint newsize = 0;

  if(newtags >= allocated)
    newsize = newtags + granularity;
  else if(newtags < allocated) {
    if((allocated-newtags) > granularity)
      newsize = newtags + granularity;
  }

  if(newsize) {
    tag = (ulong*)throw_realloc(tag, newsize*sizeof(ulong));
    allocated = newsize;
  }

  tags = newtags;
  return tag;
}


//  ------------------------------------------------------------------
//  Appends a new tag number to the end of the tag array.
//  NOTE - Does not check for duplicates or correct sequence!

ulong* gtag::append(ulong tagn) {

  resize(tags+1);
  tag[tags-1] = tagn;
  return tag;
}


//  ------------------------------------------------------------------
//  Add a new tag number at the correct position of the tag array.
//
//  Algorithm dry test situations:
//
//  []        0123456
//  tag       345789x  tags = 7 (after resize)
//  reln      1234567
//
//  Case 1 - New tag is smaller than first tag in array:
//    tagn = 2
//    reln = 1 (changed to 0)
//    tags-reln-1 = 7-0-1 = 6
//
//  Case 2 - Normal insert in middle:
//    tagn = 6
//    reln = 3
//    tags-reln-1 = 7-3-1 = 3
//
//  Case 3 - New tag is larger than last tag in array:
//    tagn = 10
//    reln = 6
//    tags-reln-1 = 7-6-1 = 0
//
//  Dry test result: Works!


ulong* gtag::add(ulong tagn) {

  // Find closest tag number
  uint reln = absolute_to_relative(tagn, GTAGN_CLOSEST);

  // Do we have it already?
  if((reln == GRELN_INVALID) or (tag[reln-1] != tagn)) {

    // Resize tag array to make room for the new number
    resize(tags+1);

    //  General rule:
    //    - tag[_reln-1] is smaller than __tagn
    //
    //  The exceptions to the rule:
    //    - no tags in array (if _reln == GRELN_INVALID)
    //    - tag[_reln-1] is larger than __tagn

    if(reln and (tag[reln-1] > tagn))
      reln--;

    // Move data to make room for the new tag number
    memmove(tag+reln+1, tag+reln, (tags-reln-1)*sizeof(ulong));

    // Copy the new tag number to the insert position
    tag[reln] = tagn;
  }

  return tag;
}


//  ------------------------------------------------------------------
//  Delete a tag number from the tag array.
//  Returns the relative tag number or GRELN_INVALID if missing.
//  NOTE - Does not resize the array.

uint gtag::delete_absolute(ulong absolute) {

  return delete_relative(absolute_to_relative(absolute));
}


//  ------------------------------------------------------------------

uint gtag::delete_relative(uint reln) {

  if(reln) {
    memmove(tag+reln-1, tag+reln, (tags-reln)*sizeof(ulong));
    tags--;
  }
  return reln;
}


//  ------------------------------------------------------------------

uint gtag::delete_absolute_resize(ulong tagn) {

  uint reln = delete_absolute(tagn);
  resize(tags);
  return reln;
}


//  ------------------------------------------------------------------

void gtag::sort() {

  qsort(tag, tags, sizeof(ulong), (StdCmpCP)gtag_compare);
}


//  ------------------------------------------------------------------

void gtag::eliminate_duplicates() {

  if(tags > 1) {
    uint before = tags;
    ulong last = tag[0];
    for(uint n=1; n<tags; n++) {
      if(last == tag[n]) {
        memmove(&tag[n-1], &tag[n], (tags-n)*sizeof(ulong));
        tags--;
        n--;
      }
      last = tag[n];
    }
    if(before != tags)
      resize(tags);
  }
}


//  ------------------------------------------------------------------
//  Return the tag number corresponding to a relative tag number.
//  Returns GTAGN_INVALID, if the requested number did not exist.

ulong gtag::relative_to_absolute(uint reln) {

  if(tag and reln and (reln <= tags))
    return tag[reln-1];
  else
    return GTAGN_INVALID;
}


//  ------------------------------------------------------------------
//  Return the relative tag number corresponding to a tag number.
//  Returns GRELN_INVALID if the requested number did not exist.
//  If __closest is true, the closest smaller tag number is returned.
//  If no tags exist at all, GRELN_INVALID is returned.

uint gtag::absolute_to_relative(ulong tagn, bool closest) {

  uint lastreln = GRELN_INVALID;

  if(tag) {

    if(tagn and tags)
      if(tagn > tag[tags-1])
        return closest ? tags : GRELN_INVALID;

    if(tags and tagn) {

      register long mid;
      register long left = 0;
      register long right = tags;

      do {
        mid = (left+right)/2;
        if(tagn < tag[(uint)mid])
          right = mid - 1;
        else if(tagn > tag[(uint)mid])
          left = mid + 1;
        else
          return (uint)(mid + 1);
      } while(left < right);

      lastreln = (uint)(left + 1);
      if(tagn == tag[(uint)left])
        return lastreln;
    }
  }

  // Not found
  return closest ? lastreln : GRELN_INVALID;
}


//  ------------------------------------------------------------------

uint gtag::absolute_to_relative(ulong tagn) {

  return absolute_to_relative(tagn, GTAGN_EXACT);
}


//  ------------------------------------------------------------------

int _stdc gtag_compare(const ulong* a, const ulong* b) {

  return compare_two(*a, *b);
}


//  ------------------------------------------------------------------

