
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gthing.h,v 1.1.1.1 1998/11/21 19:38:09 odinn Exp $
//  ------------------------------------------------------------------
//  Thing class.
//  ----------------------------------------------------------------
/*

Usage:
  Set block_size to a suitable value.
  Set auto_increment if you want the block automatically enlarged as needed.
  Call the initialize() member.
  The object is now ready to use.

To use as an array:
  Set auto_pack_index to true.
  Allocation:    Use insert() and add().
  De-allocation: Use remove() or erase().

To use as a bag of handle-accessed chunks:
  Set auto_pack_index to false.
  Allocation:    Use allocate().
  De-allocation: Use free().

In any case:
  Re-allocation: Use reallocate().

*/
//  ------------------------------------------------------------------

#ifndef __G_THING_H
#define __G_THING_H

#include <gdefs.h>


//  ------------------------------------------------------------------

class thing_base {

protected:

  // Refresh the free_space variable.
  void calc_free_space();

  // Refresh the index pointer.
  void calc_index();

  // Increase the size of the block.
  // s: The number of bytes to add to the block.
  void enlarge_block(uint s);

  // Adjust the offsets in the index.
  // o: Adjust all offsets >= this value.
  // d: The number of bytes to add to each offset. May be negative.
  void adjust_index_offsets(uint o, int d);

  // Return true the entry is the last in the index.
  int is_last(int h) const { return h == index_size-1; }

public:

  // Allocation index structure
  struct idx {
    uint offset;
    uint length;
  };

  // Pointer to the block.
  char* block;

  // Size of the block.
  uint block_size;

  // Space which is allocated.
  uint allocated;

  // Space which can be allocated.
  uint free_space;

  // Index to the allocations.
  idx* index;

  // Number of allocations (including any freed entries).
  int index_size;

  // Number of free entries in the index.
  int index_free;

  // If true, freed entries are automatically removed from the index.
  int auto_pack_index;

  // If non-zero, the memory block is automatically enlarged by a
  // multiple of the auto_increment value when needed.
  uint auto_increment;

  // constructor and destructor
  thing_base(uint b, uint i);
  ~thing_base();

  // Initial memory block allocation etc.
  // The block_size must be set first.
  void initialize();

  // De-allocate memory block.
  void reset();

  // Manually pack the index (to be used when auto_pack_index is false).
  void pack_index();

  // Allocate from the block.
  // Return a handle, or -1 if out of space.
  int allocate(uint s);

  // Allocate and insert at a particular position in the index.
  // NOTE: Invalidates all handles from h and up.
  // Return a handle, or -1 if out of space.
  int insert(int h, uint s);

  // Allocate and put at the end of the index.
  // Return a handle, or -1 if out of space.
  int add(uint s) { return insert(index_size, s); }

  // Change the size of a chunk.
  // Return a handle to the chunk, or -1 if out of space.
  int reallocate(int h, uint s);

  // De-allocate a chunk.
  // If auto_pack_index is false, the handle will point to a free
  // entry, which may be re-used in a later allocation.
  void free(int h, int r=false);

  // De-allocate all chunks.
  // If auto_pack_index is false, the index entries will be left with
  // free entries, which may be re-used in a later allocation.
  void free_all(int r=false);

  // De-allocate a chunk.
  // Also removes the index entry, regardless of the
  // auto_pack_index setting.
  void remove(int h) { free(h, true); }
  void erase(int h) { remove(h); }

  // De-allocate all chunks.
  // Also removes the index entries.
  void remove_all() { free_all(true); }

  // Return the offset of a chunk.
  uint offset(uint h) const { return index[h].offset; }

  // Return the length of a chunk.
  uint length(uint h) const { return index[h].length; }

  // Return a pointer to the chunk.
  char* pointer(uint h) const { return block+index[h].offset; }

  // Return a pointer to the chunk, using the subscript operator.
  char* operator[](uint h) const { return pointer(h); }

  // Return true if there are no allocations.
  int empty() const { return size() == 0; }

  // Return number of allocations.
  uint size() const { return index_size - index_free; }

  // Return handle to first element.
  int begin() const { return 0; }

  // Return a past-the-end handle.
  int end() const { return index_size + 1; }

  // Allocate and put at the front.
  int push_front(uint s) { return insert(begin(), s); }

  // Allocate and put at the back.
  int push_back(uint s) { return insert(end(), s); }

  // Remove from the front.
  void pop_front() { erase(begin()); }

  // Remove from the back.
  void pop_back() { erase(end()-1); }
};


//  ------------------------------------------------------------------

class string_thing : public thing_base {

public:

  int allocate(char* s);
  int insert(int h, char* s);
  int add(char* s);
  int add(char* s, int n);
  int reallocate(int h, char* s);
};


//  ------------------------------------------------------------------

#endif

//  ------------------------------------------------------------------

