
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gutlbags.h,v 1.1.1.1 1998/11/21 19:38:09 odinn Exp $
//  ------------------------------------------------------------------
//  Bag classes.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------

#ifndef __GUTLBAGS_H
#define __GUTLBAGS_H


//  ------------------------------------------------------------------

#include <gdefs.h>
#include <string.h>


//  ------------------------------------------------------------------
//  Other classes used

class gfile;


//  ------------------------------------------------------------------

class gmembag {

protected:

  struct entry {
    uint offset;
    uint length;
  };

  enum { entry_size = sizeof(entry) };

  char* block;            // Pointer to allocated block
  uint block_size;        // Size of allocated block
  uint free_offset;       // Offset to free data space
  uint free_space;        // Size of free data space
  entry* item;            // Pointer to entries list
  entry* curr_item;       // Pointer to current item
  int curr_no;            // Current item number (1-based)
  int increment;          // Number of bytes to auto-add if out of space

  void enlarge_block(uint space_wanted, uint current_free_space);

public:

  int free_entries;       // Number of free entries
  int items;              // Total number of items (including free entries)
  char* current;          // Pointer to current item data

  gmembag();
  ~gmembag();

  void initialize(uint start_size, int increment_size=0);
  void destroy();
  void reset();

  uint free_space_left();

  char* free_space_pointer();

  int add(void* data, uint length);
  int change(int handle, void* data, uint new_length);
  void remove(int handle);

  char* get_pointer();
  char* get_pointer(int handle);

  uint get_length();
  uint get_length(int handle);

  char* operator[](int index);

  int first();
  int next();

  void save(gfile& f);
  void load(gfile& f);
};


//  ------------------------------------------------------------------

class gstrbag : public gmembag {

public:

  int add(char* str);
  int change(int handle, char* str);
  int find(char* str);
  int findi(char* str);
};


//  ------------------------------------------------------------------

inline uint gmembag::free_space_left() {

  return free_space - (free_entries ? 0 : entry_size);
}


//  ------------------------------------------------------------------

inline char* gmembag::free_space_pointer() {

  return block + free_offset;
}


//  ------------------------------------------------------------------

inline char* gmembag::get_pointer() {

  return block + curr_item->offset;
}


//  ------------------------------------------------------------------

inline char* gmembag::get_pointer(int handle) {

  return block + item[handle].offset;
}


//  ------------------------------------------------------------------

inline uint gmembag::get_length() {

  return curr_item->length;
}


//  ------------------------------------------------------------------

inline uint gmembag::get_length(int handle) {

  return item[handle].length;
}


//  ------------------------------------------------------------------

inline char* gmembag::operator[](int index) {

  return get_pointer(index);
}


//  ------------------------------------------------------------------

inline int gstrbag::add(char* str) {

  return gmembag::add(str, strlen(str)+1);
}


//  ------------------------------------------------------------------

inline int gstrbag::change(int handle, char* str) {

  return gmembag::change(handle, str, strlen(str)+1);
}


//  ------------------------------------------------------------------

class gbag_base {

protected:

  struct entry {
    uint offset;
    uint length;
  };

  enum { entry_size = sizeof(entry) };

  enum { stack_size = 3 };

  struct stack {
    int    curr_no;
    entry* curr_item;
  };

  char* block;            // Pointer to allocated block
  uint block_size;        // Size of allocated block
  uint free_offset;       // Offset to free data space
  uint free_space;        // Size of free data space
  entry* item;            // Pointer to entries list
  entry* curr_item;       // Pointer to current item
  int curr_no;            // Current item number (1-based)
  int increment;          // Number of bytes to auto-add if out of space
  stack stk[stack_size];  // Small stack for push/pop of the current item
  int stk_ptr;            // Stack pointer

  void enlarge_block(uint space_wanted, uint current_free_space);

public:

  int free_entries;       // Number of free entries
  int items;              // Total number of items (including free entries)
  char* curr_data;        // Pointer to current item data

  gbag_base();
  ~gbag_base();

  void initialize(uint start_size, int increment_size=0);
  void destroy();
  void reset();

  uint free_space_left() const;

  char* free_space_pointer() const;

  int add(const void* data, uint length);
  int change(int handle, const void* data, uint new_length);
  void remove(int handle);

  const char* get_pointer() const;
  const char* get_pointer(int handle) const;

  uint get_length() const;
  uint get_length(int handle) const;

  const char* operator[](int index) const;

  int first();
  int next();

  void push();
  void pop();

  void save(gfile& f);
  void load(gfile& f);
};


//  ------------------------------------------------------------------

template<class T>
class gbag : public gbag_base {

public:

  T* current;

  int add(const T& data)  { int tmp = gbag_base::add(&data, sizeof(T)); current = (T*)curr_data; return tmp; }

  int first()  { int tmp = gbag_base::first(); if(tmp) current = (T*)curr_data; return tmp; }
  int next()   { int tmp = gbag_base::next(); if(tmp) current = (T*)curr_data; return tmp; }

  void pop()  { gbag_base::pop(); current = (T*)curr_data; }

  const T& operator[](int index) const { return *((T*)get_pointer(index)); }
};


//  ------------------------------------------------------------------

class gstrbagx : public gbag_base {

public:

  int add(const char* str);
  int change(int handle, const char* str);
  int find(const char* str);
  int findi(const char* str);

  const char* get(int handle) const { return get_pointer(handle); }
  const char* set(int handle, const char* str) { change(handle, str); return get(handle); }
};


//  ------------------------------------------------------------------
//  Bag of integers

typedef gbag<int> gintbag;


//  ------------------------------------------------------------------

inline uint gbag_base::free_space_left() const {

  return free_space - (free_entries ? 0 : entry_size);
}


//  ------------------------------------------------------------------

inline char* gbag_base::free_space_pointer() const {

  return block + free_offset;
}


//  ------------------------------------------------------------------

inline const char* gbag_base::get_pointer() const {

  return block + curr_item->offset;
}


//  ------------------------------------------------------------------

inline const char* gbag_base::get_pointer(int handle) const {

  return block + item[handle].offset;
}


//  ------------------------------------------------------------------

inline uint gbag_base::get_length() const {

  return curr_item->length;
}


//  ------------------------------------------------------------------

inline uint gbag_base::get_length(int handle) const {

  return item[handle].length;
}


//  ------------------------------------------------------------------

inline const char* gbag_base::operator[](int index) const {

  return get_pointer(index);
}


//  ------------------------------------------------------------------

inline int gstrbagx::add(const char* str) {

  return gbag_base::add(str, strlen(str)+1);
}


//  ------------------------------------------------------------------

inline int gstrbagx::change(int handle, const char* str) {

  return gbag_base::change(handle, str, strlen(str)+1);
}


//  ------------------------------------------------------------------

#endif

//  ------------------------------------------------------------------

