
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gvector.h,v 1.1.1.1 1998/11/21 19:38:09 odinn Exp $
//  ------------------------------------------------------------------
//  Simple STL-style vector class.
//  ------------------------------------------------------------------


#ifndef __GVECTOR_H
#define __GVECTOR_H


//  ------------------------------------------------------------------

#include <gutlmisc.h>
#include <gmemory.h>
#include <gfile.h>


//  ------------------------------------------------------------------

template <class T>
class gvector {

public:

  typedef T value_type;
//  typedef gallocator<T>                     vector_allocator;
  typedef value_type&                       reference;
  typedef const value_type&                 const_reference;
  typedef value_type*                       iterator;
  typedef const value_type*                 const_iterator;
  typedef size_t                            size_type;

protected:

  gallocator<T> the_allocator;

  iterator start;
  iterator finish;
  iterator end_of_storage;

  void insert_aux(iterator position, const T& x) {
    if(finish != end_of_storage) {
      construct(finish, *(finish - 1));
      copy_backward(position, finish - 1, finish);
      *position = x;
      ++finish;
    }
    else {
      #if defined(__BORLANDC__)
      const size_type CHUNKSIZE = sizeof(T) >= 1024 ? 1 : 1024 / sizeof(T);
      #else
      size_type CHUNKSIZE = 1024 / sizeof(T);
      if(CHUNKSIZE == 0)
        CHUNKSIZE = 1;
      #endif
      size_type len = size() ? 2 * size() : CHUNKSIZE;
      iterator tmp = the_allocator.allocate(len);
      uninitialized_copy(begin(), position, tmp);
      construct((tmp + (position - begin())), x);
      uninitialized_copy(position, end(), tmp + (position - begin()) + 1);
      destroy(begin(), end());
      the_allocator.deallocate(begin());
      end_of_storage = tmp + len;
      finish = tmp + size() + 1;
      start = tmp;
    }
  }

public:

  gvector() : start(0), finish(0), end_of_storage(0) { }
  gvector(const gvector<T>& x) {
    start = the_allocator.allocate(x.end() - x.begin());
    finish = uninitialized_copy(x.begin(), x.end(), start);
    end_of_storage = finish;
  }

  ~gvector () {
    destroy(start, finish); the_allocator.deallocate(start);
  }

  gvector<T>& operator=(const gvector<T>& x) {
    if(&x == this)
      return *this;
    if(x.size() > capacity()) {
      destroy(start, finish);
      the_allocator.deallocate(start);
      start = the_allocator.allocate(x.end() - x.begin());
      end_of_storage = uninitialized_copy(x.begin(), x.end(), start);
    }
    else if(size() >= x.size()) {
      iterator i = copy(x.begin(), x.end(), begin());
      destroy(i, finish);
    }
    else {
      copy(x.begin(), x.begin() + size(), begin());
      uninitialized_copy(x.begin() + size(), x.end(), begin() + size());
    }
    finish = begin() + x.size();
    return *this;
  }

  iterator       begin()        { return start;  }
  const_iterator begin() const  { return start;  }
  iterator       end()          { return finish; }
  const_iterator end()   const  { return finish; }
  iterator       last()         { return end()-1; }
  const_iterator last()  const  { return end()-1; }

  size_type size()     const  { return size_type(end() - begin()); }
  size_type max_size() const  { return the_allocator.max_size(); }

  void resize(size_type new_size) {
    T value;
    if(new_size > size())
      insert(end(), new_size - size(), value);
    else if(new_size < size())
      erase(begin() + new_size, end());
  }

  void resize (size_type new_size, T value) {
    if(new_size > size())
      insert(end(), new_size - size(), value);
    else if(new_size < size())
      erase(begin() + new_size, end());
  }

  size_type capacity() const  { return size_type(end_of_storage - begin()); }
  bool      empty() const     { return begin() == end();                    }

  void reserve(size_type n) {
    if(capacity() < n) {
      iterator tmp = the_allocator.allocate(n);
      uninitialized_copy(begin(), end(), tmp);
      destroy(start, finish);
      the_allocator.deallocate(start);
      finish = tmp + size();
      start = tmp;
      end_of_storage = begin() + n;
    }
  }

  reference       operator[](size_type n)       { return *(begin() + n); }
  const_reference operator[](size_type n) const { return *(begin() + n); }
  reference       at(size_type n)               { return *(begin() + n); }
  const_reference at(size_type n) const         { return *(begin() + n); }
  reference       front()                       { return *begin();       }
  const_reference front() const                 { return *begin();       }
  reference       back()                        { return *(end() - 1);   }
  const_reference back() const                  { return *(end() - 1);   }

  void push_back(const T& x) {
    if(finish != end_of_storage) {
      construct(finish, x);
      finish++;
    }
    else
      insert_aux(end(), x);
  }
  void pop_back() {
    --finish; destroy(finish);
  }

  iterator insert(iterator position) {
    size_type n = position - begin();
    T x;
    if(finish != end_of_storage and position == end()) {
      construct(finish, x);
      finish++;
    }
    else
      insert_aux(position, x);
    return begin() + n;
  }

  iterator insert(iterator position, const T& x) {
    size_type n = position - begin();
    if(finish != end_of_storage and position == end()) {
      construct(finish, x);
      finish++;
    }
    else
      insert_aux(position, x);
    return begin() + n;
  }

  void insert(iterator position, size_type n, const T& x) {
    if(n == 0)
      return;
    if(size_type(end_of_storage - finish) >= n) {
      if(size_type(end() - position) > n) {
        uninitialized_copy(end() - n, end(), end());
        copy_backward(position, end() - n, end());
        fill(position, position + n, x);
      }
      else {
        uninitialized_copy(position, end(), position + n);
        fill(position, end(), x);
        uninitialized_fill_n(end(), n - (end() - position), x);
      }
      finish += n;
    }
    else {
      size_type len = size() + maximum_of_two(size(), n);
      iterator tmp = the_allocator.allocate(len);
      uninitialized_copy(begin(), position, tmp);
      uninitialized_fill_n(tmp + (position - begin()), n, x);
      uninitialized_copy(position, end(), tmp + (position - begin() + n));
      destroy(begin(), end());
      the_allocator.deallocate(begin());
      end_of_storage = tmp + len;
      finish = tmp + size() + n;
      start = tmp;
    }
  }

  void erase (iterator position) {
    if(position + 1 != end())
      copy(position + 1, end(), position);
    --finish;
    destroy(finish);
  }

  void erase (iterator first, iterator last) {
    iterator i = copy(last, end(), first);
    destroy(i, finish);
    finish = finish - (last - first);
  }

  typedef int (* _stdc compare_type)(const_iterator, const_iterator);

  void qsort(compare_type compare)  { ::qsort(start, size(), sizeof(T), (StdCmpCP)compare); }

  iterator lfind(const_reference key, compare_type compare) { size_t num = size(); return (iterator)::lfind(&key, start, &num, sizeof(T), (StdCmpCP)compare); }
  iterator bfind(const_reference key, compare_type compare) { return (iterator)::bfind(&key, start, size(), sizeof(T), (StdCmpCP)compare); }
  iterator bsearch(const_reference key, compare_type compare) { return (iterator)::bsearch(&key, start, size(), sizeof(T), (StdCmpCP)compare); }

};


//  ------------------------------------------------------------------

#if 0
template <class T>
void gvector_save(gfile& f, const T& v) {

  f << v.size();
  T::const_iterator i = v.begin();
  while(i != v.end()) {
    i->save(f);
    i++;
  }
}


//  ------------------------------------------------------------------

template <class T>
void gvector_load(gfile& f, T& v) {

  int size;
  f >> size;
  v.resize(0);
  v.resize(size);
  T::iterator i = v.begin();
  while(i != v.end()) {
    i->load(f);
    i++;
  }
}
#endif

//  ------------------------------------------------------------------

template <class T>
int _stdc compare_types(const T* a, const T* b) {

  return compare_two(*a, *b);
}

int _stdc compare_int(const int* a, const int* b);
int _stdc compare_uint(const uint* a, const uint* b);
int _stdc compare_long(const long* a, const long* b);
int _stdc compare_ulong(const ulong* a, const ulong* b);


//  ------------------------------------------------------------------

#endif

//  ------------------------------------------------------------------

