
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Library General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Library General Public License for more details.
//
//  You should have received a copy of the GNU Library General Public
//  License along with this library; if not, write to the Free
//  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gs_im130.h,v 1.1.1.1 1998/11/21 19:38:11 odinn Exp $
//  ------------------------------------------------------------------

/* ============================================================== */
/*                                                                */
/*               IMAIL DEFINITIONS and STRUCTURES                 */
/*                                                                */
/* ============================================================== */

/*
 * Copyright (C) 1990-1992 by Fabiano Fabris. All rights reserved.
 *
 * The contents of this file are subject to change without notice!
 * These structures are those implemented in IMAIL 1.30, and are not
 * compatible with previous versions.
 * Fields marked 'reserved' should not be used.
 * Fields marked 'rsv_iutx' (where x is a number) are reserved for
 * IUTIL or for future version of IMAIL.
 * A few notes:
 * 1) Test the semaphore file IMBUSY.NOW (in the IMAIL "home" directory)
 *    before writing to any of the IMAIL config files. If the file
 *    exists, wait for it to be removed.
 * 2) It is an ERROR to write to any of the configuration files
 *    or to the message base(s) while IMTHINGS is running; IMTHINGS
 *    (particularly the message base maint functions) requires exclusive
 *    access to the message base(s).
 */

#include <direct.h>
#include <io.h>
#include <time.h>

#define IMAIL_MAJ_VERSION    1
#define IMAIL_MIN_VERSION    30
#define IM_PRD_CODE          0x4B

// --------------------------------------------------------
//  Data type definitions
// --------------------------------------------------------

#define word                 unsigned short
#define byte                 unsigned char
#define boolean              short
#define integer              short
typedef unsigned char        bit;

// --------------------------------------------------------
//  Internal limits
// --------------------------------------------------------

#define MAXAKAS              16      // Max # of addresses
#define MAXPACKERS           11      // Max # of packer def
#define MAXEXPORT            60      // Max export defs
#define MAXVIA               40      // max nodes packed via
#define MAXEXCEPT            10      // max EXCEPT nodes
#define MAXPACK              32      // max default pack cmd
#define MAXFWDLINK           10      // max fwd link structs
#define MAXNOIMPT            20      // max # names for IMPORT
#define ZONESDOM             10      // zones per domain entry
#define MAXTAG               51      // max areatag length
#define MAXNAME              37      // max namefield lenght
#define MAXPACKER            50      // max packer length
#define MAXORIGIN            64      // max origin length

// --------------------------------------------------------
//  Log Style Definitions
// --------------------------------------------------------

#define LOG_NORMAL           0       // normal log level
#define LOG_VERBOSE          1       // verbose log level


// --------------------------------------------------------
//  Environment Type  (FroDo or Bink)
// --------------------------------------------------------

#define ENV_FRODO            0       // FrontDoor/D'Bridge
#define ENV_BINK             1       // Binkley

// --------------------------------------------------------
//  File Attach Message Status (used in IMAIL.ND)
// --------------------------------------------------------

#define S_NORMAL             0       // used in IMAIL.ND
#define S_HOLD               1
#define S_CRASH              2

// --------------------------------------------------------
//  Security settings (used in IMAIL.ND)
// --------------------------------------------------------

#define SEC_NONE             0
#define SEC_RELAXED          1
#define SEC_FULL             2

// --------------------------------------------------------
//  Message Base Types
// --------------------------------------------------------

#ifndef MSGTYPE_SDM
  #define MSGTYPE_SDM     0x01
#endif
#ifndef MSGTYPE_SQUISH
  #define MSGTYPE_SQUISH  0x02
#endif
#ifndef MSGTYPE_HUDSON
  #define MSGTYPE_HUDSON  0x03
#endif
#ifndef MSGTYPE_QBBS
  #define MSGTYPE_QBBS    0x03
#endif
#ifndef MSGTYPE_PASSTH
  #define MSGTYPE_PASSTH  0x0F
#endif
#ifndef MSGTYPE_ECHO
  #define MSGTYPE_ECHO    0x80
#endif
#ifndef MSGTYPE_LOCAL
  #define MSGTYPE_LOCAL   0x90
#endif
#ifndef MSGTYPE_NET
  #define MSGTYPE_NET     0xA0
#endif


// --------------------------------------------------------
//  Macros to make life easier
// --------------------------------------------------------

#define BASEMASK          0x0F
#define TYPEMASK          0xF0

#define IsSdm(Type)       ((Type & BASEMASK) == MSGTYPE_SDM)
#define IsMsg(Type)       ((Type & BASEMASK) == MSGTYPE_SDM)
#define IsSquish(Type)    ((Type & BASEMASK) == MSGTYPE_SQUISH)
#define IsHudson(Type)    ((Type & BASEMASK) == MSGTYPE_HUDSON)
#define IsQbbs(Type)      ((Type & BASEMASK) == MSGTYPE_QBBS)
#define IsPassth(Type)    ((Type & BASEMASK) == MSGTYPE_PASSTH)
#define IsEcho(Type)      ((Type & TYPEMASK) == MSGTYPE_ECHO)
#define IsLocal(Type)     ((Type & TYPEMASK) == MSGTYPE_LOCAL)
#define IsNet(Type)       ((Type & TYPEMASK) == MSGTYPE_NET)


// --------------------------------------------------------
//  Exit Errorlevels & Error Messages
// --------------------------------------------------------

#define E_NOERR              0       // no error
#define E_ELOCK              239     // File locking error
#define E_AOPEN              240     // Error opening IMAIL.AX
#define E_BOPEN              241     // Error opening IMAIL.BX
#define E_NOIDX              242     // Index file missing/corrupt
#define E_NOCFG              243     // IMAIL.CF not found
#define E_NOARE              244     // IMAIL.AR not found
#define E_NONOD              245     // IMAIL.ND not found
#define E_BADCF              246     // Error in IMAIL.CF
#define E_BADVR              247     // Bad version of IMAIL.CF
#define E_EOPEN              248     // Error opening file
#define E_EREAD              249     // Error reading file
#define E_EWRIT              250     // Error writing file
#define E_CMDPR              251     // Command Line Parameter error
#define E_FILNF              252     // File not found
#define E_MEMRY              253     // Memory Allocation error
#define E_DISKF              254     // Insufficient disk space
#define E_UNKWN              255     // Unknown error

// === ERRORLEVELs returned by TOSS and/or SCAN ===

#define T_NOMAIL             0x00    // no mail processed
#define T_NET                0x01    // Net mail
#define T_ECHO               0x02    // Echo mail
#define T_BAD                0x04    // Bad and/or dupe mail
#define T_QBBS               0x08    // Qbbs message base changed
#define T_MSG                0x10    // *.MSG message base changed
#define T_SQUISH             0x20    // Squish message base changed
#define T_RSV_IUT            0x40    // reserved


// --------------------------------------------------------
//  Special values for 'ALL'
// --------------------------------------------------------

#define ZONE_ALL             56685u
#define NET_ALL              56685u
#define NODE_ALL             56685u
#define POINT_ALL            56685u

// --------------------------------------------------------
//  Misc other definitions required
// --------------------------------------------------------

#define FA_ANYFILE FA_RDONLY+FA_HIDDEN+FA_SYSTEM+FA_ARCH

#define true           1
#define false          0
#define BLANK          ' '


// --------------------------------------------------------
//  In case your compiler doesn't have these ...
// --------------------------------------------------------

#define _MAXPATH            80
#define _MAXDRIVE           3
#define _MAXDIR             66
#define _MAXFILE            9
#define _MAXEXT             5


// --------------------------------------------------------
//  Structs used in IMAIL Configuration files
// --------------------------------------------------------

struct naddress {                // std node number ...
  word         zone;             //  Zone Number
  word         net;              //  Net Number
  word         node;             //  Node Number
  word         point;            //  Point Number
};

typedef struct expt {
  word         zone;
  word         net;
  word         node;
  word         point;
  boolean      doexpt;
} expt;

struct pack_struct {                        // used in pack routing ...
  struct naddress dst;                      // pack via this node
  struct naddress nodes[MAXVIA];            // nodes to pack
  struct naddress except[MAXEXCEPT];        // exceptions
};

struct eaddress {                           // used in Area Manager ...
  struct naddress dstn;                     // node number
  bit             exp_only:1;               // export only flag
  bit             imp_only:1;               // import only flag
  bit             paused:1;                 // echo mail paused
  bit             rsvd1:5;                  // reserved
};

struct fwd_link {                           // used for forward request nodes ...
  char         areasfile[_MAXFILE+_MAXEXT];   // name of areas file
  char         toprogram[10];               // name of area manager
  char         password[21];                // area manager password
  struct naddress uplink;                   // address of uplink
};

struct dom {
  char         domain[21];                  // name of domain
  char         outbound[_MAXPATH];           // root outbound path
  word         zones[ZONESDOM];             // Zones in this domain
  byte         akas[MAXAKAS];               // =my= AKAs in this domain
};

struct im_stats {                           // Used mainly by IUTIL
  word         th_day;                      // this day
  word         la_day;                      // last day
  word         th_week;                     // this week
  word         la_week;                     // last week
  word         th_month;                    // this month
  word         la_month;                    // last month
  word         th_year;                     // this year
  word         la_year;                     // last year
};


// --------------------------------------------------------
//  IMAIL.CF Structure
// --------------------------------------------------------

typedef struct im_config_type {
  byte         im_ver_maj;                  // Major Version
  byte         im_ver_min;                  // Minor Version
  char         sysop[MAXNAME];              // name of sysop
  byte         badm_type;                   // type of BADM msg base
  byte         dupe_type;                   // type of dupe msg base
  byte         rsv_iut1;                    // reserved
  struct naddress aka[MAXAKAS];             // the AKAs
  struct dom   domains[MAXAKAS];            // domain names & zones
  byte         prod[10];                    // Type2+ product codes
  byte         badmsgs;                     // bad message board
  byte         dupebrd;                     // dupe message board
  byte         rsv_iut2;                    // reserved
  char         netmail[_MAXPATH];            // net mail subdirectory
  char         netfile[_MAXPATH];            // inbound files directory
  char         in_pkt[_MAXPATH];             // Directory for inbound PKTs
  char         out_pkt[_MAXPATH];            // Directory for outbound PKTs
  char         outbound[_MAXPATH];           // outbound directory
  char         quickbbs[_MAXPATH];           // QuickBBS system directory
  char         badm_dir[_MAXPATH];           // Bad Msg directory
  char         dupe_dir[_MAXPATH];           // Dupes directory
  char         rsv_iut3[_MAXPATH];           // reserved
  char         rsc_iut4[_MAXPATH];           // reserved
  char         logfilename[_MAXPATH];        // Log file name
  char         alnk_help[_MAXFILE+_MAXEXT];   // AreaLink help text
  char         new_areas[_MAXFILE+_MAXEXT];   // name of file for new areas
  word         pwd_expire_days;             // days before pwd expr'd
  word         max_pkt_size;                // max size of pkt to create
  word         max_dupes;                   // max dupes kept in dbase
  byte         log_level;                   // logging level
  bit          rtnrecpt:1;                  // True if to send rtn recpt
  bit          checkbad:1;                  // True if to check badmsgs
  bit          ARCmail06:1;                 // ARCmail 0.6 compatibility
  bit          environment:1;               // FroDo or Binkley
  bit          pkt_open:1;                  // Keep PKT files open
  bit          auto_add:1;                  // Auto-add unknown areas
  bit          multi_tasking:1;             // true if multi-tasking
  bit          forward_for:1;               // forward PKTs not for us
  bit          singleextract:1;             // extract 1 bundle at a time
  bit          trunc_sent:1;                // 1 = Trunc 0 = Delete
  bit          keep_recpt:1;                // keep areafix message
  bit          prod_names:1;                // use the FTSC product list
  bit          swap_ems:1;                  // swap to EMS
  bit          swap_ext:1;                  // swap to extended memory
  bit          newarea_passth:1;            // set new areas to passth
  bit          rsv_iut5:1;                  // reserved
  bit          direct_video:1;              // use direct screen writing
  bit          bit_rsvd:7;                  // reserved
  char         newarea_group;               // new areas default group
  char         dflt_origin[MAXORIGIN];      // default origin line
  char         before_toss[_MAXPATH];        // call before proc. PKTs
  long         last_run;                    // last run of IUTIL STAT
  byte         sem_timeout;                 // timeout waiting for IMBUSY
  byte         via_line;                    // add Via Line to netmails
  byte         rsv_iut6;                    // reserved
  char         no_import[MAXNOIMPT][38];    // ignore when importing
  char         packers[MAXPACKERS][MAXPACKER]; // Packer defintions
  char         arcunpak[MAXPACKER];         // cmd to de-arc .ARC files
  char         arjunpak[MAXPACKER];         // cmd to de-arc .ARJ files
  char         pkpakunpak[MAXPACKER];       // cmd to de-arc .PKA files
  char         pkzipunpak[MAXPACKER];       // cmd to de-arc .ZIP files
  char         lharcunpak[MAXPACKER];       // cmd to de-arc .LZH files
  char         zoounpak[MAXPACKER];         // cmd to de-arc .ZOO files
  char         pakunpak[MAXPACKER];         // cmd to de-arc .PAK files
  char         sqzunpak[MAXPACKER];         // cmd to de-arc .SQZ files
  char         unkunpak[MAXPACKER];         // cmd to de-arc unknown files
  struct pack_struct pack[MAXPACK];         // default PACK routing
  struct fwd_link fwd[MAXFWDLINK];          // forward link requests
  char         filler[160];                 // reserved
} im_config_type;


// --------------------------------------------------------
//  IMAIL.AR Structure
// --------------------------------------------------------

typedef struct areas_record_type {
  char         aname[MAXTAG];        // area name
  char         comment[61];          // area comment
  char         origin[MAXORIGIN];    // origin line to use
  char         group;                // area group
  char         o_addr;               // address for origin
  char         use_akas[MAXAKAS];    // addresses for seen-bys
  byte         msg_base_type;        // message base type
  byte         brd;                  // board number
  char         msg_path[_MAXPATH];    // MSG directory
  bit          active:1;             // flag area active
  bit          zone_gate:1;          // Zone-gate stripping
  bit          tiny_seen:1;          // tiny seen-by flag
  bit          secure:1;             // secure flag
  bit          keep_seen:1;          // keep seen-by flag
  bit          deleted:1;            // flag deleted area
  bit          auto_added:1;         // flag auto-added record
  bit          rsv_iut1:1;           // reserved
  bit          rsv_iut2:1;           // reserved
  bit          rsvd:7;               // reserved
  byte         iutil_bits;           // bits for IUTIL
  byte         user_bits;            // 8 user-available bits
  byte         days;                 // days to keep messages
  word         msgs;                 // num messages to keep
  struct im_stats t_stats;           // TOSS stats
  struct im_stats s_stats;           // SCAN stats
  struct im_stats d_stats;           // dupe statistics
  time_t       creation;             // date/time of creation
  time_t       update;               // last update by STAT
  time_t       marked;               // used by IUTIL
  char         filler[10];
  struct eaddress export[MAXEXPORT]; // export list
} areas_record_type;

    /* --- Notes ---------------------------------------

    1) The entries in 'o_addr' and 'use_akas' are indexes into
       the list of AKAs in IMAIL.CF, minus 1. eg:
          im_config.aka[o_addr-1]
       A value of 0 means 'no address'
    2) the 'user_bits' entry allows third-part software to store
       extra information in IMAIL.AR. Their meaning is program-
       specific, so be careful when making use of them!

    ------------------------------------------------- */


// --------------------------------------------------------
//  IMAIL.AX & IMAIL.BX Structures
// --------------------------------------------------------

    /* --- Notes ---------------------------------------

    1) These are "true" index files, created by BTree routines;
       they cannot be used unless you have the BTree source
       code as implemented in IMAIL.

    ------------------------------------------------- */

// --------------------------------------------------------
//  IMAIL.ND Structure
// --------------------------------------------------------

typedef struct node_record_type {
  struct naddress dstn;              // Node to pack for
  char         sysop[MAXNAME];       // name of sysop
  char         domain;               // index to domain
  char         pwd[21];              // AreaLink password
  char         status;               // Status of file attach msg
  char         program;              // packer to use ("packers")
  char         groups[27];           // groups node can request
  word         capability;           // capability word for node
  bit          allow_remote:1;       // allow remote maint.
  bit          direct:1;             // Mark ARCmail as direct
  bit          auto_cap:1;           // true = auto-detect capability
  bit          auto_added:1;         // record added by IMAIL
  bit          newarea_add:1;        // auto add to new areas
  bit          newarea_create:1;     // create new areas?
  bit          rescan_ok:1;          // allow node to rescan
  bit          notify:1;             // Send notify messages?
  bit          security:2;           // Define security handling
  bit          rsvd:6;               // reserved bits
  word         user_bits;            // 16 user-available bits
  char         newarea_group;        // new areas default group
  char         pkt_pwd[9];           // PKT password
  long         last_pwd_change;      // time ALNK pwd last changed
  char         filler[20];
  struct node_record_type *next_node;
  struct node_record_type *prev_node;
} node_record_type;


    /* --- Notes ---------------------------------------

    1) The last 2 fields are not saved to IMAIL.ND
    2) the 'user_bits' entry allows third-part software to store
       extra information in IMAIL.ND. Their meaning is program-
       specific, so be careful when making use of them!

    ------------------------------------------------- */


// --------------------------------------------------------
//  IMAIL.DP Structure
// --------------------------------------------------------

    /* --- Notes ---------------------------------------

    1) The dupe database consists simply of a series of dword-sized values,
       which are a 32-bit CRC of (in order) the following data:
          Addressee's name
          Sender's name
          Date
          Subject
          Address in MSGID (if any)

    ------------------------------------------------- */


