
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gmohuds1.cpp,v 1.1.1.1 1998/11/21 19:38:12 odinn Exp $
//  ------------------------------------------------------------------
//  Hudson / Goldbase msgbase handling
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------

#include <stdlib.h>

#include <gdbgerr.h>
#include <gmemdbg.h>
#include <gdbgtrk.h>
#include <gstrall.h>

#include <gmoprot.h>
#include <gmohuds.h>


//  ------------------------------------------------------------------

HudsWide* __FAR hudsonwide = NULL;

#define __HUDSON


//  ------------------------------------------------------------------

void HudsArea::data_open() {

  wide = hudsonwide;
}


//  ------------------------------------------------------------------

void HudsArea::data_close() {

}


//  ------------------------------------------------------------------

void HudsWide::exit() {

  delete hudsonwide->user;
  throw_release(hudsonwide->msgidxptr);
}


//  ------------------------------------------------------------------

void HudsExit() {

  if(hudsonwide)
    hudsonwide->exit();
  throw_release(hudsonwide);
}


//  ------------------------------------------------------------------

void HudsInit(const char* path, const char* syspath, int userno, long sizewarn, int ra2usersbbs) {

  // Initialize msgbase-wide data
  hudsonwide = (HudsWide*)throw_calloc(1, sizeof(HudsWide));
  hudsonwide->path = path;
  hudsonwide->syspath = syspath;
  hudsonwide->userno = userno;

  #ifdef __HUDSON
  hudsonwide->sizewarn = sizewarn;
  hudsonwide->ra2usersbbs = ra2usersbbs;
  #else
  NW(sizewarn);
  NW(ra2usersbbs);
  #endif

  hudsonwide->init();
}


//  ------------------------------------------------------------------

void HudsWide::init() {

  GFTRK("HudsInit");

  if(WideDebug)
    WideLog->printf("- Begin init for %s.", path);

  fhusr = -1;
  fhtxt = fhhdr = fhidx = -1;
  fhinf = fhlrd = fhtoi = -1;
  isopen = 0;
  islocked = FALSE;
  timesposted = 0;
  msgidxsize = 0;
  msgidxptr = NULL;
  pmscan = NULL;
  scn = NULL;

  // Open complete msgbase, create if none exists
  if(NOT fexist(AddPath(path, "msghdr" HUDS_EXT))) {
    WideLog->printf("* Creating new msgbase at %s", path);
    raw_open(O_CREAT);
    if(filelength(fhinf) == 0) {
      memset(&msginfo, 0, sizeof(HudsInfo));
      write(fhinf, &msginfo, sizeof(HudsInfo));
    }
    if(filelength(fhlrd) == 0) {
      memset(lastrec, 0, sizeof(HudsLast));
      write(fhlrd, lastrec, sizeof(HudsLast));
    }
  }
  else {
    raw_open();
  }

  // Lock msgbase to ensure data integrity
  lock();

  // Check if MSGTXT.BBS is approaching dangerous size
  #ifdef __HUDSON
  if(filelength(fhtxt) > sizewarn)
    HudsSizewarn();
  #endif

  // Check for mismatch between the header and the index files
  long _hdrsize = filelength(fhhdr)/(long)sizeof(HudsHdr);
  long _idxsize = filelength(fhidx)/(long)sizeof(HudsIdx);
  long _toisize = filelength(fhtoi)/(long)sizeof(HudsToIdx);
  if((_hdrsize != _idxsize) OR (_hdrsize != _toisize)) {
    raw_close();
    HGWarnRebuild();
    WideLog->ErrIndex();
    WideLog->printf("! The " HUDS_NAME " msgbase files do not have the same number of records.");
    WideLog->printf(": %smsghdr" HUDS_EXT "   (%lu records).", path, _hdrsize);
    WideLog->printf(": %smsgidx" HUDS_EXT "   (%lu records).", path, _idxsize);
    WideLog->printf(": %smsgtoidx" HUDS_EXT " (%lu records).", path, _toisize);
    WideLog->printf("+ Advice: You need to run a msgbase index rebuild utility.");
    IndexErrorExit();
  }

  // Detect USERS.BBS format unless user has configured it
  #ifdef __HUDSON
  if(ra2usersbbs == 0) {

    // Get size of USERS.BBS
    long len = filelength(fhusr);

    // Does size match Hudson format?
    int hudsmatch = (len % sizeof(HudsUsers)) == 0;
    if(hudsmatch)
      ra2usersbbs = 1;

    // Does size match RA2 format?
    int ra2match = (len % sizeof(RA2Users)) == 0;
    if(ra2match)
      ra2usersbbs = 2;

    // If it matches both of them
    if(hudsmatch AND ra2match) {

      // Check version in CONFIG.RA to make sure
      Path rapath, file;
      char* ptr = getenv("RA");
      if(ptr)
        AddBackslash(strcpy(rapath, ptr));
      MakePathname(file, rapath, "messages.ra");
      int fh = ::sopen(file, O_RDONLY|O_BINARY, WideSharemode, S_STDRD);
      if(fh != -1) {
        word VersionID = 0;
        read(fh, &VersionID, sizeof(word));
        if(VersionID >= 0x200)
          ra2usersbbs = 2;
        else
          ra2usersbbs = 1;
        ::close(fh);
      }
    }

    // If it does not match either of them
    if(NOT hudsmatch AND NOT ra2match) {
      WideLog->ErrIndex();
      WideLog->printf("! The users.bbs file has an incorrect size.");
      WideLog->printf(": %susers.bbs, %lu bytes.", path, len);
      WideLog->printf(": Should be %lu bytes if it's in RA2 format.", (len/(long)sizeof(RA2Users))*(long)sizeof(RA2Users));
      WideLog->printf(": Should be %lu bytes if it's in Hudson format.", (len/(long)sizeof(HudsUsers))*(long)sizeof(HudsUsers));
      if(ra2usersbbs)
        WideLog->printf(": It appears to be in %s format.", (ra2usersbbs == 2) ? "RA2" : "Hudson");
      WideLog->printf("+ Advice: Run a userbase packing utility.");
      IndexErrorExit();
    }
  }
  #endif

  #ifdef __HUDSON
  fhuix = fhuxi = -1;
  if(ra2usersbbs == 2) {
    if(WideDebug)
      WideLog->printf("- Using a RA2 format userbase.");
    user = new RA2User;
  }
  else {
    if(WideDebug)
      WideLog->printf("- Using a Hudson format userbase.");
    user = new HudsonUser;
  }
  #else
  user = new GoldbaseUser;
  #endif
  throw_new(user);

  // Open RA2 files
  #ifdef __HUDSON
  if(ra2usersbbs == 2) {
    RA2User* _user2 = (RA2User*)user;
    _user2->idxfh = fhuix = test_open("usersidx.bbs", O_CREAT);
    _user2->xifh = fhuxi = test_open("usersxi.bbs", O_CREAT);
  }
  #endif

  // Find user
  const char* _username = WideUsername[0];
  if(userno == -1) {
    user->fh = fhusr;
    user->find(_username);
    if(NOT user->found) {
      WideLog->printf("* User \"%s\" not found in %susers" HUDS_EXT ".", _username, path);
      user->add(_username);
      WideLog->printf("* Now added with user number %u.", user->index);
    }
    userno = user->index;
  }

  if(WideDebug)
    WideLog->printf("- Using user record number %u.", userno);

  // Close RA2 files
  #ifdef __HUDSON
  if(ra2usersbbs == 2) {
    ::close(fhuix);
    ::close(fhuxi);
  }
  #endif

  // Unlock and close
  unlock();
  raw_close();

  if(WideDebug)
    WideLog->printf("- End init for %s.", path);

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void HudsWide::save_lastread(HudsBoard board, HudsMsgn msgno) {

  GFTRK("HudsSaveLast");

  // Update lastread record
  HudsMsgn _lastread = lastrec[board-1] = msgno;
  lseekset(fhlrd, (long)userno*(long)sizeof(HudsLast));
  write(fhlrd, lastrec, sizeof(HudsLast));

  // Update user record
  user->fh = fhusr;
  user->moveto(userno);
  if(user->lastread() < _lastread)
    user->lastread(_lastread);
  if(timesposted) {
    user->inctimesposted(timesposted);
    timesposted = 0;
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void HudsArea::open() {

  GFTRK("HudsOpen");

  data_open();
  wide->isopen++;
  if(wide->isopen == 1) {
    wide->raw_open();
    wide->refresh();
    wide->timesposted = 0;
  }
  if(wide->isopen > 2) {
    WideLog->ErrTest();
    goto error;
  }

  isopen++;
  if(isopen > 2) {
    WideLog->ErrTest();
    goto error;
  }

  scan();
  goto done;

error:
  WideLog->printf("! Trying to open a " HUDS_NAME " msgbase more than twice.");
  WideLog->printf(": %s, board %u.", echoid(), board());
  WideLog->printf("+ Info: This indicates a serious bug.");
  WideLog->printf("+ Advice: Report to the Author immediately.");
  TestErrorExit();

done:
  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void HudsArea::close() {

  GFTRK("HudsClose");

  if(isopen) {
    if(isopen == 1) {
      wide->save_lastread((HudsBoard)board(), (HudsMsgn)Msgn->CvtReln(lastread));
      Msgn->Reset();
    }
    isopen--;
  }
  else {
    WideLog->ErrTest();
    goto error;
  }

  if(wide->isopen) {
    if(wide->isopen == 1) {
      if(wide->islocked)
        wide->unlock();
      wide->raw_close();
      throw_release(wide->msgidxptr);
    }
    wide->isopen--;
  }
  else {
    WideLog->ErrTest();
    goto error;
  }
  goto done;

error:
  WideLog->printf("! Trying to close an already closed " HUDS_NAME " msgbase.");
  WideLog->printf(": %s, board %u.", echoid(), board());
  WideLog->printf("+ Info: This indicates a potentially serious bug.");
  WideLog->printf("+ Advice: Report to the Author immediately.");
  TestErrorExit();

done:
  data_close();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void HudsArea::suspend() {

  GFTRK("HudsSuspend");

  wide->save_lastread((HudsBoard)board(), (HudsMsgn)Msgn->CvtReln(lastread));
  wide->raw_close();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void HudsArea::resume() {

  GFTRK("HudsResume");

  wide->raw_open();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

