
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gmbarea.h,v 1.1.1.1 1998/11/21 19:38:14 odinn Exp $
//  ------------------------------------------------------------------
//  Area structures and classes.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------

#ifndef __GMBAREA_H
#define __GMBAREA_H


//  ------------------------------------------------------------------

#include <gtag.h>
#include <glog.h>
#include <gvectorp.h>
#include <gedacfg.h>
#include <gmsg.h>


//  ------------------------------------------------------------------
//  Area base class

class gmb_area {

protected:

  int refs;
  int open_refs;
  int lock_refs;
  gareacfg* cfg;

public:

  gmb_area(gareacfg* c)                   { init(c); }
  virtual ~gmb_area();

  void  init(gareacfg* c);

  gtag           msgnset;
  gtag           pmrkset;

  uint           scan_count;
  uint           pmscan_count;

  uint           lastread;
  ulong          lastreadentry;

  int            references() const       { return refs; }
  int            open_references() const  { return open_refs; }
  int            lock_references() const  { return lock_refs; }

  bool           isopen() const           { return open_refs > 0; }
  bool           islocked() const         { return lock_refs > 0; }

  int            areaid() const           { return cfg->areaid(); }
  uint           board() const            { return cfg->board(); }
  uint           type() const             { return cfg->type(); }
  uint           msgbase() const          { return cfg->msgbase(); }
  const gstring& echoid() const           { return cfg->echoid(); }
  const gstring& path() const             { return cfg->path(); }

  bool           isfts1() const           { return cfg->isfts1(); }
  bool           isopus() const           { return cfg->isopus(); }
  bool           isezycom() const         { return cfg->isezycom(); }
  bool           isfido() const           { return cfg->isfido(); }
  bool           isgoldbase() const       { return cfg->isgoldbase(); }
  bool           ishudson() const         { return cfg->ishudson(); }
  bool           isjam() const            { return cfg->isjam(); }
  bool           ispcboard() const        { return cfg->ispcboard(); }
  bool           issquish() const         { return cfg->issquish(); }
  bool           iswildcat() const        { return cfg->iswildcat(); }
  bool           isadeptxbbs() const      { return cfg->isadeptxbbs(); }
  bool           isseparator() const      { return cfg->isseparator(); }

  bool           isnet() const            { return cfg->isnet(); }
  bool           isecho() const           { return cfg->isecho(); }
  bool           islocal() const          { return cfg->islocal(); }
  bool           isemail() const          { return cfg->isemail(); }
  bool           isnewsgroup() const      { return cfg->isnewsgroup(); }
  bool           isinternet() const       { return cfg->isinternet(); }
  bool           isqwk() const            { return cfg->isqwk(); }
  bool           issoup() const           { return cfg->issoup(); }


  void set_areaid(int a)                  { cfg->set_areaid(a); }
  void set_board(uint b)                  { cfg->set_board(b); }
  void set_type(uint t)                   { cfg->set_type(t); }
  void set_msgbase(uint m)                { cfg->set_msgbase(m); }
  void set_echoid(const gstring& s)       { cfg->set_echoid(s); }
  void set_path(const gstring& s)         { cfg->set_path(s); }


  virtual void activate() = 0;
  virtual void deactivate() = 0;

  virtual void open() = 0;
  virtual void close() = 0;

  virtual void save_lastread() = 0;

  virtual void suspend() = 0;
  virtual void resume() = 0;

  virtual void lock() = 0;
  virtual void unlock() = 0;

  virtual void scan() = 0;
  virtual void scan_area() = 0;
  virtual void scan_area_pm() = 0;

  virtual bool load_hdr(gmsg& msg) = 0;
  virtual bool load_msg(gmsg& msg) = 0;

  virtual void save_hdr(int mode, gmsg& msg) = 0;
  virtual void save_msg(int mode, gmsg& msg) = 0;

  virtual void del_msg(gmsg& msg) = 0;

  virtual void new_msgno(gmsg& msg) = 0;
  virtual bool user_lookup(gstring& lookfor) = 0;
  virtual bool renumber() = 0;

  virtual void update_timesread(gmsg& msg) = 0;

  virtual void hexdump_msg(gmsg& msg, gstring& dump) = 0;

  virtual void set_highwater_mark() { }
  virtual void reset_highwater_mark() { }

};


//  ------------------------------------------------------------------

class gmb_area_manager {

protected:

  gvectp<gmb_area> active;

public:

  gmb_area_manager();
  virtual ~gmb_area_manager();

  virtual void initialize() = 0;
  virtual void reset() = 0;

  virtual void wide_open() = 0;
  virtual void wide_close() = 0;

  gmb_area* find(gareacfg* a) const;
  gmb_area* find_inactive() const;

  virtual gmb_area* new_area(gareacfg* c) = 0;
  virtual void init_area(gmb_area* a, gareacfg* c) = 0;

  gmb_area* activate(gareacfg* c);
  void deactivate(gmb_area* a) const { a->deactivate(); }

};


//  ------------------------------------------------------------------

class gmb_adeptxbbs_manager;
class gmb_ezycom_manager;
class gmb_fido_manager;
class gmb_goldbase_manager;
class gmb_hudson_manager;
class gmb_jam_manager;
class gmb_pcboard_manager;
class gmb_squish_manager;
class gmb_wildcat_manager;


//  ------------------------------------------------------------------
//  Personal mail

const int PM_STARTUP  = 0x0001;
const int PM_ALLNAMES = 0x0002;
const int PM_LISTONLY = 0x0004;


//  ------------------------------------------------------------------

const int TIMESREAD_NO   = 0x00;
const int TIMESREAD_YES  = 0x01;
const int TIMESREAD_LOCK = 0x02;


//  ------------------------------------------------------------------

extern glog*        gmb_wide_log;
extern int          gmb_wide_debug;
extern int          gmb_wide_canlock;
extern long         gmb_wide_msgsize;
extern const char** gmb_wide_username;
extern int          gmb_wide_usernames;
extern uint         gmb_wide_sharemode;
extern int          gmb_wide_dispsoftcr;
extern int          gmb_wide_personalmail;
extern int          gmb_wide_timesread;


//  ------------------------------------------------------------------

int PopupLocked(long tries, int isopen, const char* file);


//  ------------------------------------------------------------------

#endif

//  ------------------------------------------------------------------

