
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gmbfido1.cpp,v 1.1.1.1 1998/11/21 19:38:13 odinn Exp $
//  ------------------------------------------------------------------
//  Fido/Opus/FTSC (*.MSG) type handling.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------

#include <stdlib.h>

#include <gdbgerr.h>
#include <gmemdbg.h>
#include <gdbgtrk.h>

#include <gmbfido.h>


//  ------------------------------------------------------------------

gmb_fido_manager::gmb_fido_manager() {

  userno = 0;
  fidohwmarks = false;
  fidonullfix = false;
  fidolastread = "LASTREAD";
}


//  ------------------------------------------------------------------

gmb_fido_manager::~gmb_fido_manager() {

  reset();
}


//  ------------------------------------------------------------------

void gmb_fido_manager::initialize() {

  const char* _username = gmb_wide_username[0];
  if(userno == -1) {
    Path userfile;
    PathCopy(userfile, AddPath(squishuserpath.c_str(), "USER.BBS"));
    user.fh = ::sopen(userfile, O_RDWR|O_CREAT|O_BINARY, gmb_wide_sharemode, S_STDRW);
    if(user.fh != -1) {
      user.find(_username);
      if(not user.found) {
        gmb_wide_log->printf("* User \"%s\" not found in %s.", _username, userfile);
        user.add(_username);
        gmb_wide_log->printf("* Now added with user number %u.", user.index);
      }
      ::close(user.fh);
    }
    userno = user.index;
  }
}


//  ------------------------------------------------------------------

void gmb_fido_manager::reset() {

  for(int n=0; n<active.size(); n++)
    delete active[n];
  active.reset();
}


//  ------------------------------------------------------------------

gmb_area* gmb_fido_manager::new_area(gareacfg* c) {

  gmb_fido_area* area = new gmb_fido_area(c, this);
  throw_new(area);
  return area;
}


//  ------------------------------------------------------------------

void gmb_fido_manager::init_area(gmb_area* a, gareacfg* c) {

  gmb_fido_area* area = (gmb_fido_area*)a;
  area->init(c, this);
}


//  ------------------------------------------------------------------

void gmb_fido_area::activate() {

  refs++;
  if(refs == 1) {
    open_refs = lock_refs = 0;
    highwatermark = 0;
  }
}


//  ------------------------------------------------------------------

void gmb_fido_area::deactivate() {

  if(refs) {
    if(refs == 1) {
      if(open_refs) {
        open_refs = 1;
        close();
      }
    }
    refs--;
  }
}


//  ------------------------------------------------------------------

char* gmb_fido_area::build_msgname(char* __buf, ulong __msgno) {

  sprintf(__buf, "%s%lu.msg", path().c_str(), __msgno);
  return __buf;
}


//  ------------------------------------------------------------------

int gmb_fido_area::test_open(const char* __file, int __openmode, int __sharemode, int __fail) {

  GFTRK("FidoTestOpen");

  int _fh;
  long _tries = 0;

  do {

    _fh = ::sopen(__file, __openmode, __sharemode, S_STDRW);
    if(_fh == -1) {

      // Store error values from sopen()
      int _xerrno = errno;
      int _x_doserrno = _doserrno;

      // Tell the world
      if((errno != EACCES) or (PopupLocked(++_tries, false, __file) == false)) {

        // Return instead of halting if requested
        if(errno != EACCES and not __fail) {
          GFTRK(NULL);
          return _fh;
        }

        // User requested to exit
        errno = _xerrno;
        _doserrno = _x_doserrno;
        gmb_wide_log->ErrOpen();
        gmb_wide_log->printf("! A Fido message file could not be opened.");
        gmb_wide_log->printf(": %s", __file);
        gmb_wide_log->ErrOSInfo();
        OpenErrorExit();
      }
    }
  } while(_fh == -1);

  // Remove the popup window
  if(_tries)
    PopupLocked(0, 0, NULL);

  GFTRK(NULL);

  return _fh;
}


//  ------------------------------------------------------------------

void gmb_fido_area::open() {

  GFTRK("FidoOpen");

  open_refs++;
  if(open_refs == 1) {

    // No need to do anything here
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_fido_area::close() {

  GFTRK("FidoClose");

  if(open_refs) {
    if(open_refs == 1) {
      msgnset.reset();
      pmrkset.reset();
    }
    open_refs--;
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_fido_area::save_lastread() {

  GFTRK("FidoSaveLastread");

  int fh = ::sopen(AddPath(path().c_str(), wide->fidolastread.c_str()), O_RDWR|O_CREAT|O_BINARY, gmb_wide_sharemode, S_STDRW);
  if(fh != -1) {
    word _lastread = (word)msgnset.relative_to_absolute(lastread);
    lseekset(fh, wide->userno, sizeof(word));
    write(fh, &_lastread, sizeof(word));
    ::close(fh);
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_fido_area::suspend() {

  GFTRK("FidoSuspend");

  save_lastread();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_fido_area::resume() {

  GFTRK("FidoResume");

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

