
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gmbfido3.cpp,v 1.1.1.1 1998/11/21 19:38:13 odinn Exp $
//  ------------------------------------------------------------------
//  Fido/Opus/FTSC (*.MSG) type handling.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------

#include <sys/types.h>
#include <sys/stat.h>

#include <gmemdbg.h>
#include <gdbgtrk.h>
#include <gfilutil.h>
#include <gstrall.h>

#include <gmbfido.h>


//  ------------------------------------------------------------------

bool gmb_fido_area::load_message(int mode, gmsg& msg, FidoHdr& hdr) {

  // Build message filename
  Path _msgfile;
  build_msgname(_msgfile, msg.msgno);

  // Open the message file
  int _fh = test_open(_msgfile, O_RDONLY|O_BINARY, gmb_wide_sharemode);
  if(_fh == -1) {
    GFTRK(NULL);
    return false;
  }

  // Get file info
  struct stat st;
  fstat(_fh, &st);

  // Read header
  memset(&hdr, 0, sizeof(FidoHdr));
  read(_fh, &hdr, sizeof(FidoHdr));

  msg.link.to_set(hdr.replyto);
  msg.link.first_set(hdr.reply1st);

  msg.cost      = hdr.cost;
  msg.timesread = hdr.timesread;

  msg.origname.copyn(hdr.by, sizeof(hdr.by));
  msg.destname.copyn(hdr.to, sizeof(hdr.to));
  msg.subject.copyn(hdr.re, sizeof(hdr.re));

  msg.origaddr.zone  = hdr.ftsc.origzone;
  msg.origaddr.net   = hdr.orignet;
  msg.origaddr.node  = hdr.orignode;
  msg.origaddr.point = hdr.ftsc.origpoint;

  if(isnet()) {
    msg.destaddr.zone  = hdr.ftsc.destzone;
    msg.destaddr.net   = hdr.destnet;
    msg.destaddr.node  = hdr.destnode;
    msg.destaddr.point = hdr.ftsc.destpoint;
  }

  if(isopus()) {
    msg.origaddr.zone  =
    msg.destaddr.zone  =
    msg.origaddr.point =
    msg.destaddr.point = 0;
  }

  // Convert datetime
  if(isopus()) {
    msg.date.written = hdr.opus.written;
    msg.date.arrived = hdr.opus.arrived;
  }
  else {
    msg.date.arrived = 0;
  }
  msg.date.written.convert_fts1(hdr.datetime);

  // Transfer attributes
  msg.attr.pvt(hdr.attr & FIDO_PVT);
  msg.attr.cra(hdr.attr & FIDO_CRASH);
  msg.attr.rcv(hdr.attr & FIDO_RECEIVED);
  msg.attr.snt(hdr.attr & FIDO_SENT);
  msg.attr.att(hdr.attr & FIDO_ATTACH);
  msg.attr.trs(hdr.attr & FIDO_TRANSIT);
  msg.attr.orp(hdr.attr & FIDO_ORPHAN);
  msg.attr.k_s(hdr.attr & FIDO_KILLSENT);
  msg.attr.loc(hdr.attr & FIDO_LOCAL);
  msg.attr.hld(hdr.attr & FIDO_HOLD);
  msg.attr.rsv(hdr.attr & FIDO_RESERVED);
  msg.attr.frq(hdr.attr & FIDO_FREQ);
  msg.attr.rrq(hdr.attr & FIDO_RETRECREQ);
  msg.attr.rrc(hdr.attr & FIDO_RETREC);
  msg.attr.arq(hdr.attr & FIDO_AUDITREQ);
  msg.attr.urq(hdr.attr & FIDO_UPDREQ);
  msg.attr.lok(not (st.st_mode & S_IWUSR));

  // Set the unsent attribute
  if(isnet())
  if(isnet() or (isecho() and not wide->fidohwmarks))
    msg.attr.uns((hdr.attr & FIDO_LOCAL) and not (hdr.attr & FIDO_SENT));
  else if(isecho() and wide->fidohwmarks) {
    msg.attr.uns(highwatermark <  msg.msgno);
    msg.attr.snt(highwatermark >= msg.msgno);
  }
  else
    msg.attr.uns0();

  // If message text is used
  if(mode & GMSG_TXT) {

    // Get length of message text and adjust if necessary
    long filelen = filelength(_fh);
    long textlen = filelen >= sizeof(FidoHdr) ? filelen - sizeof(FidoHdr) : 0;
    if(textlen > gmb_wide_msgsize)
      textlen = gmb_wide_msgsize;

    // Allocate space for the message text
    msg.text.reserve(textlen);

    // Read the message text
    read(_fh, msg.text.pointer(), (uint)textlen);

    // Fix msgs handled by brain-dead programs
    if(*msg.text.pointer() == NUL)
      *msg.text.pointer() = LF;

    // Fix complete msg
    if(wide->fidonullfix) {
      msg.text.strip_nul = true;
      msg.text.nul_terminates = false;
    }

    // Convert to lines
    msg.text.raw_to_lines(textlen);
  }

  // Close the message file
  ::close(_fh);

  GFTRK(NULL);

  // Success
  return true;
}


//  ------------------------------------------------------------------

bool gmb_fido_area::load_hdr(gmsg& msg) {

  GFTRK("FidoLoadHdr");

  FidoHdr _hdr;
  return load_message(GMSG_HDR, msg, _hdr);
}


//  ------------------------------------------------------------------

bool gmb_fido_area::load_msg(gmsg& msg) {

  GFTRK("FidoLoadMsg");

  FidoHdr _hdr;
  return load_message(GMSG_HDRTXT, msg, _hdr);
}


//  ------------------------------------------------------------------

