
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gmbfido4.cpp,v 1.1.1.1 1998/11/21 19:38:13 odinn Exp $
//  ------------------------------------------------------------------
//  Fido/Opus/FTSC (*.MSG) type handling.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------

#include <sys/types.h>
#include <sys/stat.h>
#if defined(__WATCOMC__)
#include <sys/utime.h>
#else
#include <utime.h>
#endif

#include <gdbgtrk.h>
#include <gstrall.h>

#include <gmbfido.h>


//  ------------------------------------------------------------------

void gmb_fido_area::lock() {

  GFTRK("FidoLock");

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_fido_area::unlock() {

  GFTRK("FidoUnlock");

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_fido_area::save_message(int mode, gmsg& msg, FidoHdr& hdr) {

  // Build message filename
  Path msgfile;
  if(mode & GMSG_NEW) {

    // Get a new message number
    uint cnt = 0;
    do {
      msg.msgno = msgnset.relative_to_absolute(msgnset.size()) + (++cnt);
    } while(fexist(build_msgname(msgfile, msg.msgno)));

    // Never create a new 1.MSG in echo or local areas.
    if(msg.msgno == 1) {
      if(not isnet()) {
        msg.msgno = 2;
        build_msgname(msgfile, msg.msgno);
      }
    }

    // Avoid collision with an existing msg
    while(fexist(msgfile)) {
      build_msgname(msgfile, ++msg.msgno);
      // here we should signal that a rescan is needed
    }
  }
  else {
    build_msgname(msgfile, msg.msgno);
  }

  // Delete message, if so ordered
  if(mode & GMSG_DELETE) {
    if(ForceDeleteFile(msgfile)) {
      chmod(msgfile, S_STDRW);
      ForceDeleteFile(msgfile);
    }
    GFTRK(NULL);
    return;
  }

  // Get message file attributes
  struct stat st;
  if(stat(msgfile, &st) == 0) {
    bool readonly = not (st.st_mode & S_IWUSR);

    // If the file is read-only, make it read-write-able
    if(readonly)
      chmod(msgfile, S_STDRW);
  }

  // Determine file open mode
  int omode = O_RDWR|O_CREAT|O_BINARY;
  if(mode & GMSG_TXT) {

    // Add truncate, in case a changed msg is shorter
    omode |= O_TRUNC;
  }

  // Open the message file
  int fh = test_open(msgfile, omode, gmb_wide_sharemode, true);

  // Get date/time of message file
  if(not (mode & GMSG_NEW)) {
    if(not msg.attr.upd()) {
      fstat(fh, &st);
    }
  }

  memset(&hdr, 0, sizeof(FidoHdr));

  // Transfer attributes
  hdr.attr |= (word)(msg.attr.pvt() ? FIDO_PVT       : 0);
  hdr.attr |= (word)(msg.attr.cra() ? FIDO_CRASH     : 0);
  hdr.attr |= (word)(msg.attr.rcv() ? FIDO_RECEIVED  : 0);
  hdr.attr |= (word)(msg.attr.snt() ? FIDO_SENT      : 0);
  hdr.attr |= (word)(msg.attr.att() ? FIDO_ATTACH    : 0);
  hdr.attr |= (word)(msg.attr.trs() ? FIDO_TRANSIT   : 0);
  hdr.attr |= (word)(msg.attr.orp() ? FIDO_ORPHAN    : 0);
  hdr.attr |= (word)(msg.attr.k_s() ? FIDO_KILLSENT  : 0);
  hdr.attr |= (word)(msg.attr.loc() ? FIDO_LOCAL     : 0);
  hdr.attr |= (word)(msg.attr.hld() ? FIDO_HOLD      : 0);
  hdr.attr |= (word)(msg.attr.rsv() ? FIDO_RESERVED  : 0);
  hdr.attr |= (word)(msg.attr.frq() ? FIDO_FREQ      : 0);
  hdr.attr |= (word)(msg.attr.rrq() ? FIDO_RETRECREQ : 0);
  hdr.attr |= (word)(msg.attr.rrc() ? FIDO_RETREC    : 0);
  hdr.attr |= (word)(msg.attr.arq() ? FIDO_AUDITREQ  : 0);
  hdr.attr |= (word)(msg.attr.urq() ? FIDO_UPDREQ    : 0);

  strxcpy(hdr.to, msg.destname.c_str(), sizeof(hdr.to));
  strxcpy(hdr.by, msg.origname.c_str(), sizeof(hdr.by));
  strxcpy(hdr.re, msg.subject.c_str(), sizeof(hdr.re));

  hdr.ftsc.origzone  = msg.origaddr.zone;
  hdr.orignet        = msg.origaddr.net;
  hdr.orignode       = msg.origaddr.node;
  hdr.ftsc.origpoint = msg.origaddr.point;

  hdr.ftsc.destzone  = msg.destaddr.zone;
  hdr.destnet        = msg.destaddr.net;
  hdr.destnode       = msg.destaddr.node;
  hdr.ftsc.destpoint = msg.destaddr.point;

  hdr.replyto   = (word)msg.link.to();
  hdr.reply1st  = (word)msg.link.first();

  hdr.cost      = (word)msg.cost;
  hdr.timesread = (word)msg.timesread;

  if(isopus()) {
    hdr.opus.written = msg.date.written.opustime();
    hdr.opus.arrived = msg.date.arrived.opustime();
  }
  strcpy(hdr.datetime, msg.date.written.fts1());

  // Write message header
  write(fh, &hdr, sizeof(FidoHdr));

  // If message text is used
  if(mode & GMSG_TXT) {

    // Write the message text
    char* txt = msg.text.lines_to_raw();
    write(fh, txt, strlen(txt)+1);
  }

  // Close the message file
  ::close(fh);

  // Reset date/time of message file
  if(not (mode & GMSG_NEW)) {
    if(not msg.attr.upd()) {
      struct utimbuf t;
      t.actime = st.st_atime;
      t.modtime = st.st_mtime;
      utime(msgfile, &t);
    }
  }

  // If the message has locked status, make it read-only
  if(msg.attr.lok())
    chmod(msgfile, S_STDRD);

  // Update internal array if the msg is new
  if(mode & GMSG_NEW) {

    // Add the msgno to the index
    msgnset.append(msg.msgno);
    scan_count++;
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_fido_area::save_hdr(int mode, gmsg& msg) {

  GFTRK("FidoSaveHdr");

  FidoHdr hdr;
  save_message(mode|GMSG_HDR, msg, hdr);
}


//  ------------------------------------------------------------------

void gmb_fido_area::save_msg(int mode, gmsg& msg) {

  GFTRK("FidoSaveMsg");

  FidoHdr hdr;
  save_message(mode|GMSG_HDRTXT, msg, hdr);
}


//  ------------------------------------------------------------------

void gmb_fido_area::del_msg(gmsg& msg) {

  GFTRK("FidoDelMsg");

  FidoHdr hdr;
  save_message(GMSG_HDR | GMSG_DELETE, msg, hdr);
}


//  ------------------------------------------------------------------

void gmb_fido_area::new_msgno(gmsg& msg) {

  GFTRK("FidoNewMsgno");

  uint n = 0;
  Path buf;
  do {
    msg.msgno = msgnset.relative_to_absolute(msgnset.size()) + (++n);
  } while(fexist(build_msgname(buf, msg.msgno)));

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_fido_area::update_timesread(gmsg& msg) {

  GFTRK("gmb_fido_area::update_timesread");

  Path filename;
  build_msgname(filename, msg.msgno);

  int fh = test_open(filename, O_RDWR|O_BINARY, gmb_wide_sharemode);
  if(fh != -1) {
    FidoHdr hdr;
    ::read(fh, &hdr, sizeof(FidoHdr));
    hdr.timesread = (word)msg.timesread;
    ::lseekset(fh, 0);
    ::write(fh, &hdr, sizeof(FidoHdr));
    ::close(fh);
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

