
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gmbhuds1.cpp,v 1.1.1.1 1998/11/21 19:38:13 odinn Exp $
//  ------------------------------------------------------------------
//  Hudson / Goldbase msgbase handling.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------

#include <stdlib.h>

#include <gdbgerr.h>
#include <gmemdbg.h>
#include <gdbgtrk.h>
#include <gstrall.h>

#include <gmbprot.h>
#include <gmbhuds.h>


//  ------------------------------------------------------------------

#define __HUDSON


//  ------------------------------------------------------------------

gmb_hudson_manager::gmb_hudson_manager() {

}


//  ------------------------------------------------------------------

gmb_hudson_manager::~gmb_hudson_manager() {

  reset();
}


//  ------------------------------------------------------------------

gmb_area* gmb_hudson_manager::new_area(gareacfg* c) {

  gmb_hudson_area* area = new gmb_hudson_area(c, this);
  throw_new(area);
  return area;
}


//  ------------------------------------------------------------------

void gmb_hudson_manager::init_area(gmb_area* a, gareacfg* c) {

  gmb_hudson_area* area = (gmb_hudson_area*)a;
  area->init(c, this);
}


//  ------------------------------------------------------------------

void gmb_hudson_manager::initialize() {

  wide.init();
}


//  ------------------------------------------------------------------

void gmb_hudson_manager::reset() {

  wide.reset();
}


//  ------------------------------------------------------------------

void gmb_hudson_widearea::init() {

  GFTRK("HudsInit");

  if(gmb_wide_debug)
    gmb_wide_log->printf("- Begin init for %s.", path.c_str());

  fhusr = -1;
  fhtxt = fhhdr = fhidx = -1;
  fhinf = fhlrd = fhtoi = -1;
  timesposted = 0;
  msgidxsize = 0;
  msgidxptr = NULL;
  pmscantotal = 0;
  pmscan = NULL;
  scn = NULL;
  just_scanning = false;

  // Open complete msgbase, create if none exists
  if(not fexist(AddPath(path.c_str(), "MSGHDR" HUDS_EXT))) {
    gmb_wide_log->printf("* Creating new msgbase at %s", path.c_str());
    raw_open(O_CREAT);
    if(filelength(fhinf) == 0) {
      memset(&msginfo, 0, sizeof(HudsInfo));
      write(fhinf, &msginfo, sizeof(HudsInfo));
    }
    if(filelength(fhlrd) == 0) {
      memset(lastrec, 0, sizeof(HudsLast));
      write(fhlrd, lastrec, sizeof(HudsLast));
    }
  }
  else {
    raw_open();
  }

  // Lock msgbase to ensure data integrity
  lock();

  // Check if MSGTXT.BBS is approaching dangerous size
  #ifdef __HUDSON
  if(filelength(fhtxt) > sizewarn)
    HudsSizewarn();
  #endif

  // Check for mismatch between the header and the index files
  long _hdrsize = filelength(fhhdr)/(long)sizeof(HudsHdr);
  long _idxsize = filelength(fhidx)/(long)sizeof(HudsIdx);
  long _toisize = filelength(fhtoi)/(long)sizeof(HudsToIdx);
  if((_hdrsize != _idxsize) or (_hdrsize != _toisize)) {
    raw_close();
    HGWarnRebuild();
    gmb_wide_log->ErrIndex();
    gmb_wide_log->printf("! The " HUDS_NAME " msgbase files do not have the same number of records.");
    gmb_wide_log->printf(": %sMSGHDR" HUDS_EXT "   (%lu records).", path.c_str(), _hdrsize);
    gmb_wide_log->printf(": %sMSGIDX" HUDS_EXT "   (%lu records).", path.c_str(), _idxsize);
    gmb_wide_log->printf(": %sMSGTOIDX" HUDS_EXT " (%lu records).", path.c_str(), _toisize);
    gmb_wide_log->printf("+ Advice: You need to run a msgbase index rebuild utility.");
    IndexErrorExit();
  }

  // Detect USERS.BBS format unless user has configured it
  #ifdef __HUDSON
  if(ra2usersbbs == 0) {

    // Get size of USERS.BBS
    long len = filelength(fhusr);

    // Does size match Hudson format?
    int hudsmatch = (len % sizeof(HudsUsers)) == 0;
    if(hudsmatch)
      ra2usersbbs = 1;

    // Does size match RA2 format?
    int ra2match = (len % sizeof(RA2Users)) == 0;
    if(ra2match)
      ra2usersbbs = 2;

    // If it matches both of them
    if(hudsmatch and ra2match) {

      // Check version in CONFIG.RA to make sure
      Path rapath, file;
      char* ptr = getenv("RA");
      if(ptr)
        AddBackslash(strcpy(rapath, ptr));
      MakePathname(file, rapath, "MESSAGES.RA");
      int fh = ::sopen(file, O_RDONLY|O_BINARY, gmb_wide_sharemode, S_STDRD);
      if(fh != -1) {
        word VersionID = 0;
        read(fh, &VersionID, sizeof(word));
        if(VersionID >= 0x200)
          ra2usersbbs = 2;
        else
          ra2usersbbs = 1;
        ::close(fh);
      }
    }

    // If it does not match either of them
    if(not hudsmatch and not ra2match) {
      gmb_wide_log->ErrIndex();
      gmb_wide_log->printf("! The USERS.BBS file has an incorrect size.");
      gmb_wide_log->printf(": %sUSERS.BBS, %lu bytes.", path.c_str(), len);
      gmb_wide_log->printf(": Should be %lu bytes if it's in RA2 format.", (len/(long)sizeof(RA2Users))*(long)sizeof(RA2Users));
      gmb_wide_log->printf(": Should be %lu bytes if it's in Hudson format.", (len/(long)sizeof(HudsUsers))*(long)sizeof(HudsUsers));
      if(ra2usersbbs)
        gmb_wide_log->printf(": It appears to be in %s format.", (ra2usersbbs == 2) ? "RA2" : "Hudson");
      gmb_wide_log->printf("+ Advice: Run a userbase packing utility.");
      IndexErrorExit();
    }
  }
  #endif

  #ifdef __HUDSON
  fhuix = fhuxi = -1;
  if(ra2usersbbs == 2) {
    if(gmb_wide_debug)
      gmb_wide_log->printf("- Using a RA2 format userbase.");
    user = &ra2user;
  }
  else {
    if(gmb_wide_debug)
      gmb_wide_log->printf("- Using a Hudson format userbase.");
    user = &hudsonuser;
  }
  #else
  user = &goldbaseuser;
  #endif

  // Open RA2 files
  #ifdef __HUDSON
  if(ra2usersbbs == 2) {
    RA2User* _user2 = (RA2User*)user;
    _user2->idxfh = fhuix = test_open("USERSIDX.BBS", O_CREAT);
    _user2->xifh = fhuxi = test_open("USERSXI.BBS", O_CREAT);
  }
  #endif

  // Find user
  const char* _username = gmb_wide_username[0];
  if(userno == -1) {
    user->fh = fhusr;
    user->find(_username);
    if(not user->found) {
      gmb_wide_log->printf("* User \"%s\" not found in %sUSERS" HUDS_EXT ".", _username, path.c_str());
      user->add(_username);
      gmb_wide_log->printf("* Now added with user number %u.", user->index);
    }
    userno = user->index;
  }

  if(gmb_wide_debug)
    gmb_wide_log->printf("- Using user record number %u.", userno);

  // Close RA2 files
  #ifdef __HUDSON
  if(ra2usersbbs == 2) {
    ::close(fhuix);
    ::close(fhuxi);
  }
  #endif

  // Unlock and close
  unlock();
  raw_close();

  if(gmb_wide_debug)
    gmb_wide_log->printf("- End init for %s.", path.c_str());

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_hudson_widearea::reset() {

  throw_release(msgidxptr);
}


//  ------------------------------------------------------------------

void gmb_hudson_widearea::save_lastread(HudsBoard board, HudsMsgn msgno) {

  GFTRK("HudsSaveLast");

  // Update lastread record
  HudsMsgn _lastread = lastrec[board-1] = msgno;
  lseekset(fhlrd, (long)userno*(long)sizeof(HudsLast));
  write(fhlrd, lastrec, sizeof(HudsLast));

  // Update user record
  user->fh = fhusr;
  user->moveto(userno);
  if(user->lastread() < _lastread)
    user->lastread(_lastread);
  if(timesposted) {
    user->inctimesposted(timesposted);
    timesposted = 0;
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_hudson_widearea::activate() {

  refs++;
  if(refs == 1) {

  }
}


//  ------------------------------------------------------------------

void gmb_hudson_widearea::deactivate() {

  if(refs) {
    if(refs == 1) {
      if(open_refs) {
        open_refs = 1;
        close();
      }
    }
    refs--;
  }
}


//  ------------------------------------------------------------------

void gmb_hudson_area::activate() {

  refs++;
  if(refs == 1) {
    wide->activate();
  }
}


//  ------------------------------------------------------------------

void gmb_hudson_area::deactivate() {

  if(refs) {
    if(refs == 1) {
      if(open_refs) {
        open_refs = 1;
        close();
        wide->deactivate();
      }
    }
    refs--;
  }
}


//  ------------------------------------------------------------------

void gmb_hudson_area::open() {

  GFTRK("gmb_hudson_area::open");

  open_refs++;
  if(open_refs)
    wide->open();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_hudson_area::close() {

  GFTRK("gmb_hudson_area::close");

  if(open_refs) {
    if(open_refs == 1) {
      if(lock_refs) {
        lock_refs = 1;
        unlock();
      }
      wide->close();
      msgnset.reset();
      pmrkset.reset();
    }
    open_refs--;
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_hudson_area::save_lastread() {

  wide->save_lastread((HudsBoard)board(), (HudsMsgn)msgnset.relative_to_absolute(lastread));
}


//  ------------------------------------------------------------------

void gmb_hudson_area::suspend() {

  GFTRK("HudsSuspend");

  save_lastread();
  wide->raw_close();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_hudson_area::resume() {

  GFTRK("HudsResume");

  wide->raw_open();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

