
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gmbjamm4.cpp,v 1.1.1.1 1998/11/21 19:38:13 odinn Exp $
//  ------------------------------------------------------------------
//  JAM msgbase implementation, save/delete.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------

#include <gmemdbg.h>
#include <gdbgtrk.h>
#include <gstrall.h>
#include <gtxtpara.h>
#include <gutlmisc.h>

#include <gmbjamm.h>


//  ------------------------------------------------------------------

void gmb_jam_area::raw_lock() {

  if(gmb_wide_canlock) {
    long tries = 0;
    while(::lock(fhjhr, 0, 1) == -1) {
      if(PopupLocked(++tries, true, path().c_str()) == false) {
        gmb_wide_log->ErrLock();
        raw_close();
        gmb_wide_log->printf("! A JAM msgbase file could not be locked.");
        gmb_wide_log->printf(": %s.JHR.", path().c_str());
        gmb_wide_log->ErrOSInfo();
        LockErrorExit();
      }
    }
    if(tries)
      PopupLocked(0, 0, NULL);
  }
}


//  ------------------------------------------------------------------

void gmb_jam_area::lock() {

  GFTRK("gmb_jam_area::lock");

  lock_refs++;
  if(lock_refs == 1) {
    raw_lock();
    lseekset(fhjhr, 0);
    read(fhjhr, &hdrinfo, sizeof(JamHdrInfo));
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_jam_area::unlock() {

  GFTRK("gmb_jam_area::unlock");

  if(lock_refs) {
    if(lock_refs == 1) {
      if(gmb_wide_canlock)
        ::unlock(fhjhr, 0, 1);
      lseekset(fhjhr, 0);
      write(fhjhr, &hdrinfo, sizeof(JamHdrInfo));
    }
    lock_refs--;
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_jam_area::save_lastread() {

  GFTRK("gmb_jam_area::save_lastread");

  // Lock area while we are updating
  lock();

  // Update .JLR record
  lastrec.lastread = msgnset.relative_to_absolute(lastread);
  if(lastrec.lastread > lastrec.highread)
    lastrec.highread = lastrec.lastread;

  // Seek to users lastread position
  lseekset(fhjlr, lastpos, sizeof(JamLast));

  // Write the .JLR record
  write(fhjlr, &lastrec, sizeof(JamLast));

  // Update header info record
  hdrinfo.modcounter++;
  if(hdrinfo.basemsgnum == 0)
    hdrinfo.basemsgnum = 1;

  // Seek to beginning of the .JHR
  lseekset(fhjhr, 0);

  // Write header info record
  write(fhjhr, &hdrinfo, sizeof(JamHdrInfo));

  // Update userfile
  if(timesposted) {

    // Currently not supported ...
    timesposted = 0;
  }

  // Unlock area
  unlock();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_jam_area::add_subfield(JamHdr& hdr, byte*& subfield, word loid, word hiid, const gstring& data) {

  uint datlen = data.length();
  subfield = (byte*)throw_realloc(subfield, (uint)hdr.subfieldlen+sizeof(JamSubFieldHdr)+datlen);
  JamSubField* subfieldptr = (JamSubField*)(subfield + (uint)hdr.subfieldlen);
  subfieldptr->loid = loid;
  subfieldptr->hiid = hiid;
  subfieldptr->datlen = datlen;
  memcpy(subfieldptr->buffer, data.c_str(), datlen);
  hdr.subfieldlen += sizeof(JamSubFieldHdr) + datlen;
}


//  ------------------------------------------------------------------

void gmb_jam_area::save_message(int mode, gmsg& msg, JamHdr& hdr) {

  // Lock while updating
  lock();

  // Reset header
  memset(&hdr, 0, sizeof(JamHdr));

  // Init paragraph class
  GParagraph para;
  if(gmb_wide_dispsoftcr)
    para.softcr = SOFTCR;

  // Set default msg text offset and length
  hdr.offset = msg.jam.offset;
  hdr.txtlen = msg.jam.txtlen;

  // Get message text size
  char* txtcpy = NULL;
  char* txt = msg.text.lines_to_raw();
  if(txt and (mode & GMSG_TXT)) {

    // Work on a copy of the original msg text
    txtcpy = throw_strdup(txt);

    // Convert message text to a paragraph list
    para.ConvertText(txtcpy, strlen(txtcpy));

    // Mark control lines and calculate pure message text size
    hdr.txtlen = para.CheckCtrlLines();
  }

  // Determine where to write the message text
  if((mode & GMSG_NEW) or (hdr.txtlen > msg.jam.txtlen))
    hdr.offset = filelength(fhjdt);

  byte* subfield = NULL;
  char buf[1024];
  hdr.subfieldlen = msg.jam.subfieldlen;
  
  if(mode & GMSG_TXT) {

    // Build the subfields
    hdr.subfieldlen = 0;

    // Process addressing kludges
    int line = 0;
    gftnaddrs msgid_address;
    GParaData* pdptr = para.paraidx;
    while(line < para.lines) {
      if(pdptr->control > CTRL_KLUDGE) {
        word fmpt, topt;
        char buf1[201], buf2[201];
        switch(pdptr->control) {
          case CTRL_INTL:
            fmpt = msg.origaddr.point;
            topt = msg.destaddr.point;
            sscanf(pdptr->text+6, "%s %s", buf1, buf2);
            msg.destaddr.set(buf1);
            msg.origaddr.set(buf2);
            msg.origaddr.point = fmpt;
            msg.destaddr.point = topt;
            break;
          case CTRL_FMPT:
            msg.origaddr.point = atow(pdptr->text+6);
            break;
          case CTRL_TOPT:
            msg.destaddr.point = atow(pdptr->text+6);
            break;
          case CTRL_MSGID:
            msgid_address = pdptr->text+8;
            break;
        }
      }
      pdptr++;
      line++;
    }

    if(msgid_address.valid() and msgid_address.equal4d(msg.origaddr) and not msg.origaddr.domain)
      msg.origaddr.domain = msgid_address.domain;

    if(msg.origname)
      add_subfield(hdr, subfield, JAMSUB_SENDERNAME, 0, msg.origname);

    if(msg.origaddr.valid()) {
      strcpy(buf, msg.origaddr.c_str());
      add_subfield(hdr, subfield, JAMSUB_OADDRESS, 0, buf);
    }

    if(msg.destname)
      add_subfield(hdr, subfield, JAMSUB_RECEIVERNAME, 0, msg.destname);

    if(msg.destaddr.valid()) {
      strcpy(buf, msg.destaddr.c_str());
      add_subfield(hdr, subfield, JAMSUB_DADDRESS, 0, buf);
    }

    if(msg.subject)
      add_subfield(hdr, subfield, JAMSUB_SUBJECT, 0, msg.subject);

    // Convert kludges
    line = 0;
    pdptr = para.paraidx;
    while(line < para.lines) {
      if(pdptr->control > CTRL_KLUDGE) {
        uint offset = 0;
        word loid = 0;
        switch(pdptr->control) {
          case CTRL_INTL:
          case CTRL_FMPT:
          case CTRL_TOPT:
            // Strip these. Data is stored in the address subfields
            break;
          case CTRL_MSGID:
            loid = JAMSUB_MSGID;
            offset = 8;
            msg.kludge.msgid = pdptr->text+offset;
            break;
          case CTRL_REPLY:
            loid = JAMSUB_REPLYID;
            offset = 8;
            msg.kludge.reply = pdptr->text+offset;
            break;
          case CTRL_PID:
            loid = JAMSUB_PID;
            offset = 6;
            break;
          case CTRL_VIA:
            loid = JAMSUB_TRACE;
            offset = 6;
            break;
          case CTRL_SEENBY:
            loid = JAMSUB_SEENBY2D;
            offset = 9;
            break;
          case CTRL_SEENBY1:
            loid = JAMSUB_SEENBY2D;
            offset = 10;
            break;
          case CTRL_PATH:
            loid = JAMSUB_PATH2D;
            offset = 7;
            break;
          case CTRL_FLAGS:
            loid = JAMSUB_FLAGS;
            offset = 7;
            break;
          default:
            loid = JAMSUB_FTSKLUDGE;
            offset = 1;
        }
        if(offset)
          add_subfield(hdr, subfield, loid, 0, pdptr->text+offset);
      }
      pdptr++;
      line++;
    }
  }

  JamIndex idx;
  bool was_deleted = false;
  if(mode & GMSG_NEW) {
    msg.msgno = hdrinfo.basemsgnum + (filelength(fhjdx)/sizeof(JamIndex));
    hdr.messagenumber = msg.msgno;
    idx.hdroffset = filelength(fhjhr);
  }
  else {
    hdr.messagenumber = msg.msgno;
    lseekset(fhjdx, hdr.messagenumber-hdrinfo.basemsgnum, sizeof(JamIndex));
    read(fhjdx, &idx, sizeof(JamIndex));
    if(idx.hdroffset != 0xFFFFFFFFL) {
      lseekset(fhjhr, idx.hdroffset);
      JamHdr oldhdr;
      read(fhjhr, &oldhdr, sizeof(JamHdr));
      was_deleted = oldhdr.attribute & JAMATTR_DELETED ? true : false;
      if(oldhdr.subfieldlen != hdr.subfieldlen) {
        oldhdr.attribute |= JAMATTR_DELETED;
        oldhdr.txtlen = 0;
        lseekset(fhjhr, idx.hdroffset);
        write(fhjhr, &oldhdr, sizeof(JamHdr));
        idx.hdroffset = filelength(fhjhr);
      }
    }
  }

  if(idx.hdroffset != 0xFFFFFFFFL) {

    memcpy(hdr.signature, JAM_SIGNATURE, 4);
    hdr.revision = JAM_REVISION;

    hdr.replyto   = msg.link.to();
    hdr.reply1st  = msg.link.first();
    hdr.replynext = msg.link.next();

    hdr.cost      = msg.cost;
    hdr.timesread = msg.timesread;
    hdr.passwordcrc = 0xFFFFFFFFL;

    hdr.msgidcrc = strCrc32(strlwr(strcpy(buf, msg.kludge.msgid.c_str())), NO, CRC32_MASK_CCITT);
    hdr.replycrc = strCrc32(strlwr(strcpy(buf, msg.kludge.reply.c_str())), NO, CRC32_MASK_CCITT);
    idx.usercrc  = strCrc32(strlwr(strcpy(buf, msg.destname.c_str())),     NO, CRC32_MASK_CCITT);

    hdr.datewritten   = msg.date.written.unixtime();
    hdr.datereceived  = msg.date.received.unixtime();
    hdr.dateprocessed = msg.date.arrived.unixtime();

    hdr.attribute |= msg.attr.loc() ? JAMATTR_LOCAL       : 0;
    hdr.attribute |= msg.attr.trs() ? JAMATTR_INTRANSIT   : 0;
    hdr.attribute |= msg.attr.pvt() ? JAMATTR_PRIVATE     : 0;
    hdr.attribute |= msg.attr.rcv() ? JAMATTR_READ        : 0;
    hdr.attribute |= msg.attr.snt() ? JAMATTR_SENT        : 0;
    hdr.attribute |= msg.attr.k_s() ? JAMATTR_KILLSENT    : 0;
    hdr.attribute |= msg.attr.a_s() ? JAMATTR_ARCHIVESENT : 0;
    hdr.attribute |= msg.attr.hld() ? JAMATTR_HOLD        : 0;
    hdr.attribute |= msg.attr.cra() ? JAMATTR_CRASH       : 0;
    hdr.attribute |= msg.attr.imm() ? JAMATTR_IMMEDIATE   : 0;
    hdr.attribute |= msg.attr.dir() ? JAMATTR_DIRECT      : 0;
    hdr.attribute |= msg.attr.zon() ? JAMATTR_GATE        : 0;
    hdr.attribute |= msg.attr.frq() ? JAMATTR_FILEREQUEST : 0;
    hdr.attribute |= msg.attr.att() ? JAMATTR_FILEATTACH  : 0;
    hdr.attribute |= msg.attr.tfs() ? JAMATTR_TRUNCFILE   : 0;
    hdr.attribute |= msg.attr.kfs() ? JAMATTR_KILLFILE    : 0;
    hdr.attribute |= msg.attr.rrq() ? JAMATTR_RECEIPTREQ  : 0;
    hdr.attribute |= msg.attr.cfm() ? JAMATTR_CONFIRMREQ  : 0;
    hdr.attribute |= msg.attr.orp() ? JAMATTR_ORPHAN      : 0;
    hdr.attribute |= msg.attr.lok() ? JAMATTR_LOCKED      : 0;
    hdr.attribute |= msg.attr.del() ? JAMATTR_DELETED     : 0;

    if(isnet())
      hdr.attribute |= JAMATTR_TYPENET;
    else if(isecho())
      hdr.attribute |= JAMATTR_TYPEECHO;
    else
      hdr.attribute |= JAMATTR_TYPELOCAL;

    gftnaddrs daddress = msg.daddress;
    if((msg.destaddr != daddress) and daddress.valid())
      hdr.attribute |= JAMATTR_GATE;

    lseekset(fhjhr, idx.hdroffset);
    write(fhjhr, &hdr, sizeof(JamHdr));

    // Only write subfields if the msg body is written
    if(mode & GMSG_TXT)
      write(fhjhr, subfield, (uint)hdr.subfieldlen);

    // Delete msg if requested
    if(mode & GMSG_DELETE) {
      if(not was_deleted)
        hdrinfo.activemsgs--;
      hdr.attribute |= JAMATTR_DELETED;
      idx.usercrc = 0xFFFFFFFFL;
      if(wide->harddelete)
        idx.hdroffset = 0xFFFFFFFFL;
    }

    // Write msg index
    lseekset(fhjdx, hdr.messagenumber-hdrinfo.basemsgnum, sizeof(JamIndex));
    write(fhjdx, &idx, sizeof(JamIndex));

    // Free subfield buffer
    throw_release(subfield);

    // Update the header info
    if(mode & GMSG_NEW)
      hdrinfo.activemsgs++;
    hdrinfo.modcounter++;
    lseekset(fhjhr, 0);
    write(fhjhr, &hdrinfo, sizeof(JamHdrInfo));

    // If message text is used
    if(mode & GMSG_TXT) {

      char* txt = (char*)throw_malloc((uint)(hdr.txtlen+256));

      // Copy text paragraphs, excluding kludges
      int line = 0;
      char* tptr = txt;
      GParaData* pdptr = para.paraidx;
      while(line < para.lines) {
        if(pdptr->control < CTRL_KLUDGE) {
          memcpy(tptr, pdptr->text, pdptr->length);
          tptr += pdptr->length;
          *tptr++ = CR;
        }
        pdptr++;
        line++;
      }
      *tptr = NUL;

      // Seek to start position of the message text
      lseekset(fhjdt, hdr.offset);

      // Write the message text
      write(fhjdt, txt, (uint)hdr.txtlen);

      // Release the memory we have used
      throw_free(txtcpy);
      throw_free(txt);
    }

    // Update internals if new
    if(mode & GMSG_NEW) {

      // Count our msgs
      timesposted++;

      // Update internal array
      msgnset.append(msg.msgno);
      scan_count++;
    }

    // Update the NETMAIL/ECHOMAIL.JAM files if necessary (local and not sent)
    if((hdr.attribute & JAMATTR_LOCAL) and not (hdr.attribute & JAMATTR_SENT)) {
      if(isecho() or isnet()) {
        gfile fp;
        Path file;
        sprintf(file, "%s%sMAIL.JAM", wide->jampath.c_str(), isnet() ? "NET" : "ECHO");
        fp.fopen(file, "at+");
        if(fp.isopen()) {
          fp.printf("%s %lu\n", path().c_str(), msg.msgno);
          fp.fclose();
        }
      }
    }
  }
  else {
    scan();
  }

  // Unlock after updating
  unlock();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_jam_area::save_hdr(int mode, gmsg& msg) {

  GFTRK("gmb_jam_area::save_hdr");

  JamHdr hdr;
  save_message(mode | GMSG_HDR, msg, hdr);
}


//  ------------------------------------------------------------------

void gmb_jam_area::save_msg(int mode, gmsg& msg) {

  GFTRK("gmb_jam_area::save_msg");

  JamHdr hdr;
  save_message(mode | GMSG_HDRTXT, msg, hdr);
}


//  ------------------------------------------------------------------

void gmb_jam_area::del_msg(gmsg& msg) {

  GFTRK("gmb_jam_area::del_msg");

  JamHdr hdr;
  save_message(GMSG_HDR | GMSG_DELETE, msg, hdr);
}


//  ------------------------------------------------------------------

void gmb_jam_area::new_msgno(gmsg& msg) {

  GFTRK("gmb_jam_area::new_msgno");

  msg.msgno = hdrinfo.basemsgnum + (filelength(fhjdx)/sizeof(JamIndex));

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_jam_area::update_timesread(gmsg& msg) {

  GFTRK("gmb_jam_area::update_timesread");

  bool do_lock = gmb_wide_timesread & TIMESREAD_LOCK ? true : false;
  if(do_lock)
    lock();

  lseekset(fhjdx, msg.msgno-hdrinfo.basemsgnum, sizeof(JamIndex));

  JamIndex idx;
  read(fhjdx, &idx, sizeof(JamIndex));

  JamHdr hdr;
  lseekset(fhjhr, idx.hdroffset);
  read(fhjhr, &hdr, sizeof(JamHdr));

  hdr.timesread = msg.timesread;

  lseekset(fhjhr, idx.hdroffset);
  write(fhjhr, &hdr, sizeof(JamHdr));

  hdrinfo.modcounter++;
  lseekset(fhjhr, 0);
  write(fhjhr, &hdrinfo, sizeof(JamHdrInfo));

  if(do_lock)
    unlock();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

