
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gmbsqsh1.cpp,v 1.1.1.1 1998/11/21 19:38:13 odinn Exp $
//  ------------------------------------------------------------------
//  Squish msgbase handling and Maximus user functions.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------

#include <gdbgerr.h>
#include <gmemdbg.h>
#include <gdbgtrk.h>

#include <gmbsqsh.h>

              
//  ------------------------------------------------------------------

gmb_squish_manager::gmb_squish_manager() {

  userno = 0;
  direct = 0;
  recycle = SQUISHRECYCLE_YES;
  squishscan = SQS_QUICK;
}


//  ------------------------------------------------------------------

gmb_squish_manager::~gmb_squish_manager() {

  reset();
}


//  ------------------------------------------------------------------

gmb_area* gmb_squish_manager::new_area(gareacfg* c) {

  gmb_squish_area* area = new gmb_squish_area(c, this);
  throw_new(area);
  return area;
}


//  ------------------------------------------------------------------

void gmb_squish_manager::init_area(gmb_area* a, gareacfg* c) {

  gmb_squish_area* area = (gmb_squish_area*)a;
  area->init(c, this);
}


//  ------------------------------------------------------------------

void gmb_squish_manager::reset() {

  for(int n=0; n<active.size(); n++)
    delete active[n];
  active.reset();
}


//  ------------------------------------------------------------------

void gmb_squish_area::activate() {

  refs++;
  if(refs == 1) {
    open_refs = lock_refs = 0;
    softlock = false;
    fhsqd = fhsqi = -1;
    idx = NULL;
    memset(&base, 0, sizeof(base));
  }
}


//  ------------------------------------------------------------------

void gmb_squish_area::deactivate() {

  if(refs) {
    if(refs == 1) {
      if(open_refs) {
        open_refs = 1;
        close();
      }
    }
    refs--;
  }
}


//  ------------------------------------------------------------------

void gmb_squish_manager::initialize() {

  Path userfile;
  PathCopy(userfile, AddPath(userpath.c_str(), "USER.BBS"));
  const char* _username = gmb_wide_username[0];
  if(userno == -1) {
    user.fh = ::sopen(userfile, O_RDWR|O_CREAT|O_BINARY, gmb_wide_sharemode, S_STDRW);
    if(user.fh != -1) {
      user.find(_username);
      if(not user.found) {
        gmb_wide_log->printf("* User \"%s\" not found in %s.", _username, userfile);
        user.add(_username);
        gmb_wide_log->printf("* Now added with user number %u.", user.index);
      }
      close(user.fh);
    }
    userno = user.index;
  }
}


//  ------------------------------------------------------------------

void gmb_squish_area::raw_close() {

  GFTRK("SquishRawClose");

  if(fhsqi != -1)  ::close(fhsqi);  fhsqi = -1;
  if(fhsqd != -1)  ::close(fhsqd);  fhsqd = -1;

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

int gmb_squish_area::test_open(const char* __file) {

  GFTRK("SquishTestOpen");

  int _fh;
  long _tries = 0;

  do {

    _fh = ::sopen(__file, O_RDWR|O_BINARY|O_CREAT, gmb_wide_sharemode, S_STDRW);
    if(_fh == -1) {

      // Tell the world
      if(PopupLocked(++_tries, false, __file) == false) {

        // User requested to exit
        gmb_wide_log->ErrOpen();
        raw_close();
        gmb_wide_log->printf("! A Squish msgbase file could not be opened.");
        gmb_wide_log->printf(": %s.", __file);
        gmb_wide_log->ErrOSInfo();
        OpenErrorExit();
      }
    }
  } while(_fh == -1);

  // Remove the popup window
  if(_tries)
    PopupLocked(0, 0, NULL);

  GFTRK(NULL);

  return _fh;
}


//  ------------------------------------------------------------------

void gmb_squish_area::raw_open() {

  GFTRK("SquishRawOpen");

  fhsqd = test_open(AddPath(path().c_str(), ".SQD"));
  fhsqi = test_open(AddPath(path().c_str(), ".SQI"));

  GFTRK(NULL);
}


//  ------------------------------------------------------------------
//  Open the Squish message base

void gmb_squish_area::open() {

  GFTRK("SquishOpen");

  open_refs++;
  if(open_refs == 1) {
    raw_open();
    refresh();
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_squish_area::close() {

  GFTRK("SquishClose");

  if(open_refs) {
    if(open_refs == 1) {
      if(lock_refs) {
        lock_refs = 1;
        unlock();
      }
      raw_close();
      msgnset.reset();
      pmrkset.reset();
      throw_release(idx);
    }
    open_refs--;
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_squish_area::save_lastread() {

  GFTRK("SquishSaveLastread");

  int fh = ::sopen(AddPath(path().c_str(), ".SQL"), O_RDWR|O_CREAT|O_BINARY, gmb_wide_sharemode, S_STDRW);
  if(fh != -1) {
    ::lseekset(fh, wide->userno, sizeof(dword));
    dword lastread_msgn = msgnset.relative_to_absolute(lastread);
    ::write(fh, &lastread_msgn, sizeof(dword));
    ::close(fh);
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_squish_area::suspend() {

  GFTRK("SquishSuspend");

  save_lastread();
  raw_close();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_squish_area::resume() {

  GFTRK("SquishResume");

  raw_open();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

