
//  ------------------------------------------------------------------
//  The Goldware Library. Copyright (C) Odinn Sorensen.
//  ------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//  ------------------------------------------------------------------
//  $Id: gmbsqsh4.cpp,v 1.1.1.1 1998/11/21 19:38:13 odinn Exp $
//  ------------------------------------------------------------------
//  Squish msgbase handling and Maximus user functions.
//  ------------------------------------------------------------------


//  ------------------------------------------------------------------

#include <gdbgerr.h>
#include <gmemdbg.h>
#include <gdbgtrk.h>
#include <gstrall.h>
#include <gcrcall.h>

#include <gmbsqsh.h>

              
//  ------------------------------------------------------------------

void gmb_squish_area::raw_lock() {

  if(gmb_wide_canlock) {
    long tries = 0;
    while(::lock(fhsqd, 0, 1) == -1) {
      if(PopupLocked(++tries, true, path().c_str()) == false) {
        gmb_wide_log->ErrLock();
        raw_close();
        gmb_wide_log->printf("! A Squish msgbase file could not be locked.");
        gmb_wide_log->printf(": %s.SQD.", path().c_str());
        gmb_wide_log->ErrOSInfo();
        LockErrorExit();
      }
    }
    if(tries)
      PopupLocked(0, 0, NULL);
  }
}


//  ------------------------------------------------------------------

void gmb_squish_area::lock() {

  GFTRK("SquishLock");

  lock_refs++;
  if(lock_refs == 1) {
    raw_lock();
    refresh();
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_squish_area::unlock() {

  GFTRK("SquishUnlock");

  if(lock_refs) {
    if(lock_refs == 1) {
      if(gmb_wide_canlock)
        ::unlock(fhsqd, 0, 1);
      ::lseekset(fhsqd, 0);
      ::write(fhsqd, &base, sizeof(SqshBase));
      ::lseekset(fhsqi, 0);
      ::write(fhsqi, idx, (uint)(base.totalmsgs*sizeof(SqshIdx)));
      ::chsize(fhsqi, base.totalmsgs*sizeof(SqshIdx));
    }
    lock_refs--;
  }

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_squish_area::read_frm(dword offset, SqshFrm* frm) {

  lseekset(fhsqd, offset);
  read(fhsqd, frm, sizeof(SqshFrm));
}


//  ------------------------------------------------------------------

void gmb_squish_area::write_frm(dword offset, SqshFrm* frm) {

  lseekset(fhsqd, offset);
  write(fhsqd, frm, sizeof(SqshFrm));
}


//  ------------------------------------------------------------------

void gmb_squish_area::upd_frm_next(dword offset, SqshFrm* frm, dword next) {

  if(offset != SQFRAME_NULL) {
    read_frm(offset, frm);
    frm->next = next;
    write_frm(offset, frm);
  }
}


//  ------------------------------------------------------------------

void gmb_squish_area::upd_frm_prev(dword offset, SqshFrm* frm, dword prev) {

  if(offset != SQFRAME_NULL) {
    read_frm(offset, frm);
    frm->prev = prev;
    write_frm(offset, frm);
  }
}


//  ------------------------------------------------------------------

void gmb_squish_area::add_to_free_chain(dword delframe, SqshFrm* delfrm) {

  // Update free frame chain
  if((base.firstfreeframe == SQFRAME_NULL) or (base.lastfreeframe == SQFRAME_NULL)) {

    // No other free frames, so create new chain
    base.firstfreeframe = base.lastfreeframe = delframe;
    delfrm->prev = delfrm->next = SQFRAME_NULL;
  }
  else {

    // Insert this frame into the chain
    delfrm->next = SQFRAME_NULL;
    delfrm->prev = base.lastfreeframe;
    SqshFrm lastfrm;
    upd_frm_next(base.lastfreeframe, &lastfrm, delframe);
    base.lastfreeframe = delframe;
  }

  // Write the deleted frame
  delfrm->type = SQFRAME_FREE;
  //gmb_wide_log->printf("- Deleted frame 0x%08lX of length %lu.", delframe, delfrm->length);
  write_frm(delframe, delfrm);
}


//  ------------------------------------------------------------------

void gmb_squish_area::delete_msg(uint reln) {

  GFTRK("SquishDeleteMsg");

  lock();

  // Load the frame to be deleted
  SqshFrm delfrm;
  dword delframe = idx[reln].offset;
  read_frm(delframe, &delfrm);

  // Chain the previous and next frames together
  SqshFrm tmpfrm;
  upd_frm_next(delfrm.prev, &tmpfrm, delfrm.next);
  upd_frm_prev(delfrm.next, &tmpfrm, delfrm.prev);

  // Update base data if the first or last frame was deleted
  if(delframe == base.firstframe)
    base.firstframe = delfrm.next;
  if(delframe == base.lastframe)
    base.lastframe = delfrm.prev;

  // Add the deleted msg to the free frame chain
  add_to_free_chain(delframe, &delfrm);

  // Remove deleted message number from the indexes
  msgnset.delete_relative(reln+1);
  dword tomove = base.totalmsgs - reln - 1;
  memmove(idx+reln, idx+reln+1, (uint)(tomove*sizeof(SqshIdx)));

  // Update base data
  base.highestmsg--;
  base.totalmsgs--;

  // Update area data
  if(lastread)
    lastread--;
  
  unlock();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_squish_area::init_frm(SqshFrm* frm) {

  memset(frm, 0, sizeof(SqshFrm));
  frm->type = SQFRAME_NORMAL;
  frm->id = SQFRAMEID;
}


//  ------------------------------------------------------------------
//  Copy the text itself to a buffer, or count its length if out==NULL

uint gmb_squish_copytobuf(char* p, char* out, char** end) {

  if(out)
    *out++ = CTRL_A;

  uint len = 1;

  while((*p==CR) or (*p==LF) or (*p==SOFTCR))
    p++;

  while((*p==CTRL_A) or (strncmp(p, "AREA:", 5)==0)) {

    // Skip over the first ^A
    if(*p == CTRL_A)
      p++;

    while(*p and (*p != CR) and (*p != LF) and (*p != SOFTCR)) {
      if(out)
        *out++ = *p;
      len++;
      p++;
    }

    if(out)
      *out++ = CTRL_A;

    len++;

    while((*p==LF) or (*p==SOFTCR))
      p++;
    if(*p == CR)
      p++;
    while((*p==LF) or (*p==SOFTCR))
      p++;
  }

  // Nul-term the string
  if(out)
    *out = NUL;

  len++;

  // Make sure to leave no trailing CTRL_A's.
  if(out and (out[-1]==CTRL_A))
    out[-1] = NUL;
  

  // Now store the new end location of the kludge lines
  if(end)
    *end = p;
  
  return len;
}


//  ------------------------------------------------------------------

char* gmb_squish_copytocontrolbuf(char* txt, char** newtext, uint* length) {

  // Figure out how long the control info is
  uint ctlsize = gmb_squish_copytobuf(txt, NULL, NULL);

  // Allocate memory for it
  char* cbuf = (char*)throw_calloc(1, ctlsize+20);

  // Now copy the text itself
  char* end;
  gmb_squish_copytobuf(txt, cbuf, &end);

  if(length)
    *length -= (uint)(end-txt);
  if(newtext)
    *newtext = end;

  return cbuf;
}


//  ------------------------------------------------------------------

void gmb_squish_area::excess_frm(dword lastframe, dword newframe, SqshFrm* newfrm, dword totsize) {

  // Is the excess length large enough for a frame and message header?
  dword excesslength = newfrm->length - totsize;
  if(excesslength >= sizeof(SqshFrm)) {

    // Calculate frame offset of the excess frame
    dword exframe = newframe + totsize + sizeof(SqshFrm);

    // Adjust base data if this becomes the last free frame
    if(lastframe == base.lastfreeframe)
      base.lastfreeframe = exframe;

    // Setup the excess frame and write it
    SqshFrm exfrm;
    init_frm(&exfrm);
    exfrm.type = SQFRAME_FREE;
    exfrm.next = newfrm->next;
    exfrm.prev = newframe;
    newfrm->next = exframe;
    newfrm->length = totsize;
    exfrm.length = excesslength - sizeof(SqshFrm);
    write_frm(exframe, &exfrm);
    SqshFrm tmpfrm;
    upd_frm_prev(exfrm.next, &tmpfrm, exframe);
    //gmb_wide_log->printf("- Created excess free frame 0x%08lX of length %lu.", exframe, exfrm.length);
  }
}


//  ------------------------------------------------------------------

uint gmb_squish_area::find_msgn(ulong tagn) {

  if(idx) {

    register SqshIdx* tag = idx;
    register uint tags = (uint)base.totalmsgs;

    if(tagn and tags and (tagn > tag[tags-1].msgno))
      return 0;

    if(tags and tagn) {

      register long mid;
      register long left = 0;
      register long right = tags;

      do {
        mid = (left+right)/2;
        if(tagn < tag[(uint)mid].msgno)
          right = mid - 1;
        else if(tagn > tag[(uint)mid].msgno)
          left = mid + 1;
        else
          return (uint)(mid + 1);
      } while(left < right);

      if(tagn == tag[(uint)left].msgno)
        return (uint)(left + 1);
    }
  }

  return 0;
}


//  ------------------------------------------------------------------

void gmb_squish_area::save_message(int mode, gmsg& msg, SqshHdr& hdr) {

  lock();

  uint reln = (mode & GMSG_NEW) ? 0 : (msgnset.absolute_to_relative(msg.msgno) - 1);

  // Reset header
  memset(&hdr, 0, sizeof(SqshHdr));

  // Convert attributes
  hdr.attr |= MSGUID;
  hdr.attr |= msg.attr.pvt() ? MSGPRIVATE : 0;
  hdr.attr |= msg.attr.cra() ? MSGCRASH   : 0;
  hdr.attr |= msg.attr.rcv() ? MSGREAD    : 0;
  hdr.attr |= msg.attr.snt() ? MSGSENT    : 0;
  hdr.attr |= msg.attr.att() ? MSGFILE    : 0;
  hdr.attr |= msg.attr.trs() ? MSGFWD     : 0;
  hdr.attr |= msg.attr.orp() ? MSGORPHAN  : 0;
  hdr.attr |= msg.attr.k_s() ? MSGKILL    : 0;
  hdr.attr |= msg.attr.loc() ? MSGLOCAL   : 0;
  hdr.attr |= msg.attr.hld() ? MSGHOLD    : 0;
  hdr.attr |= msg.attr.rsv() ? MSGXX2     : 0;
  hdr.attr |= msg.attr.frq() ? MSGFRQ     : 0;
  hdr.attr |= msg.attr.rrq() ? MSGRRQ     : 0;
  hdr.attr |= msg.attr.rrc() ? MSGCPT     : 0;
  hdr.attr |= msg.attr.arq() ? MSGARQ     : 0;
  hdr.attr |= msg.attr.urq() ? MSGURQ     : 0;
  hdr.attr |= msg.attr.prn() ? MSGPRINTED : 0;
  hdr.attr |= msg.attr.lok() ? MSGLOK     : 0;
  hdr.attr |= msg.timesread  ? MSGSEEN    : 0;

  if(msg.attr.scn() and not msg.attr.uns())
    hdr.attr |= MSGSCANNED;

  if(msg.attr.dir() and wide->direct)
    hdr.attr |= MSGCRASH | MSGHOLD;

  strxcpy(hdr.from, msg.origname.c_str(), sizeof(hdr.from));
  strxcpy(hdr.to,   msg.destname.c_str(), sizeof(hdr.to));
  strxcpy(hdr.subj, msg.subject.c_str(), sizeof(hdr.subj));

  hdr.orig.zone  = msg.origaddr.zone;
  hdr.orig.net   = msg.origaddr.net;
  hdr.orig.node  = msg.origaddr.node;
  hdr.orig.point = msg.origaddr.point;

  hdr.dest.zone  = msg.destaddr.zone;
  hdr.dest.net   = msg.destaddr.net;
  hdr.dest.node  = msg.destaddr.node;
  hdr.dest.point = msg.destaddr.point;

  hdr.replyto    = msg.link.to();
  hdr.replies[0] = msg.link.first();
  for(int r=1; r<=8; r++)
    hdr.replies[r] = msg.link.list(r-1);
  hdr.umsgid = (mode & GMSG_NEW) ? base.nextmsgno : msg.msgno;

  hdr.date_written = msg.date.written.opustime();
  hdr.date_arrived = msg.date.arrived.opustime();
  hdr.utc_offset = (short)msg.date.written.utc_offset_minutes();
  strcpy(hdr.ftsc_date, msg.date.written.fts1());

  // Setup some local variables for speed
  dword hash = strHash32(hdr.to) | ((hdr.attr & MSGREAD) ? 0x80000000LU : 0);

  // Writing msg text?
  if(mode & GMSG_TXT) {

    char* txt = msg.text.lines_to_raw();
    uint usize = strlen(txt) + 1;
    char* ctl = gmb_squish_copytocontrolbuf(txt, &txt, &usize);
    dword txtsize = strlen(txt) + 1;
    dword ctlsize = strlen(ctl) + 1;
    dword totsize = sizeof(SqshHdr) + ctlsize + txtsize;

    SqshFrm oldfrm, newfrm;
    dword newframe = SQFRAME_NULL;
    dword oldframe = idx ? idx[reln].offset : base.endframe;
    if(not (mode & GMSG_NEW)) {

      // Get the original frame and see if there is still room for the msg
      read_frm(oldframe, &oldfrm);
      if(oldfrm.length >= totsize) {
        newframe = oldframe;
        newfrm = oldfrm;
      }
    }

    // It's a new message or the changed message doesn't fit
    if(newframe == SQFRAME_NULL) {

      // If there is a max msgs limit and are we writing a new
      // msg, delete msgs to (hopefully) make room for this msg
      if(base.maxmsgs and (mode & GMSG_NEW))
        while(base.maxmsgs <= base.totalmsgs)
          delete_msg((uint)base.protmsgs);

      // Locate a free frame, if possible
      newframe = base.firstfreeframe;
      //gmb_wide_log->printf("- Looking for a frame of at least length %lu.", totsize);
      while(1) {

        // At end of free frames?
        if(newframe == SQFRAME_NULL) {
          newframe = base.endframe;
          init_frm(&newfrm);
          //gmb_wide_log->printf("- Allocated new frame 0x%08lX of length %lu.", newframe, totsize);
          break;
        }

        // Is this frame large enough in itself?
        read_frm(newframe, &newfrm);
        //gmb_wide_log->printf("- Found free frame 0x%08lX of length %lu.", newframe, newfrm.length);
        if(newfrm.length >= totsize) {

          // Create excess frame if possible
          if(wide->recycle == SQUISHRECYCLE_YES) {
            excess_frm(newframe, newframe, &newfrm, totsize);
            //gmb_wide_log->printf("- Frame was large enough (%lu bytes wasted).", newfrm.length - totsize);
          }
          break;
        }

        // If two frames are adjacent, try to merge them to make more room
        if(wide->recycle and (wide->recycle != SQUISHRECYCLE_MSGAPI2)) {
          dword lastframe = SQFRAME_NULL;
          while((newfrm.next == (newframe+newfrm.length+sizeof(SqshFrm))) and (newfrm.length < totsize)) {
            SqshFrm lastfrm;
            read_frm(newfrm.next, &lastfrm);
            newfrm.length += lastfrm.length + sizeof(SqshFrm);
            //gmb_wide_log->printf("- Merged frames 0x%08lX and 0x%08lX. New length: %lu.", newframe, newfrm.next, newfrm.length);
            lastframe = newfrm.next;
            newfrm.next = lastfrm.next;
          }

          // Did we get a large enough frame?
          if(newfrm.length >= totsize) {

            // Create excess frame if possible
            if(wide->recycle == SQUISHRECYCLE_YES) {
              excess_frm(lastframe, newframe, &newfrm, totsize);
              //gmb_wide_log->printf("- Merged frame was large enough (%lu bytes wasted).", newfrm.length - totsize);
            }

            // If one of the frames in our chain was the last free frame,
            // set the last free frame to the one we've merged it into,
            // for later a clean up effort.
            if(lastframe == base.lastfreeframe)
              base.lastfreeframe = newframe;

            // Got a free frame
            break;
          }
        }

        // Go to next free frame and try again
        newframe = newfrm.next;
      }

      // If this was the first frame (ie. the first one pointed to by
      // firstfreeframe, which means that the first frame found was long
      // enough to hold the message), then set the free pointer to the
      // start of the new free chain.
      if(newframe == base.firstfreeframe)
        base.firstfreeframe = newfrm.next;
      if(newframe == base.lastfreeframe)
        base.lastfreeframe = newfrm.prev;

      // Now update the linked list of free frames, to remove the current
      // frame from the free-frame list, if necessary.  We only need to do
      // this if the current frame wasn't just being appended to the .SQD
      // file, since there would be no links to update in that case.
      if(newframe != base.endframe) {
        SqshFrm tmpfrm;
        upd_frm_next(newfrm.prev, &tmpfrm, newfrm.next);
        upd_frm_prev(newfrm.next, &tmpfrm, newfrm.prev);
      }

      if(mode & GMSG_NEW) {

        // Link the frame to the last frame
        newfrm.prev = base.lastframe;
        newfrm.next = SQFRAME_NULL;
        SqshFrm tmpfrm;
        upd_frm_next(newfrm.prev, &tmpfrm, newframe);
        if(base.firstframe == SQFRAME_NULL)
          base.firstframe = newframe;
        base.lastframe = newframe;
      }
      else {

        // Rewriting old message
        newfrm.next = oldfrm.next;
        newfrm.prev = oldfrm.prev;
        add_to_free_chain(oldframe, &oldfrm);
        SqshFrm tmpfrm;
        upd_frm_next(newfrm.prev, &tmpfrm, newframe);
        upd_frm_prev(newfrm.next, &tmpfrm, newframe);
        if(base.firstframe == oldframe)
          base.firstframe = newframe;
        if(base.lastframe == oldframe)
          base.lastframe = newframe;
      }

      // Set the frame length only if this is a brand new frame
      if(newframe == base.endframe) {
        newfrm.length = totsize;
        base.endframe += sizeof(SqshFrm) + totsize;
      }
    }

    // Set sizes in the frame
    newfrm.totsize = totsize;
    newfrm.ctlsize = ctlsize;
    newfrm.type = SQFRAME_NORMAL;

    // Write frame, header, control info and message text
    write_frm(newframe, &newfrm);
    write(fhsqd, &hdr, sizeof(SqshHdr));
    write(fhsqd, ctl, (uint)ctlsize);
    write(fhsqd, txt, (uint)txtsize);
    throw_free(ctl);

    // Update internal arrays if new
    if(mode & GMSG_NEW) {
      base.highestmsg++;
      reln = (uint)(base.totalmsgs++);
      msg.msgno = base.nextmsgno++;
      msgnset.append(msg.msgno);
      scan_count++;
      idx = (SqshIdx*)throw_realloc(idx, (uint)(base.totalmsgs*sizeof(SqshIdx)));
    }

    // Update index
    SqshIdx* idxp = idx + reln;
    idxp->offset = newframe;
    idxp->msgno = msg.msgno;
    idxp->hash = hash;
  }
  else {

    // Just update the header
    idx[reln].hash = hash;
    lseekset(fhsqd, idx[reln].offset+sizeof(SqshFrm));
    write(fhsqd, &hdr, sizeof(SqshHdr));
  }

  // Adjust the highwatermark if required
  if(msg.attr.uns())
    if(base.highwatermark >= msg.msgno)
      base.highwatermark = msg.msgno - 1;

  unlock();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_squish_area::save_hdr(int mode, gmsg& msg) {

  GFTRK("SquishSaveHdr");

  SqshHdr hdr;
  save_message(mode|GMSG_HDR, msg, hdr);
}


//  ------------------------------------------------------------------

void gmb_squish_area::save_msg(int mode, gmsg& msg) {

  GFTRK("SquishSaveMsg");

  SqshHdr hdr;
  save_message(mode|GMSG_HDRTXT, msg, hdr);
}


//  ------------------------------------------------------------------

void gmb_squish_area::del_msg(gmsg& msg) {

  GFTRK("SquishDelMsg");

  delete_msg(msgnset.absolute_to_relative(msg.msgno)-1);

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

void gmb_squish_area::new_msgno(gmsg& msg) {

  msg.msgno = base.nextmsgno;
}


//  ------------------------------------------------------------------

void gmb_squish_area::update_timesread(gmsg& msg) {

  GFTRK("gmb_squish_area::update_timesread");

  int do_lock = gmb_wide_timesread & TIMESREAD_LOCK;
  if(do_lock) {
    lock_refs++;
    raw_lock();
  }

  uint reln = msgnset.absolute_to_relative(msg.msgno) - 1;

  dword frame = idx[reln].offset;

  SqshHdr hdr;
  ::lseekset(fhsqd, frame+sizeof(SqshFrm));
  ::read(fhsqd, &hdr, sizeof(SqshHdr));

  hdr.attr |= msg.timesread ? MSGSEEN : 0;

  ::lseekset(fhsqd, frame+sizeof(SqshFrm));
  ::write(fhsqd, &hdr, sizeof(SqshHdr));

  if(do_lock)
    unlock();

  GFTRK(NULL);
}


//  ------------------------------------------------------------------

